#include <Platform_Lib/SubtitleParser/SubtitleParser.h>
#include <string.h>
#include <Platform_Lib/SubtitleParser/SRTParser.h>
#include <Platform_Lib/SubtitleParser/SMIParser.h>
#include <Platform_Lib/SubtitleParser/SSAParser.h>
#include <Platform_Lib/SubtitleParser/IDXSUBParser.h>
#include <Platform_Lib/SubtitleParser/TEXTSUBParser.h>
#include <Platform_Lib/SubtitleParser/LRCParser.h>
#include <Platform_Lib/SubtitleParser/SubViewerParser.h>
#include <Platform_Lib/Graphics/DirectGraphics.h>
#include <Platform_Lib/FontRenderer/render.h>
#include <Utility/ResourceMgr/resource.h>

// temporary solution for HTML text
static void filter_out_tags(char *input, char *output, size_t output_size) {
	char *pivot1, *pivot2;
	bool valid_tag;

	output[0] = '\0';
	pivot1 = &(input[0]);

	while(*pivot1 != '\0' && output_size > 0) {
		if((pivot2 = strchr(pivot1, '<')) != NULL) {
			if(*(pivot2 + 1) == ' ')
				valid_tag = false;
			else
				valid_tag = true;

			if(pivot2 != pivot1) {
				size_t copySize = (size_t)(pivot2 - pivot1) > output_size ? output_size : (pivot2 - pivot1);

				strncat(output, pivot1, copySize);
				output_size -= copySize;
			}
		}
		else {
			strncat(output, pivot1, output_size);
			break;
		}

		do {
			pivot2++;
		} while(*pivot2 != '>' && *pivot2 != '\0');

		if(*pivot2 == '\0' || valid_tag == false) {
			size_t copySize = (size_t)(pivot2 - pivot1 + 1) > output_size ? output_size : (pivot2 - pivot1 + 1);

			strncat(output, pivot1, copySize);
			output_size -= copySize;
		}

		if(*pivot2 == '>')
			pivot1 = pivot2+1;
	}
	return;
}

void replace_pattern(char *input, char *output, size_t output_size, char *needle, char *pattern) {
	char *pivot1, *pivot2;

	size_t needleSize = strlen(needle);
	size_t patternSize = strlen(pattern);

	if(strlen(input) > 2048)
		return;

	output[0] = '\0';
	pivot1 = &(input[0]);

	while(*pivot1 != '\0' && output_size > 0) {
		if((pivot2 = strstr(pivot1, needle)) != NULL) {
			size_t copySize;

			if(pivot2 != pivot1) {
				copySize = (size_t)(pivot2 - pivot1) > output_size ? output_size : (pivot2 - pivot1);

				strncat(output, pivot1, copySize);
				output_size -= copySize;
			}

			copySize = patternSize < output_size ? patternSize : output_size;

			strncat(output, pattern, copySize);
			pivot1 = pivot2 + needleSize;
		}
		else {
			strncat(output, pivot1, output_size);
			break;
		}
	}
	return;
}

SubtitleParser::SubtitleParser(char *sub_filename, SUBTITLE_TYPE type,SUBTITLE_STYLE_SHEET sheet) : 
 mType(type),
 mSS(sheet),
 mEngine(NULL) {
	switch(mType) {
		case SUBTITLE_SRT:
			try {
				mEngine = new SRTParser();
			}
			catch(...) {
				throw -1;
			}
			break;
		 case SUBTITLE_SMI:
        	try {
				mEngine = new SMIParser();
			}
			catch(...) {
				throw -1;
			}
			break;
		case SUBTITLE_SSA:
			try {
				mEngine	 = new SSAParser();
			}
			catch(...) {
				throw -1;
			}
			break;
		case SUBTITLE_ASS:
			try {
				mEngine	 = new SSAParser();
			}
			catch(...) {
				throw -1;
			}
			break;
		case SUBTITLE_IDXSUB:
			try {
				mEngine = new IDXSUBParser();
			}
			catch(...) {
				throw -1;
			}
			break;
		case SUBTITLE_TEXTSUB:
			try {
				mEngine = new TEXTSUBParser();
			}
			catch(...) {
				throw -1;
			}
			break;
		case SUBTITLE_LRC:
			try {
				mEngine = new LRCParser();
			}
			catch(...) {
				throw -1;
			}
			break;
		case SUBTITLE_SUBVIEWER:
			try {
				mEngine = new SubViewerParser();
			}
			catch(...) {
				throw -1;
			}
			break;
		default:
			printf("WARNING: Not Supported Subtitle Format\n");
			throw -1;
	}
	if(mSS.SubFrom == SUBTITLE_EXTERNAL){
		if(mEngine->Open(sub_filename) == false) {
			delete mEngine;
			throw -1;
		}
	}
		
}

SubtitleParser::~SubtitleParser() {
	if(mEngine != NULL) {
		if(mSS.SubFrom == SUBTITLE_EXTERNAL){
		mEngine->Close();
		}
		delete mEngine;
		mEngine = NULL;
	}
}
class TEXTSUBParser;
int SubtitleParser::GetSubtitle(int timeMs, SUBTITLE_STYLE_SHEET *stylesheet, char **text,float fps, void *SubBuf) {
	if(mEngine != NULL){
		if(mType != SUBTITLE_TEXTSUB){
			if(mType==SUBTITLE_SRT)
				return dynamic_cast<SRTParser *>(mEngine)->GetSubtitle(timeMs, stylesheet, text,SubBuf);
			else if(mType == SUBTITLE_SUBVIEWER) 
				return dynamic_cast<SubViewerParser *>(mEngine)->GetSubtitle(timeMs, stylesheet, text, SubBuf);
			else if((mType==SUBTITLE_SSA)||(mType==SUBTITLE_ASS))
				return dynamic_cast<SSAParser *>(mEngine)->GetSubtitle(timeMs, stylesheet, text,SubBuf);
			else 
				return mEngine->GetSubtitle(timeMs, stylesheet, text);
		}
		else
			return dynamic_cast<TEXTSUBParser *>(mEngine)->GetSubtitle(timeMs, stylesheet, text,fps);
	}
	else
		return false;
}

int SubtitleParser::SubtitleSeek(int timeMs, SUBTITLE_STYLE_SHEET *stylesheet, char **text,float fps,void *SubBuf) {
	if(mEngine != NULL){
		if(mType != SUBTITLE_TEXTSUB){
			if(mType==SUBTITLE_SRT)
				return dynamic_cast<SRTParser *>(mEngine)->SubtitleSeek(timeMs, stylesheet, text,SubBuf);
			else if(mType == SUBTITLE_SUBVIEWER) 
				return dynamic_cast<SubViewerParser *>(mEngine)->SubtitleSeek(timeMs, stylesheet, text, SubBuf);
			else if((mType==SUBTITLE_SSA)||(mType==SUBTITLE_ASS))
				return dynamic_cast<SSAParser *>(mEngine)->SubtitleSeek(timeMs, stylesheet, text,SubBuf);
			else 
				return mEngine->SubtitleSeek(timeMs, stylesheet, text);
		}
		else
			return dynamic_cast<TEXTSUBParser *>(mEngine)->GetSubtitle(timeMs, stylesheet, text,fps);
	}
	else
		return false;
}
int SubtitleParser::SwitchLang(char *LangName){
	if(mEngine != NULL){
		if(mType == SUBTITLE_IDXSUB){
			return dynamic_cast<IDXSUBParser *>(mEngine)->SwitchLang(LangName);
		}
		else if(mType == SUBTITLE_SMI){
			return dynamic_cast<SMIParser *>(mEngine)->SwitchLang(LangName);
		}
	}
	return -1;
}

int SubtitleParser::GetDefaultLang(char *LangName){
	if(mEngine != NULL){
		if(mType == SUBTITLE_IDXSUB){
			return dynamic_cast<IDXSUBParser *>(mEngine)->GetDefaultLang(LangName);
		}
		else if(mType == SUBTITLE_SMI){
			return dynamic_cast<SMIParser *>(mEngine)->GetDefaultLang(LangName);
		}
	}
	return -1;
}

int SubtitleParser::GetLangNum(){
	if(mEngine != NULL) {
		if(mType == SUBTITLE_IDXSUB)
			return dynamic_cast<IDXSUBParser *>(mEngine)->GetLangNum();
		else if(mType == SUBTITLE_SMI)
			return dynamic_cast<SMIParser *>(mEngine)->GetLangNum();
	}

	return -1;
}

	
HRESULT SubtitleParser::DrawOntoSurface(SURFACEHANDLE handleDisplay,
					int left,
					int top,
					int width,
					int height,
					int from_x,
					int from_y,
					SUBTITLE_STYLE_SHEET stylesheet,
					char *string) {
	FontRenderer *pFont;
	int FLAG = H_CENTER|AM_MULTILINE|MIXED_SCHEME;

	if(stylesheet.fontStyle.encoding == DBCS_GBK)
		(ResourceMgr::GetInstance())->GetFontRenderer(&pFont, OSD_LANG_SCHINESE);
	else if(stylesheet.fontStyle.encoding == DBCS_BIG5 || stylesheet.fontStyle.encoding == UTF8)
		(ResourceMgr::GetInstance())->GetFontRenderer(&pFont, OSD_LANG_CHINESE);
	else
		(ResourceMgr::GetInstance())->GetFontRenderer(&pFont, OSD_LANG_ENGLISH);

	pFont->SetHeight(stylesheet.fontStyle.fontSize);

	if(stylesheet.fontStyle.alignment == 0) {
		unsigned int lineNum, yOffset;
		int fontSize = pFont->GetLineHeight();
		
		lineNum = pFont->GetTextLineNumber(string, 
						stylesheet.fontStyle.encoding,
						width, 
						FLAG);

		if(lineNum > 1)
			yOffset = fontSize * (lineNum * 4 + 3) / 4;
		else
			yOffset = fontSize * 11 / 4;

		if((unsigned int)height >= yOffset) {
			top = height - yOffset;
			height = yOffset;
		}

		FLAG |= V_TOP;
	}
	else if(stylesheet.fontStyle.alignment == 1)
		FLAG |= V_BOTTOM;
	else if(stylesheet.fontStyle.alignment == 2)
		FLAG |= V_CENTER;
	else if(stylesheet.fontStyle.alignment == 3)
		FLAG |= V_TOP;

	if(stylesheet.fontStyle.outline == true)
		FLAG |= EFF_OUTLINE_THICK;

	if(stylesheet.type == PLAIN) {
		MYRECT srcRect = {left, top, width, height};
		MYRECT surRect, dstRect;

		SURFACEDESC surface;

		DG_GetSurfaceDesc(handleDisplay, &surface);
		surRect.x = 0;
		surRect.y = 0;
		surRect.w = surface.dwWidth;
		surRect.h = surface.dwHeight;

		dstRect = GetIntersectRect(&srcRect, &surRect);

		DG_DrawRectangle(handleDisplay, 
						dstRect.x, 
						dstRect.y, 
						dstRect.w, 
						dstRect.h, 
						stylesheet.fontStyle.backColor);

		if(stylesheet.fontStyle.outline == true) 
			pFont->DG_TextOut(handleDisplay, 
						dstRect.x,
						dstRect.y,
						dstRect.w,
						dstRect.h, 
						stylesheet.fontStyle.primaryColor,
						stylesheet.fontStyle.outlineColor,
						string,
						stylesheet.fontStyle.encoding,
						FLAG);
		else
			pFont->DG_TextOut(handleDisplay, 
						dstRect.x,
						dstRect.y,
						dstRect.w,
						dstRect.h, 
						stylesheet.fontStyle.primaryColor,
						stylesheet.fontStyle.backColor,
						string,
						stylesheet.fontStyle.encoding,
						FLAG);

		return S_OK;
	}
	else if(stylesheet.type == HTML) {
		char buffer[2048], buffer2[2048];
		MYRECT srcRect = {left, top, width, height};
		MYRECT surRect, dstRect;

		SURFACEDESC surface;

		DG_GetSurfaceDesc(handleDisplay, &surface);
		surRect.x = 0;
		surRect.y = 0;
		surRect.w = surface.dwWidth;
		surRect.h = surface.dwHeight;

		dstRect = GetIntersectRect(&srcRect, &surRect);

		// replace &nbsp; with ' ' [string -> buffer]
		replace_pattern(string, buffer, 2048, "&nbsp;", " ");
		// replace <br> with '\n' [buffer -> buffer2]
		replace_pattern(buffer, buffer2, 2048, "<br>", "\n");
		// filter out <> </> tags [buffer2 -> buffer]
		filter_out_tags(buffer2, buffer, 2048);
		// replace &gt; <> </> tags [buffer -> buffer2]
		replace_pattern(buffer, buffer2, 2048, "&gt;", ">");
		// replace &lt; <> </> tags [buffer2 -> buffer]
		replace_pattern(buffer2, buffer, 2048, "&lt;", "<");

		DG_DrawRectangle(handleDisplay, 
						dstRect.x, 
						dstRect.y, 
						dstRect.w, 
						dstRect.h, 
						stylesheet.fontStyle.backColor);

		if(stylesheet.fontStyle.outline == true) 
			pFont->DG_TextOut(handleDisplay, 
						dstRect.x,
						dstRect.y,
						dstRect.w,
						dstRect.h, 
						stylesheet.fontStyle.primaryColor,
						stylesheet.fontStyle.outlineColor,
						buffer,
						stylesheet.fontStyle.encoding,
						FLAG);
		else
			pFont->DG_TextOut(handleDisplay, 
						dstRect.x,
						dstRect.y,
						dstRect.w,
						dstRect.h, 
						stylesheet.fontStyle.primaryColor,
						stylesheet.fontStyle.backColor,
						buffer,
						stylesheet.fontStyle.encoding,
						FLAG);

		return S_OK;
	}
	else
		if(stylesheet.type == BITMAP) 
		{
		SURFACEDESC surface;
		char *psurf,*pindex,*pAlpha;
		DG_GetSurfaceDesc(handleDisplay, &surface);
		psurf = (char *) surface.lpSurface;
		if((left >= (int)surface.dwWidth)||(top >= (int)surface.dwHeight))
			return E_FAIL;
		if((from_x >= (int)stylesheet.imagewidth)||(from_y >= (int)stylesheet.imageheight ))
			return E_FAIL;
		
		if((from_x+width) >= (int)stylesheet.imagewidth)
			width = stylesheet.imagewidth - from_x;
		if((left+width) >= (int)surface.dwWidth)
			width = surface.dwWidth - left;
			
		if((from_y+height) >= (int)stylesheet.imageheight)
			height = stylesheet.imageheight - from_y;
		if((top+height) >= (int)surface.dwHeight)
			height = surface.dwHeight - top;
		
		pindex = string+stylesheet.imagewidth*from_y+from_x;
		pAlpha= string+stylesheet.imageheight *stylesheet.imagewidth+stylesheet.imagewidth*from_y+from_x;

	
		DG_Lock(handleDisplay, 0, 0,surface.dwWidth, surface.dwHeight, 0);
		
	if(surface.pixelFormat == Format_8) {	// RGB332
		int8_t R,G,B;
		
		for(int i=0;i<height;i++)
		{
			for(int j = 0;j<width;j++)
			{
			 unsigned char index = (unsigned char)*pindex++;
			R = (stylesheet.palette[index]>>16) & 0xe0;
			G = (stylesheet.palette[index]>>8) & 0xe0;
			B =  stylesheet.palette[index] & 0xc0;
		 	*(psurf+surface.lPitch*(i+top)+(j+left)) =  R|(G>>3)|(B>>6);
			}
			pindex +=(stylesheet.imagewidth-width);
		}
		
	}
	else if(surface.pixelFormat == Format_16) {	// RGB565
		int16_t R,G,B;
//		printf("11111111111111111\n");
		for(int i=0;i<height;i++)
		{
			for(int j = 0;j<width;j++)
			{
			 	unsigned char index = (unsigned char)*pindex++;
			 	unsigned char Alpha = (unsigned char)*pAlpha++;
			//	printf("alpha %d\n",Alpha);
			 	if(Alpha == 0)
			 		*((int16_t*)(psurf+surface.lPitch*(i+top)+2*(j+left)))=RESERVED_COLOR_KEY;
			 	else 
			 	{
					R = (stylesheet.palette[index]>>16) & 0xf8;
					G = (stylesheet.palette[index]>>8) & 0xfc;
					B =  stylesheet.palette[index] & 0xf8;
				 	*((int16_t*)(psurf+surface.lPitch*(i+top)+2*(j+left))) =  (int16_t)((R<<8)|(G<<3)|(B>>3));
				}
			}
			pindex +=(stylesheet.imagewidth-width);
			pAlpha +=(stylesheet.imagewidth-width);
		}
	}
	else if(surface.pixelFormat == Format_32 ) {	// A888
		int32_t A,R,G,B;
		for(int i=0;i<height;i++)
		{
			for(int j = 0;j<width;j++)
			{
			 unsigned char index = (unsigned char)*pindex++;
			 unsigned char Alpha = (unsigned char)*pAlpha++;
			A = Alpha&0xff;
			R = (stylesheet.palette[index]>>16)&0xff;
			G = (stylesheet.palette[index]>>8)&0xff ;
			B =  stylesheet.palette[index] &0xff;
		 	*((int32_t*)(psurf+surface.lPitch*(i+top)+4*(j+left))) =  (int32_t)((A<<24)|(R<<16)|(G<<8)|B);
			}
			pindex +=(stylesheet.imagewidth-width);
			pAlpha +=(stylesheet.imagewidth-width);
		}
	}
	else if(surface.pixelFormat == Format_32_888A) {	// 888A
		int32_t A,R,G,B;
		for(int i=0;i<height;i++)
		{
			for(int j = 0;j<width;j++)
			{
			 unsigned char index = (unsigned char)*pindex++;
			 unsigned char Alpha = (unsigned char)*pAlpha++;
			A = Alpha&0xff;
			R = (stylesheet.palette[index]>>16)&0xff;
			G = (stylesheet.palette[index]>>8)&0xff ;
			B =  stylesheet.palette[index] &0xff;
		 	*((int32_t*)(psurf+surface.lPitch*(i+top)+4*(j+left))) =  (int32_t)((R<<24)|(G<<16)|(B<<8)|A);
			}
			pindex +=(stylesheet.imagewidth-width);
			pAlpha +=(stylesheet.imagewidth-width);
		}
	}
		DG_Unlock(handleDisplay);
		return S_OK;
	}
	else
		return E_FAIL;
}

