/*
* INTEL CONFIDENTIAL
* Copyright (c) 2002, 2003 Intel Corporation.  All rights reserved.
* 
* The source code contained or described herein and all documents
* related to the source code ("Material") are owned by Intel
* Corporation or its suppliers or licensors.  Title to the
* Material remains with Intel Corporation or its suppliers and
* licensors.  The Material contains trade secrets and proprietary
* and confidential information of Intel or its suppliers and
* licensors. The Material is protected by worldwide copyright and
* trade secret laws and treaty provisions.  No part of the Material
* may be used, copied, reproduced, modified, published, uploaded,
* posted, transmitted, distributed, or disclosed in any way without
* Intel's prior express written permission.

* No license under any patent, copyright, trade secret or other
* intellectual property right is granted to or conferred upon you
* by disclosure or delivery of the Materials, either expressly, by
* implication, inducement, estoppel or otherwise. Any license
* under such intellectual property rights must be express and
* approved by Intel in writing.
* 
* $Workfile: RTKMR_MicroStack.c
* $Revision: #1.0.1868.18043
* $Author:   Intel Corporation, Intel Device Builder
* $Date:     2005年7月27日
*
*
*
*/


#if defined(WIN32) || defined(_WIN32_WCE)
#	ifndef MICROSTACK_NO_STDAFX
#		include "stdafx.h"
#	endif
char* RTKMR_PLATFORM = "WINDOWS";
#else
char* RTKMR_PLATFORM = "POSIX";
#endif

#if defined(WIN32)
#define _CRTDBG_MAP_ALLOC
#endif

#if defined(WINSOCK2)
#	include <winsock2.h>
#	include <ws2tcpip.h>
#elif defined(WINSOCK1)
#	include <winsock.h>
#	include <wininet.h>
#endif

#include "ILibParsers.h"
#include "RTKMR_MicroStack.h"
#include "ILibWebServer.h"
#include "ILibWebClient.h"
#include "ILibAsyncSocket.h"

#if defined(WIN32)
#include <crtdbg.h>
#endif

#define UPNP_HTTP_MAXSOCKETS 5
#define UPNP_PORT 1900
#define UPNP_GROUP "239.255.255.250"
#define RTKMR_MIN(a,b) (((a)<(b))?(a):(b))

#define LVL3DEBUG(x)

RTKMR_MicroStackToken RTKMR_CreateMicroStack(void *Chain, const char* FriendlyName,const char* UDN, const char* SerialNumber, const int NotifyCycleSeconds, const unsigned short PortNum);
/* UPnP Set Function Pointers Methods */
void (*RTKMR_FP_PresentationPage) (void* upnptoken,struct packetheader *packet);
/*! \var RTKMR_FP_AVTransport_GetDeviceCapabilities
\brief Dispatch Pointer for AVTransport >> urn:schemas-upnp-org:service:AVTransport:1 >> GetDeviceCapabilities
*/
void (*RTKMR_FP_AVTransport_GetDeviceCapabilities) (void* upnptoken,unsigned int InstanceID);
/*! \var RTKMR_FP_AVTransport_GetMediaInfo
\brief Dispatch Pointer for AVTransport >> urn:schemas-upnp-org:service:AVTransport:1 >> GetMediaInfo
*/
void (*RTKMR_FP_AVTransport_GetMediaInfo) (void* upnptoken,unsigned int InstanceID);
/*! \var RTKMR_FP_AVTransport_GetPositionInfo
\brief Dispatch Pointer for AVTransport >> urn:schemas-upnp-org:service:AVTransport:1 >> GetPositionInfo
*/
void (*RTKMR_FP_AVTransport_GetPositionInfo) (void* upnptoken,unsigned int InstanceID);
/*! \var RTKMR_FP_AVTransport_GetTransportInfo
\brief Dispatch Pointer for AVTransport >> urn:schemas-upnp-org:service:AVTransport:1 >> GetTransportInfo
*/
void (*RTKMR_FP_AVTransport_GetTransportInfo) (void* upnptoken,unsigned int InstanceID);
/*! \var RTKMR_FP_AVTransport_GetTransportSettings
\brief Dispatch Pointer for AVTransport >> urn:schemas-upnp-org:service:AVTransport:1 >> GetTransportSettings
*/
void (*RTKMR_FP_AVTransport_GetTransportSettings) (void* upnptoken,unsigned int InstanceID);
/*! \var RTKMR_FP_AVTransport_Next
\brief Dispatch Pointer for AVTransport >> urn:schemas-upnp-org:service:AVTransport:1 >> Next
*/
void (*RTKMR_FP_AVTransport_Next) (void* upnptoken,unsigned int InstanceID);
/*! \var RTKMR_FP_AVTransport_Pause
\brief Dispatch Pointer for AVTransport >> urn:schemas-upnp-org:service:AVTransport:1 >> Pause
*/
void (*RTKMR_FP_AVTransport_Pause) (void* upnptoken,unsigned int InstanceID);
/*! \var RTKMR_FP_AVTransport_Play
\brief Dispatch Pointer for AVTransport >> urn:schemas-upnp-org:service:AVTransport:1 >> Play
*/
void (*RTKMR_FP_AVTransport_Play) (void* upnptoken,unsigned int InstanceID,char* Speed);
/*! \var RTKMR_FP_AVTransport_Previous
\brief Dispatch Pointer for AVTransport >> urn:schemas-upnp-org:service:AVTransport:1 >> Previous
*/
void (*RTKMR_FP_AVTransport_Previous) (void* upnptoken,unsigned int InstanceID);
/*! \var RTKMR_FP_AVTransport_Seek
\brief Dispatch Pointer for AVTransport >> urn:schemas-upnp-org:service:AVTransport:1 >> Seek
*/
void (*RTKMR_FP_AVTransport_Seek) (void* upnptoken,unsigned int InstanceID,char* Unit,char* Target);
/*! \var RTKMR_FP_AVTransport_SetAVTransportURI
\brief Dispatch Pointer for AVTransport >> urn:schemas-upnp-org:service:AVTransport:1 >> SetAVTransportURI
*/
void (*RTKMR_FP_AVTransport_SetAVTransportURI) (void* upnptoken,unsigned int InstanceID,char* CurrentURI,char* CurrentURIMetaData);
/*! \var RTKMR_FP_AVTransport_SetPlayMode
\brief Dispatch Pointer for AVTransport >> urn:schemas-upnp-org:service:AVTransport:1 >> SetPlayMode
*/
void (*RTKMR_FP_AVTransport_SetPlayMode) (void* upnptoken,unsigned int InstanceID,char* NewPlayMode);
/*! \var RTKMR_FP_AVTransport_Stop
\brief Dispatch Pointer for AVTransport >> urn:schemas-upnp-org:service:AVTransport:1 >> Stop
*/
void (*RTKMR_FP_AVTransport_Stop) (void* upnptoken,unsigned int InstanceID);
/*! \var RTKMR_FP_ConnectionManager_ConnectionComplete
\brief Dispatch Pointer for ConnectionManager >> urn:schemas-upnp-org:service:ConnectionManager:1 >> ConnectionComplete
*/
void (*RTKMR_FP_ConnectionManager_ConnectionComplete) (void* upnptoken,int ConnectionID);
/*! \var RTKMR_FP_ConnectionManager_GetCurrentConnectionIDs
\brief Dispatch Pointer for ConnectionManager >> urn:schemas-upnp-org:service:ConnectionManager:1 >> GetCurrentConnectionIDs
*/
void (*RTKMR_FP_ConnectionManager_GetCurrentConnectionIDs) (void* upnptoken);
/*! \var RTKMR_FP_ConnectionManager_GetCurrentConnectionInfo
\brief Dispatch Pointer for ConnectionManager >> urn:schemas-upnp-org:service:ConnectionManager:1 >> GetCurrentConnectionInfo
*/
void (*RTKMR_FP_ConnectionManager_GetCurrentConnectionInfo) (void* upnptoken,int ConnectionID);
/*! \var RTKMR_FP_ConnectionManager_GetProtocolInfo
\brief Dispatch Pointer for ConnectionManager >> urn:schemas-upnp-org:service:ConnectionManager:1 >> GetProtocolInfo
*/
void (*RTKMR_FP_ConnectionManager_GetProtocolInfo) (void* upnptoken);
/*! \var RTKMR_FP_ConnectionManager_PrepareForConnection
\brief Dispatch Pointer for ConnectionManager >> urn:schemas-upnp-org:service:ConnectionManager:1 >> PrepareForConnection
*/
void (*RTKMR_FP_ConnectionManager_PrepareForConnection) (void* upnptoken,char* RemoteProtocolInfo,char* PeerConnectionManager,int PeerConnectionID,char* Direction);
/*! \var RTKMR_FP_RenderingControl_GetMute
\brief Dispatch Pointer for RenderingControl >> urn:schemas-upnp-org:service:RenderingControl:1 >> GetMute
*/
void (*RTKMR_FP_RenderingControl_GetMute) (void* upnptoken,unsigned int InstanceID,char* Channel);
/*! \var RTKMR_FP_RenderingControl_GetVolume
\brief Dispatch Pointer for RenderingControl >> urn:schemas-upnp-org:service:RenderingControl:1 >> GetVolume
*/
void (*RTKMR_FP_RenderingControl_GetVolume) (void* upnptoken,unsigned int InstanceID,char* Channel);
/*! \var RTKMR_FP_RenderingControl_GetVolumeDB
\brief Dispatch Pointer for RenderingControl >> urn:schemas-upnp-org:service:RenderingControl:1 >> GetVolumeDB
*/
void (*RTKMR_FP_RenderingControl_GetVolumeDB) (void* upnptoken,unsigned int InstanceID,char* Channel);
/*! \var RTKMR_FP_RenderingControl_GetVolumeDBRange
\brief Dispatch Pointer for RenderingControl >> urn:schemas-upnp-org:service:RenderingControl:1 >> GetVolumeDBRange
*/
void (*RTKMR_FP_RenderingControl_GetVolumeDBRange) (void* upnptoken,unsigned int InstanceID,char* Channel);
/*! \var RTKMR_FP_RenderingControl_ListPresets
\brief Dispatch Pointer for RenderingControl >> urn:schemas-upnp-org:service:RenderingControl:1 >> ListPresets
*/
void (*RTKMR_FP_RenderingControl_ListPresets) (void* upnptoken,unsigned int InstanceID);
/*! \var RTKMR_FP_RenderingControl_SelectPreset
\brief Dispatch Pointer for RenderingControl >> urn:schemas-upnp-org:service:RenderingControl:1 >> SelectPreset
*/
void (*RTKMR_FP_RenderingControl_SelectPreset) (void* upnptoken,unsigned int InstanceID,char* PresetName);
/*! \var RTKMR_FP_RenderingControl_SetMute
\brief Dispatch Pointer for RenderingControl >> urn:schemas-upnp-org:service:RenderingControl:1 >> SetMute
*/
void (*RTKMR_FP_RenderingControl_SetMute) (void* upnptoken,unsigned int InstanceID,char* Channel,int DesiredMute);
/*! \var RTKMR_FP_RenderingControl_SetVolume
\brief Dispatch Pointer for RenderingControl >> urn:schemas-upnp-org:service:RenderingControl:1 >> SetVolume
*/
void (*RTKMR_FP_RenderingControl_SetVolume) (void* upnptoken,unsigned int InstanceID,char* Channel,unsigned short DesiredVolume);
/*! \var RTKMR_FP_RenderingControl_SetVolumeDB
\brief Dispatch Pointer for RenderingControl >> urn:schemas-upnp-org:service:RenderingControl:1 >> SetVolumeDB
*/
void (*RTKMR_FP_RenderingControl_SetVolumeDB) (void* upnptoken,unsigned int InstanceID,char* Channel,short DesiredVolume);



const int RTKMR_DeviceDescriptionTemplateLengthUX = 1575;
const int RTKMR_DeviceDescriptionTemplateLength = 699;
const char RTKMR_DeviceDescriptionTemplate[699]={
    0x5A,0x3C,0x3F,0x78,0x6D,0x6C,0x20,0x76,0x65,0x72,0x73,0x69,0x6F,0x6E,0x3D,0x22,0x31,0x2E,0x30,0x22
    ,0x20,0x65,0x6E,0x63,0x6F,0x64,0x69,0x6E,0x67,0x3D,0x22,0x75,0x74,0x66,0x2D,0x38,0x22,0x3F,0x3E,0x3C
    ,0x72,0x6F,0x6F,0x74,0x20,0x78,0x6D,0x6C,0x6E,0x73,0x3D,0x22,0x75,0x72,0x6E,0x3A,0x73,0x63,0x68,0x65
    ,0x6D,0x61,0x73,0x2D,0x75,0x70,0x6E,0x70,0x2D,0x6F,0x72,0x67,0x3A,0x64,0x65,0x76,0x69,0x63,0x65,0x2D
    ,0x31,0x2D,0x30,0x22,0x3E,0x3C,0x73,0x70,0x65,0x63,0x56,0xC6,0x14,0x0B,0x3E,0x3C,0x6D,0x61,0x6A,0x6F
    ,0x72,0x3E,0x31,0x3C,0x2F,0x46,0x02,0x0A,0x3C,0x6D,0x69,0x6E,0x6F,0x72,0x3E,0x30,0x3C,0x2F,0x46,0x02
    ,0x02,0x3C,0x2F,0x8D,0x0B,0x00,0x06,0x12,0x00,0x08,0x02,0x05,0x54,0x79,0x70,0x65,0x3E,0x1B,0x1C,0x12
    ,0x3A,0x4D,0x65,0x64,0x69,0x61,0x52,0x65,0x6E,0x64,0x65,0x72,0x65,0x72,0x3A,0x31,0x3C,0x2F,0x0B,0x0E
    ,0x0F,0x3C,0x64,0x6C,0x6E,0x61,0x3A,0x58,0x5F,0x44,0x4C,0x4E,0x41,0x44,0x4F,0x43,0xC6,0x2F,0x01,0x3A
    ,0x44,0x05,0x00,0x0E,0x31,0x00,0xC4,0x09,0x00,0x11,0x31,0x0A,0x44,0x4D,0x50,0x2D,0x31,0x2E,0x30,0x30
    ,0x3C,0x2F,0x8E,0x11,0x13,0x3E,0x3C,0x66,0x72,0x69,0x65,0x6E,0x64,0x6C,0x79,0x4E,0x61,0x6D,0x65,0x3E
    ,0x25,0x73,0x3C,0x2F,0x4D,0x04,0x2E,0x3C,0x6D,0x61,0x6E,0x75,0x66,0x61,0x63,0x74,0x75,0x72,0x65,0x72
    ,0x3E,0x52,0x65,0x61,0x6C,0x74,0x65,0x6B,0x20,0x50,0x72,0x6F,0x64,0x75,0x63,0x74,0x20,0x44,0x65,0x76
    ,0x65,0x6C,0x6F,0x70,0x6D,0x65,0x6E,0x74,0x20,0x44,0x69,0x76,0x69,0x44,0x5D,0x1B,0x20,0x49,0x20,0x53
    ,0x79,0x73,0x74,0x65,0x6D,0x20,0x44,0x65,0x73,0x69,0x67,0x6E,0x20,0x44,0x65,0x70,0x74,0x2E,0x20,0x49
    ,0x56,0x3C,0x2F,0x0D,0x13,0x00,0x8D,0x16,0x10,0x55,0x52,0x4C,0x3E,0x68,0x74,0x74,0x70,0x3A,0x2F,0x2F
    ,0x77,0x77,0x77,0x2E,0x72,0x06,0x1A,0x09,0x2E,0x63,0x6F,0x6D,0x2E,0x74,0x77,0x3C,0x2F,0xD0,0x0A,0x11
    ,0x3C,0x6D,0x6F,0x64,0x65,0x6C,0x44,0x65,0x73,0x63,0x72,0x69,0x70,0x74,0x69,0x6F,0x6E,0x88,0x26,0x0B
    ,0x27,0x73,0x20,0x55,0x50,0x6E,0x50,0x2F,0x41,0x56,0x20,0xC5,0x52,0x01,0x20,0x08,0x53,0x02,0x20,0x44
    ,0x45,0x74,0x02,0x3C,0x2F,0x91,0x0E,0x00,0x06,0x13,0x00,0x85,0x3F,0x07,0x56,0x65,0x6E,0x75,0x73,0x3C
    ,0x2F,0x4A,0x04,0x00,0x07,0x07,0x0A,0x75,0x6D,0x62,0x65,0x72,0x3E,0x58,0x31,0x3C,0x2F,0x0C,0x04,0x06
    ,0x3C,0x73,0x65,0x72,0x69,0x61,0x88,0x07,0x00,0x84,0x4E,0x00,0x4D,0x04,0x0A,0x3C,0x55,0x44,0x4E,0x3E
    ,0x75,0x75,0x69,0x64,0x3A,0x44,0x55,0x03,0x55,0x44,0x4E,0x45,0x0C,0x00,0x84,0x95,0x04,0x4C,0x69,0x73
    ,0x74,0x49,0x03,0x00,0x89,0x05,0x00,0x1A,0x87,0x00,0xC7,0x0D,0x0E,0x3A,0x41,0x56,0x54,0x72,0x61,0x6E
    ,0x73,0x70,0x6F,0x72,0x74,0x3A,0x31,0x85,0x1B,0x00,0x0A,0x87,0x00,0xC7,0x16,0x02,0x49,0x64,0xC5,0x97
    ,0x00,0xD0,0x0E,0x02,0x49,0x64,0x44,0x0F,0x0B,0x5F,0x62,0x61,0x73,0x65,0x6C,0x69,0x6E,0x65,0x3C,0x2F
    ,0xCA,0x0B,0x05,0x3C,0x53,0x43,0x50,0x44,0xC4,0x5F,0x00,0x8B,0x17,0x0B,0x2F,0x73,0x63,0x70,0x64,0x2E
    ,0x78,0x6D,0x6C,0x3C,0x2F,0x88,0x07,0x08,0x3C,0x63,0x6F,0x6E,0x74,0x72,0x6F,0x6C,0x90,0x0A,0x00,0xC7
    ,0x05,0x02,0x3C,0x2F,0x0B,0x08,0x09,0x3C,0x65,0x76,0x65,0x6E,0x74,0x53,0x75,0x62,0xD0,0x15,0x00,0x05
    ,0x06,0x02,0x3C,0x2F,0xCC,0x07,0x02,0x3C,0x2F,0x10,0x42,0x00,0x6B,0x44,0x06,0x43,0x6F,0x6E,0x6E,0x65
    ,0x63,0x04,0x80,0x07,0x4D,0x61,0x6E,0x61,0x67,0x65,0x72,0xF2,0x45,0x04,0x43,0x4D,0x47,0x52,0x1E,0x46
    ,0x00,0x51,0x19,0x00,0x9F,0x47,0x00,0x12,0x0C,0x00,0x21,0x49,0x00,0xD2,0x18,0x00,0xBF,0x4A,0x00,0xD1
    ,0x8E,0x00,0xC6,0xC1,0x04,0x69,0x6E,0x67,0x43,0xC6,0x71,0x00,0x32,0x90,0x03,0x52,0x43,0x53,0x1E,0x90
    ,0x00,0xD0,0x18,0x00,0x5F,0x91,0x00,0xD1,0x0B,0x00,0xA1,0x92,0x00,0x51,0x18,0x00,0xDE,0x93,0x01,0x2F
    ,0x8D,0xDB,0x03,0x2F,0x64,0x65,0xC7,0x05,0x02,0x72,0x6F,0x00,0x00,0x03,0x6F,0x74,0x3E,0x00,0x00};
/* AVTransport */
const int RTKMR_AVTransportDescriptionLengthUX = 11182;
const int RTKMR_AVTransportDescriptionLength = 2143;
const char RTKMR_AVTransportDescription[2143] = {
    0x4E,0x48,0x54,0x54,0x50,0x2F,0x31,0x2E,0x31,0x20,0x32,0x30,0x30,0x20,0x20,0x4F,0x4B,0x0D,0x0A,0x43
    ,0x4F,0x4E,0x54,0x45,0x4E,0x54,0x2D,0x54,0x59,0x50,0x45,0x3A,0x20,0x20,0x74,0x65,0x78,0x74,0x2F,0x78
    ,0x6D,0x6C,0x3B,0x20,0x63,0x68,0x61,0x72,0x73,0x65,0x74,0x3D,0x22,0x75,0x74,0x66,0x2D,0x38,0x22,0x0D
    ,0x0A,0x53,0x65,0x72,0x76,0x65,0x72,0x3A,0x20,0x50,0x4F,0x53,0x49,0x58,0x2C,0x20,0x55,0x50,0x6E,0xC4
    ,0x12,0x13,0x30,0x2C,0x20,0x49,0x6E,0x74,0x65,0x6C,0x20,0x4D,0x69,0x63,0x72,0x6F,0x53,0x74,0x61,0x63
    ,0x6B,0x84,0x05,0x3C,0x2E,0x31,0x38,0x36,0x38,0x0D,0x0A,0x43,0x6F,0x6E,0x74,0x65,0x6E,0x74,0x2D,0x4C
    ,0x65,0x6E,0x67,0x74,0x68,0x3A,0x20,0x31,0x31,0x30,0x34,0x35,0x0D,0x0A,0x0D,0x0A,0x3C,0x3F,0x78,0x6D
    ,0x6C,0x20,0x76,0x65,0x72,0x73,0x69,0x6F,0x6E,0x3D,0x22,0x31,0x2E,0x30,0x22,0x20,0x65,0x6E,0x63,0x6F
    ,0x64,0x69,0x6E,0x67,0xC8,0x1C,0x37,0x3F,0x3E,0x3C,0x73,0x63,0x70,0x64,0x20,0x78,0x6D,0x6C,0x6E,0x73
    ,0x3D,0x22,0x75,0x72,0x6E,0x3A,0x73,0x63,0x68,0x65,0x6D,0x61,0x73,0x2D,0x75,0x70,0x6E,0x70,0x2D,0x6F
    ,0x72,0x67,0x3A,0x73,0x65,0x72,0x76,0x69,0x63,0x65,0x2D,0x31,0x2D,0x30,0x22,0x3E,0x3C,0x73,0x70,0x65
    ,0x63,0x56,0x06,0x15,0x0B,0x3E,0x3C,0x6D,0x61,0x6A,0x6F,0x72,0x3E,0x31,0x3C,0x2F,0x46,0x02,0x0A,0x3C
    ,0x6D,0x69,0x6E,0x6F,0x72,0x3E,0x30,0x3C,0x2F,0x46,0x02,0x02,0x3C,0x2F,0x8D,0x0B,0x0A,0x61,0x63,0x74
    ,0x69,0x6F,0x6E,0x4C,0x69,0x73,0x74,0x08,0x03,0x0C,0x3E,0x3C,0x6E,0x61,0x6D,0x65,0x3E,0x47,0x65,0x74
    ,0x44,0x65,0x04,0x19,0x0E,0x43,0x61,0x70,0x61,0x62,0x69,0x6C,0x69,0x74,0x69,0x65,0x73,0x3C,0x2F,0x05
    ,0x07,0x09,0x3C,0x61,0x72,0x67,0x75,0x6D,0x65,0x6E,0x74,0x07,0x0E,0x00,0x87,0x03,0x00,0x87,0x0E,0x0A
    ,0x49,0x6E,0x73,0x74,0x61,0x6E,0x63,0x65,0x49,0x44,0xC8,0x0B,0x04,0x64,0x69,0x72,0x65,0x06,0x17,0x04
    ,0x69,0x6E,0x3C,0x2F,0x8A,0x03,0x1C,0x3C,0x72,0x65,0x6C,0x61,0x74,0x65,0x64,0x53,0x74,0x61,0x74,0x65
    ,0x56,0x61,0x72,0x69,0x61,0x62,0x6C,0x65,0x3E,0x41,0x5F,0x41,0x52,0x47,0x5F,0x44,0x66,0x01,0x5F,0xCC
    ,0x12,0x00,0x15,0x0B,0x02,0x3C,0x2F,0x4A,0x1F,0x00,0xCF,0x21,0x09,0x50,0x6C,0x61,0x79,0x4D,0x65,0x64
    ,0x69,0x61,0x92,0x21,0x03,0x6F,0x75,0x74,0xE2,0x21,0x08,0x50,0x6F,0x73,0x73,0x69,0x62,0x6C,0x65,0x04
    ,0x12,0x0B,0x62,0x61,0x63,0x6B,0x53,0x74,0x6F,0x72,0x61,0x67,0x65,0xC7,0x14,0x00,0xB0,0x23,0x03,0x52
    ,0x65,0x63,0x7F,0x23,0x00,0x45,0x23,0x06,0x52,0x65,0x63,0x6F,0x72,0x64,0xFF,0x22,0x0E,0x65,0x63,0x51
    ,0x75,0x61,0x6C,0x69,0x74,0x79,0x4D,0x6F,0x64,0x65,0x73,0xBF,0x24,0x00,0x86,0x24,0x00,0x4E,0x14,0x00
    ,0xE1,0x6A,0x01,0x2F,0x4E,0x90,0x04,0x2F,0x61,0x63,0x74,0x8B,0xA3,0x00,0x8A,0xA0,0x00,0x05,0x70,0x04
    ,0x49,0x6E,0x66,0x6F,0x7F,0x9E,0x00,0x7F,0x9E,0x00,0x6E,0x9E,0x07,0x4E,0x72,0x54,0x72,0x61,0x63,0x6B
    ,0x78,0x56,0x08,0x4E,0x75,0x6D,0x62,0x65,0x72,0x4F,0x66,0x48,0x11,0x00,0x30,0xBE,0x00,0x05,0xBD,0x04
    ,0x44,0x75,0x72,0x61,0x44,0xF6,0x00,0x37,0xBF,0x07,0x43,0x75,0x72,0x72,0x65,0x6E,0x74,0xCF,0x12,0x00
    ,0xB0,0xE0,0x00,0x87,0x11,0x03,0x55,0x52,0x49,0xF7,0xE0,0x0A,0x41,0x56,0x54,0x72,0x61,0x6E,0x73,0x70
    ,0x6F,0x72,0x46,0x11,0x00,0x7A,0x20,0x08,0x4D,0x65,0x74,0x61,0x44,0x61,0x74,0x61,0x7F,0x22,0x02,0x6F
    ,0x72,0x4E,0x13,0x00,0xF1,0x86,0x02,0x65,0x78,0xFB,0x43,0x00,0x84,0x0F,0x00,0xFF,0x44,0x00,0x88,0x20
    ,0x00,0x3F,0x44,0x00,0x92,0x22,0x00,0x3A,0x45,0x04,0x50,0x6C,0x61,0x79,0x04,0xFA,0x02,0x75,0x6D,0x77
    ,0xCC,0x00,0x44,0x10,0x0B,0x62,0x61,0x63,0x6B,0x53,0x74,0x6F,0x72,0x61,0x67,0x65,0x08,0x13,0x00,0xF0
    ,0xED,0x06,0x52,0x65,0x63,0x6F,0x72,0x64,0xBD,0x22,0x00,0xC6,0x10,0x00,0x3F,0x22,0x05,0x57,0x72,0x69
    ,0x74,0x65,0x84,0xFB,0x02,0x75,0x73,0xFD,0x21,0x00,0x06,0x55,0x00,0x8D,0x13,0x00,0x61,0xF0,0x00,0x09
    ,0xF3,0x03,0x4C,0x69,0x73,0xC5,0x03,0x00,0x87,0xE7,0x00,0x08,0x02,0x00,0xC5,0xF5,0x07,0x47,0x65,0x74
    ,0x50,0x6F,0x73,0x69,0xC4,0xF1,0x04,0x49,0x6E,0x66,0x6F,0x08,0xF7,0x00,0xCE,0x0E,0x00,0x8F,0xE2,0x0A
    ,0x49,0x6E,0x73,0x74,0x61,0x6E,0x63,0x65,0x49,0x44,0x92,0xE0,0x02,0x69,0x6E,0x63,0xE0,0x0A,0x5F,0x41
    ,0x52,0x47,0x5F,0x54,0x59,0x50,0x45,0x5F,0xCC,0x12,0x00,0x30,0xE0,0x05,0x54,0x72,0x61,0x63,0x6B,0xB7
    ,0xDF,0x07,0x43,0x75,0x72,0x72,0x65,0x6E,0x74,0xC7,0x10,0x00,0xB5,0x1E,0x04,0x44,0x75,0x72,0x61,0xC4
    ,0xFB,0x00,0xBF,0x20,0x00,0xCE,0x12,0x00,0x35,0x41,0x00,0xCA,0xEB,0x00,0x3F,0x43,0x02,0x63,0x6B,0xBA
    ,0xFE,0x00,0x85,0x63,0x03,0x55,0x52,0x49,0x7F,0x64,0x00,0x89,0x11,0x00,0xB2,0xFC,0x05,0x6C,0x54,0x69
    ,0x6D,0x65,0x79,0xFB,0x06,0x6C,0x61,0x74,0x69,0x76,0x65,0xC4,0x10,0x00,0x08,0xC4,0x00,0xB2,0xFB,0x03
    ,0x41,0x62,0x73,0x3B,0x21,0x07,0x41,0x62,0x73,0x6F,0x6C,0x75,0x74,0x3F,0x21,0x08,0x52,0x65,0x6C,0x43
    ,0x6F,0x75,0x6E,0x74,0x7F,0x42,0x00,0x05,0x11,0x02,0x65,0x72,0x3D,0x43,0x00,0x3C,0x22,0x00,0x48,0x43
    ,0x00,0x32,0x22,0x00,0xC9,0xED,0x03,0x4C,0x69,0x73,0xC5,0x03,0x00,0x87,0xE1,0x00,0x08,0x02,0x00,0x85
    ,0xF0,0x02,0x47,0x65,0x04,0xFF,0x0A,0x6E,0x73,0x70,0x6F,0x72,0x74,0x49,0x6E,0x66,0x6F,0x48,0xF1,0x00
    ,0x0E,0x0F,0x00,0xCF,0xFD,0x0A,0x49,0x6E,0x73,0x74,0x61,0x6E,0x63,0x65,0x49,0x44,0x12,0xFD,0x02,0x69
    ,0x6E,0x63,0x78,0x0A,0x5F,0x41,0x52,0x47,0x5F,0x54,0x59,0x50,0x45,0x5F,0xCC,0x12,0x00,0x30,0xFD,0x00
    ,0x0A,0xEC,0x00,0x06,0x30,0x00,0x85,0xF3,0x00,0x37,0xFF,0x00,0x50,0x11,0x00,0x3F,0x23,0x00,0x05,0x23
    ,0x02,0x75,0x73,0x7F,0x23,0x00,0x89,0x11,0x00,0xB7,0x46,0x05,0x53,0x70,0x65,0x65,0x64,0x7F,0x44,0x05
    ,0x74,0x50,0x6C,0x61,0x79,0x47,0x12,0x00,0xFF,0x9C,0x00,0xD3,0x9C,0x08,0x53,0x65,0x74,0x74,0x69,0x6E
    ,0x67,0x73,0xFF,0x9D,0x00,0xFF,0x9D,0x00,0x6E,0xF4,0x00,0x44,0x44,0x03,0x4D,0x6F,0x64,0xB8,0x9A,0x00
    ,0x87,0xAD,0x00,0x8A,0x11,0x00,0xF0,0xBD,0x09,0x52,0x65,0x63,0x51,0x75,0x61,0x6C,0x69,0x74,0xBF,0x21
    ,0x00,0x04,0xCF,0x06,0x52,0x65,0x63,0x6F,0x72,0x64,0xCD,0x13,0x00,0x7F,0x79,0x00,0x07,0xE7,0x04,0x4E
    ,0x65,0x78,0x74,0x7F,0x75,0x00,0x7F,0x75,0x00,0xFF,0xAA,0x00,0xC6,0x7A,0x04,0x61,0x75,0x73,0x65,0xFF
    ,0x31,0x00,0xFF,0x31,0x00,0xFF,0x31,0x00,0x89,0xAC,0x00,0xBF,0xD8,0x00,0xBF,0xD8,0x00,0xAE,0xD8,0x05
    ,0x53,0x70,0x65,0x65,0x64,0x36,0xF9,0x08,0x54,0x72,0x61,0x6E,0x73,0x70,0x6F,0x72,0x05,0xD8,0x00,0x07
    ,0x12,0x00,0x3F,0x83,0x00,0xC8,0xFD,0x07,0x72,0x65,0x76,0x69,0x6F,0x75,0x73,0xBF,0xB5,0x00,0xBF,0xB5
    ,0x00,0x3F,0xE7,0x00,0xC6,0x57,0x03,0x65,0x65,0x6B,0xFF,0x83,0x00,0xFF,0x83,0x00,0xEE,0x83,0x04,0x55
    ,0x6E,0x69,0x74,0xBF,0xD5,0x02,0x45,0x5F,0x44,0x3D,0x04,0x4D,0x6F,0x64,0x65,0xB2,0xA3,0x05,0x54,0x61
    ,0x72,0x67,0x65,0x7F,0x20,0x00,0x47,0x20,0x00,0xC8,0x12,0x00,0x3F,0x72,0x00,0x09,0x72,0x03,0x74,0x41
    ,0x56,0x49,0xBC,0x03,0x55,0x52,0x49,0x3F,0xF9,0x00,0x3F,0xF9,0x00,0x2E,0xF9,0x06,0x43,0x75,0x72,0x72
    ,0x65,0x6E,0x8C,0x2D,0x00,0xAF,0xC8,0x00,0x8F,0x3E,0x00,0x3A,0x20,0x08,0x4D,0x65,0x74,0x61,0x44,0x61
    ,0x74,0x61,0x3F,0x22,0x01,0x72,0x0E,0x13,0x00,0xFF,0x78,0x00,0xCA,0x78,0x04,0x50,0x6C,0x61,0x79,0x46
    ,0xAE,0x00,0xBF,0xEC,0x00,0xBF,0xEC,0x00,0xAC,0xEC,0x03,0x4E,0x65,0x77,0xD0,0x2D,0x00,0x6E,0xEE,0x00
    ,0x87,0x87,0x00,0x0A,0x3F,0x00,0xBF,0xCC,0x00,0x88,0xCC,0x03,0x74,0x6F,0x70,0x7F,0xC9,0x00,0x7F,0xC9
    ,0x00,0x37,0xFE,0x00,0x87,0x87,0x00,0x46,0xF4,0x07,0x73,0x65,0x72,0x76,0x69,0x63,0x65,0xC5,0xE5,0x01
    ,0x54,0x06,0xDA,0x01,0x73,0xCC,0xE8,0x10,0x20,0x73,0x65,0x6E,0x64,0x45,0x76,0x65,0x6E,0x74,0x73,0x3D
    ,0x22,0x6E,0x6F,0x22,0x8E,0xDC,0x00,0x51,0x66,0x10,0x61,0x74,0x61,0x54,0x79,0x70,0x65,0x3E,0x73,0x74
    ,0x72,0x69,0x6E,0x67,0x3C,0x2F,0x49,0x04,0x0D,0x3C,0x61,0x6C,0x6C,0x6F,0x77,0x65,0x64,0x56,0x61,0x6C
    ,0x75,0x65,0xC7,0x9B,0x00,0x8B,0x04,0x09,0x3E,0x4E,0x4F,0x52,0x4D,0x41,0x4C,0x3C,0x2F,0x4D,0x05,0x00
    ,0xCE,0x08,0x09,0x52,0x45,0x50,0x45,0x41,0x54,0x5F,0x41,0x4C,0xD1,0x09,0x01,0x2F,0x52,0x17,0x07,0x64
    ,0x65,0x66,0x61,0x75,0x6C,0x74,0x4E,0x17,0x00,0x4D,0x05,0x03,0x3C,0x2F,0x73,0x0E,0xF8,0x00,0x24,0x3E
    ,0x13,0x52,0x65,0x63,0x6F,0x72,0x64,0x53,0x74,0x6F,0x72,0x61,0x67,0x65,0x4D,0x65,0x64,0x69,0x75,0x6D
    ,0x3F,0x3F,0x00,0x05,0x3F,0x02,0x4E,0x45,0xE3,0x34,0x00,0x2A,0x2C,0x03,0x79,0x65,0x73,0x48,0x6A,0x09
    ,0x4C,0x61,0x73,0x74,0x43,0x68,0x61,0x6E,0x67,0x24,0x69,0x00,0xB6,0x44,0x0E,0x6C,0x61,0x74,0x69,0x76
    ,0x65,0x54,0x69,0x6D,0x65,0x50,0x6F,0x73,0x69,0xC4,0xE5,0x00,0xFF,0x1A,0x00,0x5F,0x9D,0x08,0x54,0x72
    ,0x61,0x63,0x6B,0x55,0x52,0x49,0xBF,0x19,0x00,0xA4,0x19,0x04,0x44,0x75,0x72,0x61,0x7F,0x34,0x00,0xA3
    ,0xD1,0x00,0x46,0x95,0x06,0x51,0x75,0x61,0x6C,0x69,0x74,0xFF,0xD3,0x00,0xCA,0xD3,0x0D,0x54,0x5F,0x49
    ,0x4D,0x50,0x4C,0x45,0x4D,0x45,0x4E,0x54,0x45,0x44,0xBF,0x97,0x00,0x5F,0x64,0x00,0x04,0xC2,0x01,0x61
    ,0xFF,0x4A,0x00,0x60,0xDE,0x0F,0x41,0x62,0x73,0x6F,0x6C,0x75,0x74,0x65,0x43,0x6F,0x75,0x6E,0x74,0x65
    ,0x72,0x99,0x9A,0x02,0x69,0x34,0x7F,0xB4,0x00,0x49,0xB4,0x00,0xBF,0x1A,0x00,0x24,0x35,0x14,0x5F,0x41
    ,0x52,0x47,0x5F,0x54,0x59,0x50,0x45,0x5F,0x49,0x6E,0x73,0x74,0x61,0x6E,0x63,0x65,0x49,0x44,0xD1,0xE9
    ,0x01,0x75,0x7F,0x1A,0x00,0x44,0x4F,0x0A,0x56,0x54,0x72,0x61,0x6E,0x73,0x70,0x6F,0x72,0x74,0x3F,0xCE
    ,0x00,0xA7,0xE7,0x08,0x4D,0x65,0x74,0x61,0x44,0x61,0x74,0x61,0x7F,0x83,0x00,0x26,0x34,0x00,0x7F,0x1B
    ,0x00,0x60,0xE9,0x00,0xC9,0x4E,0x09,0x50,0x6C,0x61,0x79,0x53,0x70,0x65,0x65,0x64,0xFF,0xE7,0x04,0x75
    ,0x65,0x3E,0x31,0x50,0xE4,0x00,0x4D,0xEF,0x10,0x20,0x76,0x65,0x6E,0x64,0x6F,0x72,0x2D,0x64,0x65,0x66
    ,0x69,0x6E,0x65,0x64,0x20,0xBF,0xEF,0x00,0x98,0xEF,0x04,0x4E,0x65,0x78,0x74,0xBF,0x86,0x00,0x26,0xEF
    ,0x1A,0x50,0x6F,0x73,0x73,0x69,0x62,0x6C,0x65,0x52,0x65,0x63,0x6F,0x72,0x64,0x51,0x75,0x61,0x6C,0x69
    ,0x74,0x79,0x4D,0x6F,0x64,0x65,0x73,0x7F,0x1C,0x00,0x66,0x1C,0x0B,0x53,0x74,0x6F,0x72,0x61,0x67,0x65
    ,0x4D,0x65,0x64,0x69,0x7F,0xA4,0x00,0x9A,0xF2,0x13,0x62,0x73,0x6F,0x6C,0x75,0x74,0x65,0x54,0x69,0x6D
    ,0x65,0x50,0x6F,0x73,0x69,0x74,0x69,0x6F,0x6E,0xFF,0xD9,0x00,0xAA,0x6D,0x00,0x7F,0xDB,0x00,0xA1,0x6F
    ,0x07,0x6C,0x61,0x79,0x62,0x61,0x63,0x6B,0xCB,0x51,0x02,0x75,0x6D,0xFF,0xC0,0x0A,0x75,0x65,0x3E,0x4E
    ,0x45,0x54,0x57,0x4F,0x52,0x4B,0x3F,0xB7,0x00,0x58,0xED,0x00,0xC6,0x9A,0x00,0x06,0x2B,0x0B,0x57,0x72
    ,0x69,0x74,0x65,0x53,0x74,0x61,0x74,0x75,0x73,0xBF,0xEE,0x0A,0x75,0x65,0x3E,0x55,0x4E,0x4B,0x4E,0x4F
    ,0x57,0x4E,0xFF,0xE4,0x00,0xA0,0xCA,0x00,0x13,0x5D,0x00,0xFF,0xAE,0x00,0x59,0xE7,0x00,0x89,0x92,0x00
    ,0x84,0x48,0x01,0x65,0x3F,0x76,0x13,0x75,0x65,0x3E,0x50,0x41,0x55,0x53,0x45,0x44,0x5F,0x52,0x45,0x43
    ,0x4F,0x52,0x44,0x49,0x4E,0x47,0x50,0x78,0x00,0x4E,0x0B,0x06,0x4C,0x41,0x59,0x49,0x4E,0x47,0x7F,0x81
    ,0x00,0xD9,0xCA,0x0C,0x75,0x6D,0x62,0x65,0x72,0x4F,0x66,0x54,0x72,0x61,0x63,0x6B,0x12,0x7F,0x03,0x75
    ,0x69,0x34,0x3F,0xFE,0x13,0x3E,0x41,0x5F,0x41,0x52,0x47,0x5F,0x54,0x59,0x50,0x45,0x5F,0x53,0x65,0x65
    ,0x6B,0x4D,0x6F,0x64,0xBF,0x50,0x00,0x84,0xC6,0x08,0x54,0x52,0x41,0x43,0x4B,0x5F,0x4E,0x52,0xFF,0xC6
    ,0x00,0x18,0xF4,0x07,0x43,0x75,0x72,0x72,0x65,0x6E,0x74,0x45,0x45,0x00,0x3F,0x45,0x00,0x62,0x94,0x00
    ,0xFF,0xDC,0x00,0xC5,0xDC,0x02,0x4F,0x4B,0x1D,0x91,0x0E,0x45,0x52,0x52,0x4F,0x52,0x5F,0x4F,0x43,0x43
    ,0x55,0x52,0x52,0x45,0x44,0xDD,0x9B,0x10,0x20,0x76,0x65,0x6E,0x64,0x6F,0x72,0x2D,0x64,0x65,0x66,0x69
    ,0x6E,0x65,0x64,0x20,0xBF,0xF1,0x00,0xA7,0x85,0x06,0x54,0x61,0x72,0x67,0x65,0x74,0xF3,0xEF,0x08,0x2F
    ,0x73,0x65,0x72,0x76,0x69,0x63,0x65,0x85,0xE6,0x01,0x54,0x08,0x05,0x01,0x63,0x00,0x00,0x03,0x70,0x64
    ,0x3E,0x00,0x00};
/* ConnectionManager */
const int RTKMR_ConnectionManagerDescriptionLengthUX = 4808;
const int RTKMR_ConnectionManagerDescriptionLength = 1104;
const char RTKMR_ConnectionManagerDescription[1104] = {
    0x4E,0x48,0x54,0x54,0x50,0x2F,0x31,0x2E,0x31,0x20,0x32,0x30,0x30,0x20,0x20,0x4F,0x4B,0x0D,0x0A,0x43
    ,0x4F,0x4E,0x54,0x45,0x4E,0x54,0x2D,0x54,0x59,0x50,0x45,0x3A,0x20,0x20,0x74,0x65,0x78,0x74,0x2F,0x78
    ,0x6D,0x6C,0x3B,0x20,0x63,0x68,0x61,0x72,0x73,0x65,0x74,0x3D,0x22,0x75,0x74,0x66,0x2D,0x38,0x22,0x0D
    ,0x0A,0x53,0x65,0x72,0x76,0x65,0x72,0x3A,0x20,0x50,0x4F,0x53,0x49,0x58,0x2C,0x20,0x55,0x50,0x6E,0xC4
    ,0x12,0x13,0x30,0x2C,0x20,0x49,0x6E,0x74,0x65,0x6C,0x20,0x4D,0x69,0x63,0x72,0x6F,0x53,0x74,0x61,0x63
    ,0x6B,0x84,0x05,0x3B,0x2E,0x31,0x38,0x36,0x38,0x0D,0x0A,0x43,0x6F,0x6E,0x74,0x65,0x6E,0x74,0x2D,0x4C
    ,0x65,0x6E,0x67,0x74,0x68,0x3A,0x20,0x34,0x36,0x37,0x32,0x0D,0x0A,0x0D,0x0A,0x3C,0x3F,0x78,0x6D,0x6C
    ,0x20,0x76,0x65,0x72,0x73,0x69,0x6F,0x6E,0x3D,0x22,0x31,0x2E,0x30,0x22,0x20,0x65,0x6E,0x63,0x6F,0x64
    ,0x69,0x6E,0x67,0x88,0x1C,0x37,0x3F,0x3E,0x3C,0x73,0x63,0x70,0x64,0x20,0x78,0x6D,0x6C,0x6E,0x73,0x3D
    ,0x22,0x75,0x72,0x6E,0x3A,0x73,0x63,0x68,0x65,0x6D,0x61,0x73,0x2D,0x75,0x70,0x6E,0x70,0x2D,0x6F,0x72
    ,0x67,0x3A,0x73,0x65,0x72,0x76,0x69,0x63,0x65,0x2D,0x31,0x2D,0x30,0x22,0x3E,0x3C,0x73,0x70,0x65,0x63
    ,0x56,0x06,0x15,0x0B,0x3E,0x3C,0x6D,0x61,0x6A,0x6F,0x72,0x3E,0x31,0x3C,0x2F,0x46,0x02,0x0A,0x3C,0x6D
    ,0x69,0x6E,0x6F,0x72,0x3E,0x30,0x3C,0x2F,0x46,0x02,0x02,0x3C,0x2F,0x8D,0x0B,0x0A,0x61,0x63,0x74,0x69
    ,0x6F,0x6E,0x4C,0x69,0x73,0x74,0x08,0x03,0x0C,0x3E,0x3C,0x6E,0x61,0x6D,0x65,0x3E,0x43,0x6F,0x6E,0x6E
    ,0x65,0x45,0x07,0x0A,0x43,0x6F,0x6D,0x70,0x6C,0x65,0x74,0x65,0x3C,0x2F,0x45,0x06,0x09,0x3C,0x61,0x72
    ,0x67,0x75,0x6D,0x65,0x6E,0x74,0x47,0x0D,0x00,0x87,0x03,0x00,0xD1,0x0D,0x02,0x49,0x44,0x48,0x0C,0x03
    ,0x64,0x69,0x72,0x86,0x12,0x05,0x3E,0x69,0x6E,0x3C,0x2F,0x8A,0x03,0x1C,0x3C,0x72,0x65,0x6C,0x61,0x74
    ,0x65,0x64,0x53,0x74,0x61,0x74,0x65,0x56,0x61,0x72,0x69,0x61,0x62,0x6C,0x65,0x3E,0x41,0x5F,0x41,0x52
    ,0x47,0x5F,0xC4,0x65,0x01,0x5F,0x4E,0x13,0x00,0x95,0x0B,0x02,0x3C,0x2F,0x4A,0x20,0x01,0x2F,0x8E,0x26
    ,0x04,0x2F,0x61,0x63,0x74,0x0B,0x39,0x00,0x07,0x36,0x0A,0x47,0x65,0x74,0x43,0x75,0x72,0x72,0x65,0x6E
    ,0x74,0xCC,0x2A,0x01,0x73,0x71,0x37,0x01,0x73,0x92,0x37,0x03,0x6F,0x75,0x74,0xE2,0x37,0x00,0x56,0x1F
    ,0x00,0x3F,0x37,0x00,0x1C,0x37,0x03,0x6E,0x66,0x6F,0xBF,0x6E,0x00,0xBF,0x6E,0x00,0xA3,0x6E,0x00,0x4F
    ,0x91,0x05,0x52,0x63,0x73,0x49,0x44,0x37,0x58,0x00,0xCB,0x8F,0x03,0x52,0x63,0x73,0xB4,0x1F,0x0B,0x41
    ,0x56,0x54,0x72,0x61,0x6E,0x73,0x70,0x6F,0x72,0x74,0xBF,0x21,0x00,0x45,0xB1,0x00,0xCF,0x13,0x00,0x30
    ,0x43,0x08,0x50,0x72,0x6F,0x74,0x6F,0x63,0x6F,0x6C,0x0C,0x72,0x00,0xFA,0x44,0x00,0x8E,0x13,0x00,0x31
    ,0x23,0x03,0x65,0x65,0x72,0x4A,0xF8,0x07,0x4D,0x61,0x6E,0x61,0x67,0x65,0x72,0x3F,0x6A,0x03,0x50,0x45
    ,0x5F,0xD3,0x14,0x00,0xBE,0x26,0x00,0x7F,0x8F,0x00,0xBF,0xB0,0x00,0x84,0xE2,0x01,0x44,0x48,0xCE,0x00
    ,0xBF,0xB1,0x03,0x50,0x45,0x5F,0xCB,0x12,0x00,0x30,0xD2,0x00,0xC4,0xE7,0x02,0x75,0x73,0x7F,0x68,0x00
    ,0x8D,0xF3,0x00,0x88,0x14,0x00,0xA1,0xF4,0x00,0x49,0xF7,0x03,0x4C,0x69,0x73,0xC5,0x03,0x00,0x07,0xED
    ,0x00,0x08,0x02,0x00,0x05,0xFA,0x03,0x47,0x65,0x74,0xD4,0xB7,0x00,0xCE,0x0E,0x00,0x10,0x35,0x05,0x6F
    ,0x75,0x72,0x63,0x65,0xF7,0xE5,0x00,0x46,0x0F,0x00,0x7E,0xC1,0x04,0x53,0x69,0x6E,0x6B,0xF8,0x1F,0x03
    ,0x69,0x6E,0x6B,0xAF,0xE0,0x00,0x25,0x52,0x09,0x50,0x72,0x65,0x70,0x61,0x72,0x65,0x46,0x6F,0x8B,0xE7
    ,0x00,0x65,0x53,0x06,0x52,0x65,0x6D,0x6F,0x74,0x65,0x14,0x61,0x00,0x8A,0xF3,0x02,0x69,0x6E,0x6D,0xF3
    ,0x00,0x7E,0x57,0x00,0x0E,0xF2,0x07,0x4D,0x61,0x6E,0x61,0x67,0x65,0x72,0x3F,0x25,0x02,0x45,0x5F,0x93
    ,0x14,0x00,0x7E,0x26,0x02,0x49,0x44,0x3F,0x25,0x02,0x45,0x5F,0x4E,0x13,0x00,0xB0,0xF6,0x03,0x44,0x69
    ,0x72,0xCE,0x79,0x00,0x39,0x6C,0x00,0x8B,0x12,0x00,0xB0,0xC2,0x00,0x1E,0x44,0x03,0x6F,0x75,0x74,0x7F
    ,0x44,0x00,0xAC,0xE5,0x0B,0x41,0x56,0x54,0x72,0x61,0x6E,0x73,0x70,0x6F,0x72,0x74,0x7F,0x23,0x00,0x85
    ,0xB1,0x00,0xCF,0x13,0x00,0xF0,0xB1,0x03,0x52,0x63,0x73,0xFF,0x44,0x00,0x05,0xD3,0x03,0x52,0x63,0x73
    ,0x65,0x87,0x01,0x2F,0x4E,0xF9,0x02,0x2F,0x61,0x07,0xEB,0x00,0x47,0x02,0x00,0xC6,0xFE,0x07,0x73,0x65
    ,0x72,0x76,0x69,0x63,0x65,0x45,0xEE,0x01,0x54,0x06,0xE2,0x01,0x73,0x4C,0xF1,0x10,0x20,0x73,0x65,0x6E
    ,0x64,0x45,0x76,0x65,0x6E,0x74,0x73,0x3D,0x22,0x6E,0x6F,0x22,0x48,0x56,0x00,0xD8,0xF6,0x00,0x07,0xE5
    ,0x10,0x61,0x74,0x61,0x54,0x79,0x70,0x65,0x3E,0x73,0x74,0x72,0x69,0x6E,0x67,0x3C,0x2F,0x49,0x04,0x03
    ,0x3C,0x2F,0x73,0x8E,0xFD,0x00,0xAF,0x1B,0x00,0x4A,0xED,0x00,0x04,0xF6,0x02,0x75,0x73,0xA3,0x1C,0x0C
    ,0x61,0x6C,0x6C,0x6F,0x77,0x65,0x64,0x56,0x61,0x6C,0x75,0x65,0xC6,0x42,0x00,0x8C,0x04,0x05,0x3E,0x4F
    ,0x4B,0x3C,0x2F,0x4D,0x04,0x00,0xCE,0x07,0x15,0x43,0x6F,0x6E,0x74,0x65,0x6E,0x74,0x46,0x6F,0x72,0x6D
    ,0x61,0x74,0x4D,0x69,0x73,0x6D,0x61,0x74,0x63,0x68,0x9D,0x0C,0x14,0x49,0x6E,0x73,0x75,0x66,0x66,0x69
    ,0x63,0x69,0x65,0x6E,0x74,0x42,0x61,0x6E,0x64,0x77,0x69,0x64,0x74,0x9E,0x0C,0x05,0x55,0x6E,0x72,0x65
    ,0x6C,0x85,0xF7,0x07,0x43,0x68,0x61,0x6E,0x6E,0x65,0x6C,0x9F,0x0B,0x05,0x6B,0x6E,0x6F,0x77,0x6E,0x90
    ,0x2D,0x01,0x2F,0x12,0x3A,0x00,0x3F,0x5B,0x00,0x96,0xCF,0x00,0xC8,0x76,0x02,0x69,0x34,0xFF,0x75,0x00
    ,0x13,0xB5,0x00,0xFF,0x18,0x00,0xA7,0x8E,0x00,0xBF,0x1A,0x00,0x2B,0xA9,0x07,0x4D,0x61,0x6E,0x61,0x67
    ,0x65,0x72,0xFF,0xC5,0x00,0x4E,0xE1,0x03,0x79,0x65,0x73,0x88,0xE1,0x06,0x53,0x6F,0x75,0x72,0x63,0x65
    ,0x7F,0xE0,0x00,0xA6,0x1A,0x03,0x69,0x6E,0x6B,0x7F,0xFA,0x00,0x6F,0xFA,0x03,0x44,0x69,0x72,0x06,0xFA
    ,0x00,0xBF,0xF8,0x00,0x45,0xE4,0x03,0x70,0x75,0x74,0x5D,0xF9,0x06,0x4F,0x75,0x74,0x70,0x75,0x74,0xBF
    ,0xD4,0x00,0xD9,0x69,0x07,0x43,0x75,0x72,0x72,0x65,0x6E,0x74,0x4C,0xA0,0x01,0x73,0xF3,0x84,0x09,0x2F
    ,0x73,0x65,0x72,0x76,0x69,0x63,0x65,0x53,0x44,0xF5,0x01,0x54,0x08,0x05,0x01,0x63,0x00,0x00,0x03,0x70
    ,0x64,0x3E,0x00,0x00};
/* RenderingControl */
const int RTKMR_RenderingControlDescriptionLengthUX = 5290;
const int RTKMR_RenderingControlDescriptionLength = 974;
const char RTKMR_RenderingControlDescription[974] = {
    0x4E,0x48,0x54,0x54,0x50,0x2F,0x31,0x2E,0x31,0x20,0x32,0x30,0x30,0x20,0x20,0x4F,0x4B,0x0D,0x0A,0x43
    ,0x4F,0x4E,0x54,0x45,0x4E,0x54,0x2D,0x54,0x59,0x50,0x45,0x3A,0x20,0x20,0x74,0x65,0x78,0x74,0x2F,0x78
    ,0x6D,0x6C,0x3B,0x20,0x63,0x68,0x61,0x72,0x73,0x65,0x74,0x3D,0x22,0x75,0x74,0x66,0x2D,0x38,0x22,0x0D
    ,0x0A,0x53,0x65,0x72,0x76,0x65,0x72,0x3A,0x20,0x50,0x4F,0x53,0x49,0x58,0x2C,0x20,0x55,0x50,0x6E,0xC4
    ,0x12,0x13,0x30,0x2C,0x20,0x49,0x6E,0x74,0x65,0x6C,0x20,0x4D,0x69,0x63,0x72,0x6F,0x53,0x74,0x61,0x63
    ,0x6B,0x84,0x05,0x3B,0x2E,0x31,0x38,0x36,0x38,0x0D,0x0A,0x43,0x6F,0x6E,0x74,0x65,0x6E,0x74,0x2D,0x4C
    ,0x65,0x6E,0x67,0x74,0x68,0x3A,0x20,0x35,0x31,0x35,0x34,0x0D,0x0A,0x0D,0x0A,0x3C,0x3F,0x78,0x6D,0x6C
    ,0x20,0x76,0x65,0x72,0x73,0x69,0x6F,0x6E,0x3D,0x22,0x31,0x2E,0x30,0x22,0x20,0x65,0x6E,0x63,0x6F,0x64
    ,0x69,0x6E,0x67,0x88,0x1C,0x37,0x3F,0x3E,0x3C,0x73,0x63,0x70,0x64,0x20,0x78,0x6D,0x6C,0x6E,0x73,0x3D
    ,0x22,0x75,0x72,0x6E,0x3A,0x73,0x63,0x68,0x65,0x6D,0x61,0x73,0x2D,0x75,0x70,0x6E,0x70,0x2D,0x6F,0x72
    ,0x67,0x3A,0x73,0x65,0x72,0x76,0x69,0x63,0x65,0x2D,0x31,0x2D,0x30,0x22,0x3E,0x3C,0x73,0x70,0x65,0x63
    ,0x56,0x06,0x15,0x0B,0x3E,0x3C,0x6D,0x61,0x6A,0x6F,0x72,0x3E,0x31,0x3C,0x2F,0x46,0x02,0x0A,0x3C,0x6D
    ,0x69,0x6E,0x6F,0x72,0x3E,0x30,0x3C,0x2F,0x46,0x02,0x02,0x3C,0x2F,0x8D,0x0B,0x0A,0x61,0x63,0x74,0x69
    ,0x6F,0x6E,0x4C,0x69,0x73,0x74,0x08,0x03,0x10,0x3E,0x3C,0x6E,0x61,0x6D,0x65,0x3E,0x47,0x65,0x74,0x4D
    ,0x75,0x74,0x65,0x3C,0x2F,0x85,0x03,0x09,0x3C,0x61,0x72,0x67,0x75,0x6D,0x65,0x6E,0x74,0x87,0x0A,0x00
    ,0x87,0x03,0x00,0x07,0x0B,0x0A,0x49,0x6E,0x73,0x74,0x61,0x6E,0x63,0x65,0x49,0x44,0xC8,0x0B,0x04,0x64
    ,0x69,0x72,0x65,0x86,0x13,0x04,0x69,0x6E,0x3C,0x2F,0x8A,0x03,0x1C,0x3C,0x72,0x65,0x6C,0x61,0x74,0x65
    ,0x64,0x53,0x74,0x61,0x74,0x65,0x56,0x61,0x72,0x69,0x61,0x62,0x6C,0x65,0x3E,0x41,0x5F,0x41,0x52,0x47
    ,0x5F,0x84,0x62,0x01,0x5F,0xCC,0x12,0x00,0x15,0x0B,0x02,0x3C,0x2F,0x4A,0x1F,0x00,0xCF,0x21,0x07,0x43
    ,0x68,0x61,0x6E,0x6E,0x65,0x6C,0x3F,0x21,0x02,0x45,0x5F,0x09,0x12,0x00,0x71,0x20,0x05,0x75,0x72,0x72
    ,0x65,0x6E,0x0D,0x4E,0x00,0x4A,0x42,0x03,0x6F,0x75,0x74,0xA2,0x42,0x00,0xC6,0x5C,0x00,0x61,0x3E,0x01
    ,0x2F,0xCE,0x63,0x04,0x2F,0x61,0x63,0x74,0x8B,0x73,0x00,0x8A,0x70,0x05,0x56,0x6F,0x6C,0x75,0x6D,0x3F
    ,0x71,0x00,0x3F,0x71,0x00,0x3F,0x71,0x00,0x3F,0x71,0x00,0x39,0x71,0x00,0x8E,0x4E,0x00,0xAF,0x71,0x00
    ,0xC8,0x5D,0x00,0x3F,0x72,0x00,0x10,0x72,0x02,0x44,0x42,0xBF,0x72,0x00,0xBF,0x72,0x00,0xBF,0x72,0x00
    ,0xBF,0x72,0x00,0xBF,0x72,0x00,0xBC,0x72,0x02,0x44,0x42,0x3F,0x73,0x00,0x14,0x73,0x04,0x52,0x61,0x6E
    ,0x67,0xFF,0xE6,0x00,0xFF,0xE6,0x00,0xFF,0xE6,0x00,0xFF,0xE6,0x00,0xF2,0xE6,0x07,0x4D,0x69,0x6E,0x56
    ,0x61,0x6C,0x75,0x3F,0x73,0x01,0x42,0x73,0x1E,0x02,0x61,0x78,0x7F,0x1E,0x00,0x7F,0x91,0x00,0x4E,0x91
    ,0x00,0x84,0xFD,0x07,0x50,0x72,0x65,0x73,0x65,0x74,0x73,0x3F,0x90,0x00,0x3F,0x90,0x00,0x35,0xE4,0x00
    ,0x86,0x2E,0x04,0x4E,0x61,0x6D,0x65,0x44,0xCA,0x00,0x37,0xE6,0x00,0x50,0x11,0x00,0xBF,0xE7,0x00,0x87
    ,0xE7,0x05,0x53,0x65,0x6C,0x65,0x63,0xC7,0x56,0x00,0xBF,0xE6,0x00,0xBF,0xE6,0x00,0xAE,0xE6,0x00,0xCA
    ,0x54,0x00,0x7F,0xE7,0x02,0x45,0x5F,0xCC,0x12,0x00,0x7F,0x55,0x00,0x49,0x55,0x05,0x74,0x4D,0x75,0x74
    ,0x65,0xBF,0xAA,0x00,0xBF,0xAA,0x00,0xAF,0xAA,0x06,0x68,0x61,0x6E,0x6E,0x65,0x6C,0xBF,0xCB,0x02,0x45
    ,0x5F,0x09,0x12,0x00,0xF0,0xCA,0x0A,0x44,0x65,0x73,0x69,0x72,0x65,0x64,0x4D,0x75,0x74,0xB7,0x74,0x03
    ,0x4D,0x75,0x74,0x7F,0x70,0x00,0x4D,0x70,0x05,0x56,0x6F,0x6C,0x75,0x6D,0xFF,0x70,0x00,0xFF,0x70,0x00
    ,0xFF,0x70,0x00,0xFF,0x70,0x00,0xF9,0x70,0x00,0x8E,0x4E,0x00,0xEE,0xE5,0x00,0x88,0x5D,0x00,0xFF,0x71
    ,0x00,0xD0,0x71,0x02,0x44,0x42,0x7F,0x72,0x00,0x7F,0x72,0x00,0x7F,0x72,0x00,0x7F,0x72,0x00,0x7F,0x72
    ,0x00,0x7B,0x72,0x02,0x44,0x42,0xBB,0xE4,0x00,0xC7,0xE6,0x00,0x06,0xEA,0x07,0x73,0x65,0x72,0x76,0x69
    ,0x63,0x65,0xC5,0xFC,0x01,0x54,0x46,0xF5,0x01,0x73,0xCC,0xFF,0x10,0x20,0x73,0x65,0x6E,0x64,0x45,0x76
    ,0x65,0x6E,0x74,0x73,0x3D,0x22,0x6E,0x6F,0x22,0x47,0xF2,0x00,0xD0,0x7F,0x0D,0x64,0x61,0x74,0x61,0x54
    ,0x79,0x70,0x65,0x3E,0x69,0x32,0x3C,0x2F,0x49,0x03,0x03,0x3C,0x2F,0x73,0x4E,0xD9,0x00,0xE4,0x16,0x04
    ,0x4D,0x75,0x74,0x65,0xD1,0x15,0x07,0x62,0x6F,0x6F,0x6C,0x65,0x61,0x6E,0x36,0x17,0x03,0x79,0x65,0x73
    ,0x08,0x2E,0x04,0x4C,0x61,0x73,0x74,0x04,0xF4,0x01,0x67,0xD2,0x18,0x06,0x73,0x74,0x72,0x69,0x6E,0x67
    ,0xBF,0x2F,0x00,0xD8,0xAA,0x00,0x8F,0x49,0x03,0x75,0x69,0x34,0x7F,0x1A,0x00,0x0C,0xC5,0x09,0x50,0x72
    ,0x65,0x73,0x65,0x74,0x4E,0x61,0x6D,0x64,0x35,0x0C,0x61,0x6C,0x6C,0x6F,0x77,0x65,0x64,0x56,0x61,0x6C
    ,0x75,0x65,0x47,0xEB,0x00,0x8B,0x04,0x12,0x3E,0x46,0x61,0x63,0x74,0x6F,0x72,0x79,0x44,0x65,0x66,0x61
    ,0x75,0x6C,0x74,0x73,0x3C,0x2F,0x8D,0x07,0x02,0x3C,0x2F,0xD2,0x0F,0x00,0x34,0x79,0x00,0x8A,0x2C,0x00
    ,0x04,0xA7,0x00,0xFF,0x62,0x00,0x1E,0xA9,0x00,0x13,0x5F,0x00,0xFF,0xA8,0x00,0x4D,0x79,0x00,0x44,0x93
    ,0x03,0x6E,0x65,0x6C,0x7F,0x5E,0x09,0x75,0x65,0x3E,0x4D,0x61,0x73,0x74,0x65,0x72,0x33,0x5C,0x01,0x2F
    ,0xD3,0xF0,0x03,0x2F,0x73,0x63,0x00,0x00,0x03,0x70,0x64,0x3E,0x00,0x00};



struct RTKMR_DataObject;

//
// It should not be necessary to expose/modify any of these structures. They
// are used by the internal stack
//

struct SubscriberInfo
{
    char* SID;		// Subscription ID
    int SIDLength;
    int SEQ;
    
    
    int Address;
    unsigned short Port;
    char* Path;
    int PathLength;
    int RefCount;
    int Disposing;
    
    #if defined(WIN32) || defined(_WIN32_WCE)
    unsigned int RenewByTime;
    #else
    struct timeval RenewByTime;
    #endif
    
    struct SubscriberInfo *Next;
    struct SubscriberInfo *Previous;
};
struct RTKMR_DataObject
{
    //
    // Absolutely DO NOT put anything above these 3 function pointers
    //
    ILibChain_PreSelect PreSelect;
    ILibChain_PostSelect PostSelect;
    ILibChain_Destroy Destroy;
    
    void *EventClient;
    void *Chain;
    int UpdateFlag;
    
    /* Network Poll */
    unsigned int NetworkPollTime;
    
    int ForceExit;
    char *UUID;
    char *UDN;
    char *Serial;
    
    void *WebServerTimer;
    void *HTTPServer;
    
    char *DeviceDescription;
    int DeviceDescriptionLength;
    int InitialNotify;
    
    char* AVTransport_LastChange;
    char* ConnectionManager_SourceProtocolInfo;
    char* ConnectionManager_SinkProtocolInfo;
    char* ConnectionManager_CurrentConnectionIDs;
    char* RenderingControl_LastChange;
    
    
    struct sockaddr_in addr;
    int addrlen;
    
    struct ip_mreq mreq;
    char message[4096];
    int *AddressList;
    int AddressListLength;
    
    int _NumEmbeddedDevices;
    int WebSocketPortNumber;
    
    
    
    #if defined(WIN32) || defined(_WIN32_WCE)
    SOCKET *NOTIFY_SEND_socks;
    SOCKET NOTIFY_RECEIVE_sock;
    SOCKET MSEARCH_sock;	
    unsigned int CurrentTime;
    unsigned int NotifyTime;
    #else
    int *NOTIFY_SEND_socks;
    int NOTIFY_RECEIVE_sock;
    int MSEARCH_sock;
    struct timeval CurrentTime;
    struct timeval NotifyTime;
    #endif
    
    int SID;
    int NotifyCycleTime;
    
    
    sem_t EventLock;
    struct SubscriberInfo *HeadSubscriberPtr_AVTransport;
    int NumberOfSubscribers_AVTransport;
    struct SubscriberInfo *HeadSubscriberPtr_ConnectionManager;
    int NumberOfSubscribers_ConnectionManager;
    struct SubscriberInfo *HeadSubscriberPtr_RenderingControl;
    int NumberOfSubscribers_RenderingControl;
    
};

struct MSEARCH_state
{
    char *ST;
    int STLength;
    void *upnp;
    struct sockaddr_in dest_addr;
};
struct RTKMR_FragmentNotifyStruct
{
    struct RTKMR_DataObject *upnp;
    int packetNumber;
};

/* Pre-declarations */
void RTKMR_FragmentedSendNotify(void *data);
void RTKMR_SendNotify(const struct RTKMR_DataObject *upnp);
void RTKMR_SendByeBye(const struct RTKMR_DataObject *upnp);
void RTKMR_MainInvokeSwitch();
void RTKMR_SendDataXmlEscaped(const void* RTKMR_Token, const char* Data, const int DataLength, const int Terminate);
void RTKMR_SendData(const void* RTKMR_Token, const char* Data, const int DataLength, const int Terminate);
int RTKMR_PeriodicNotify(struct RTKMR_DataObject *upnp);
void RTKMR_SendEvent_Body(void *upnptoken, char *body, int bodylength, struct SubscriberInfo *info);

/*! \fn RTKMR_GetWebServerToken(const RTKMR_MicroStackToken MicroStackToken)
\brief Converts a MicroStackToken to a WebServerToken
\para
\a MicroStackToken is the void* returned from a call to RTKMR_CreateMicroStack. The returned token, is the server token
not the session token.
\param MicroStackToken MicroStack Token
\returns WebServer Token
*/
void* RTKMR_GetWebServerToken(const RTKMR_MicroStackToken MicroStackToken)
{
    return(((struct RTKMR_DataObject*)MicroStackToken)->HTTPServer);
}


#define RTKMR_BuildSsdpResponsePacket(outpacket,outlength,ipaddr,port,EmbeddedDeviceNumber,USN,USNex,ST,NTex,NotifyTime)\
{\
    *outlength = sprintf(outpacket,"HTTP/1.1 200 OK\r\nLOCATION: http://%d.%d.%d.%d:%d/\r\nEXT:\r\nSERVER: %s, UPnP/1.0, Intel MicroStack/1.0.1868\r\nUSN: uuid:%s%s\r\nCACHE-CONTROL: max-age=%d\r\nST: %s%s\r\n\r\n" ,(ipaddr&0xFF),((ipaddr>>8)&0xFF),((ipaddr>>16)&0xFF),((ipaddr>>24)&0xFF),port,RTKMR_PLATFORM,USN,USNex,NotifyTime,ST,NTex);\
}
#define RTKMR_BuildSsdpNotifyPacket(outpacket,outlength,ipaddr,port,EmbeddedDeviceNumber,USN,USNex,NT,NTex,NotifyTime)\
{\
    *outlength = sprintf(outpacket,"NOTIFY * HTTP/1.1\r\nLOCATION: http://%d.%d.%d.%d:%d/\r\nHOST: 239.255.255.250:1900\r\nSERVER: %s, UPnP/1.0, Intel MicroStack/1.0.1868\r\nNTS: ssdp:alive\r\nUSN: uuid:%s%s\r\nCACHE-CONTROL: max-age=%d\r\nNT: %s%s\r\n\r\n",(ipaddr&0xFF),((ipaddr>>8)&0xFF),((ipaddr>>16)&0xFF),((ipaddr>>24)&0xFF),port,RTKMR_PLATFORM,USN,USNex,NotifyTime,NT,NTex);\
}




void RTKMR_SetDisconnectFlag(RTKMR_SessionToken token,void *flag)
{
    ((struct ILibWebServer_Session*)token)->Reserved10=flag;
}


/*! \fn RTKMR_IPAddressListChanged(RTKMR_MicroStackToken MicroStackToken)
\brief Tell the underlying MicroStack that an IPAddress may have changed
\param MicroStackToken Microstack
*/
void RTKMR_IPAddressListChanged(RTKMR_MicroStackToken MicroStackToken)
{
    ((struct RTKMR_DataObject*)MicroStackToken)->UpdateFlag = 1;
    ILibForceUnBlockChain(((struct RTKMR_DataObject*)MicroStackToken)->Chain);
}

//
//	Internal underlying Initialization, that shouldn't be called explicitely
// 
// <param name="state">State object</param>
// <param name="NotifyCycleSeconds">Cycle duration</param>
// <param name="PortNumber">Port Number</param>
void RTKMR_Init(struct RTKMR_DataObject *state,const int NotifyCycleSeconds,const unsigned short PortNumber)
{
    int ra = 1;
    int i;
    struct sockaddr_in addr;
    struct ip_mreq mreq;
    unsigned char TTL = 4;
    
    /* Complete State Reset */
    memset(state,0,sizeof(struct RTKMR_DataObject));
    
    /* Setup Notification Timer */
    state->NotifyCycleTime = NotifyCycleSeconds;
    
    #if defined(WIN32) || defined(_WIN32_WCE)
    state->CurrentTime = GetTickCount() / 1000;
    state->NotifyTime = state->CurrentTime  + (state->NotifyCycleTime/2);
    #else
    gettimeofday(&(state->CurrentTime),NULL);
    (state->NotifyTime).tv_sec = (state->CurrentTime).tv_sec  + (state->NotifyCycleTime/2);
    #endif
    
    memset((char *)&(state->addr), 0, sizeof(state->addr));
    state->addr.sin_family = AF_INET;
    state->addr.sin_addr.s_addr = htonl(INADDR_ANY);
    state->addr.sin_port = (unsigned short)htons(UPNP_PORT);
    state->addrlen = sizeof(state->addr);
    
    
    /* Set up socket */
    state->AddressListLength = ILibGetLocalIPAddressList(&(state->AddressList));
    
    #if defined(WIN32) || defined(_WIN32_WCE)
    state->NOTIFY_SEND_socks = (SOCKET*)malloc(sizeof(int)*(state->AddressListLength));
    #else
    state->NOTIFY_SEND_socks = (int*)malloc(sizeof(int)*(state->AddressListLength));
    #endif
    
    state->NOTIFY_RECEIVE_sock = socket(AF_INET, SOCK_DGRAM, 0);
    memset((char *)&(addr), 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = htonl(INADDR_ANY);
    addr.sin_port = (unsigned short)htons(UPNP_PORT);
    if (setsockopt(state->NOTIFY_RECEIVE_sock, SOL_SOCKET, SO_REUSEADDR,(char*)&ra, sizeof(ra)) < 0)
    {
        printf("Setting SockOpt SO_REUSEADDR failed\r\n");
        exit(1);
    }
    if (bind(state->NOTIFY_RECEIVE_sock, (struct sockaddr *) &(addr), sizeof(addr)) < 0)
    {
        printf("Could not bind to UPnP Listen Port\r\n");
        exit(1);
    }
    
    //
    // Iterate through all the current IP Addresses
    //
    for(i=0;i<state->AddressListLength;++i)
    {
#if 1 ////def OLD_SOCKET
        state->NOTIFY_SEND_socks[i] = socket(AF_INET, SOCK_DGRAM, 0);
        memset((char *)&(addr), 0, sizeof(addr));
        addr.sin_family = AF_INET;
        addr.sin_addr.s_addr = state->AddressList[i];
        addr.sin_port = (unsigned short)htons(UPNP_PORT);
#else
        state->NOTIFY_SEND_socks[i] = state->NOTIFY_RECEIVE_sock;
#endif 
        //
        // Set re-use address
        //
#if 1 ////def OLD_SOCKET
        if (setsockopt(state->NOTIFY_SEND_socks[i], SOL_SOCKET, SO_REUSEADDR,(char*)&ra, sizeof(ra)) == 0)
#endif 
        {
            //
            // Set the Multicast TTL
            //
            if (setsockopt(state->NOTIFY_SEND_socks[i], IPPROTO_IP, IP_MULTICAST_TTL,(char*)&TTL, sizeof(TTL)) < 0)
            {
                /* Ignore this case */
            }
            
            //
            // Bind the socket
            //
#if 1 ////def OLD_SOCKET
            if (bind(state->NOTIFY_SEND_socks[i], (struct sockaddr *) &(addr), sizeof(addr)) == 0)
#endif 
            {
                mreq.imr_multiaddr.s_addr = inet_addr(UPNP_GROUP);
                mreq.imr_interface.s_addr = state->AddressList[i];
                //
                // Join the multicast group
                //
                if (setsockopt(state->NOTIFY_RECEIVE_sock, IPPROTO_IP, IP_ADD_MEMBERSHIP,(char*)&mreq, sizeof(mreq)) < 0)
                {
                    /* Does not matter */
                }
            }
        }
    }
    
}
void RTKMR_PostMX_Destroy(void *object)
{
    struct MSEARCH_state *mss = (struct MSEARCH_state*)object;
    free(mss->ST);
    free(mss);
}
void RTKMR_PostMX_MSEARCH(void *object)
{
    struct MSEARCH_state *mss = (struct MSEARCH_state*)object;
    
    char *b = (char*)malloc(sizeof(char)*5000);
    int packetlength;
    struct sockaddr_in response_addr;
    int response_addrlen;
    #if defined(WIN32) || defined(_WIN32_WCE)
    SOCKET *response_socket;
    #else
    int *response_socket;
    #endif
    
    int cnt;
    int i;
    struct sockaddr_in dest_addr = mss->dest_addr;
    char *ST = mss->ST;
    int STLength = mss->STLength;
    struct RTKMR_DataObject *upnp = (struct RTKMR_DataObject*)mss->upnp;
    
    #if defined(WIN32) || defined(_WIN32_WCE)
    response_socket = (SOCKET*)malloc(upnp->AddressListLength*sizeof(int));
    #else
    response_socket = (int*)malloc(upnp->AddressListLength*sizeof(int));
    #endif
    
    //
    // Iterate through all the current IP Addresses
    //
    for(i=0;i<upnp->AddressListLength;++i)
    {
        //
        // Create a socket to respond with
        //
        response_socket[i] = socket(AF_INET, SOCK_DGRAM, 0);
        if (response_socket[i]< 0)
        {
            printf("response socket");
            exit(1);
        }
        memset((char *)&(response_addr), 0, sizeof(response_addr));
        response_addr.sin_family = AF_INET;
        response_addr.sin_addr.s_addr = upnp->AddressList[i];
        response_addr.sin_port = (unsigned short)htons(0);
        response_addrlen = sizeof(response_addr);	
        if (bind(response_socket[i], (struct sockaddr *) &(response_addr), sizeof(response_addr)) < 0)
        {
            /* Ignore if this happens */
        }
    }
    
    //
    // Search for root device
    //
    if(STLength==15 && memcmp(ST,"upnp:rootdevice",15)==0)
    {
        for(i=0;i<upnp->AddressListLength;++i)
        {
            
            RTKMR_BuildSsdpResponsePacket(b,&packetlength,upnp->AddressList[i],(unsigned short)upnp->WebSocketPortNumber,0,upnp->UDN,"::upnp:rootdevice","upnp:rootdevice","",upnp->NotifyCycleTime);
            
            
            cnt = sendto(response_socket[i], b, packetlength, 0,(struct sockaddr *) &dest_addr, sizeof(dest_addr));
        }
    }
    //
    // Search for everything
    //
    else if(STLength==8 && memcmp(ST,"ssdp:all",8)==0)
    {
        for(i=0;i<upnp->AddressListLength;++i)
        {
            RTKMR_BuildSsdpResponsePacket(b,&packetlength,upnp->AddressList[i],(unsigned short)upnp->WebSocketPortNumber,0,upnp->UDN,"::upnp:rootdevice","upnp:rootdevice","",upnp->NotifyCycleTime);
            cnt = sendto(response_socket[i], b, packetlength, 0,
            (struct sockaddr *) &dest_addr, sizeof(dest_addr));
            RTKMR_BuildSsdpResponsePacket(b,&packetlength,upnp->AddressList[i],(unsigned short)upnp->WebSocketPortNumber,0,upnp->UDN,"",upnp->UUID,"",upnp->NotifyCycleTime);
            cnt = sendto(response_socket[i], b, packetlength, 0,
            (struct sockaddr *) &dest_addr, sizeof(dest_addr));
            RTKMR_BuildSsdpResponsePacket(b,&packetlength,upnp->AddressList[i],(unsigned short)upnp->WebSocketPortNumber,0,upnp->UDN,"::urn:schemas-upnp-org:device:MediaRenderer:1","urn:schemas-upnp-org:device:MediaRenderer:1","",upnp->NotifyCycleTime);
            cnt = sendto(response_socket[i], b, packetlength, 0,
            (struct sockaddr *) &dest_addr, sizeof(dest_addr));
/*
            RTKMR_BuildSsdpResponsePacket(b,&packetlength,upnp->AddressList[i],(unsigned short)upnp->WebSocketPortNumber,0,upnp->UDN,"::urn:schemas-upnp-org:service:AVTransport:1","urn:schemas-upnp-org:service:AVTransport:1","",upnp->NotifyCycleTime);
            cnt = sendto(response_socket[i], b, packetlength, 0,
            (struct sockaddr *) &dest_addr, sizeof(dest_addr));
            RTKMR_BuildSsdpResponsePacket(b,&packetlength,upnp->AddressList[i],(unsigned short)upnp->WebSocketPortNumber,0,upnp->UDN,"::urn:schemas-upnp-org:service:ConnectionManager:1","urn:schemas-upnp-org:service:ConnectionManager:1","",upnp->NotifyCycleTime);
            cnt = sendto(response_socket[i], b, packetlength, 0,
            (struct sockaddr *) &dest_addr, sizeof(dest_addr));
            RTKMR_BuildSsdpResponsePacket(b,&packetlength,upnp->AddressList[i],(unsigned short)upnp->WebSocketPortNumber,0,upnp->UDN,"::urn:schemas-upnp-org:service:RenderingControl:1","urn:schemas-upnp-org:service:RenderingControl:1","",upnp->NotifyCycleTime);
            cnt = sendto(response_socket[i], b, packetlength, 0,
            (struct sockaddr *) &dest_addr, sizeof(dest_addr));
*/
        }
        
    }
    if(STLength==(int)strlen(upnp->UUID) && memcmp(ST,upnp->UUID,(int)strlen(upnp->UUID))==0)
    {
        for(i=0;i<upnp->AddressListLength;++i)
        {
            RTKMR_BuildSsdpResponsePacket(b,&packetlength,upnp->AddressList[i],(unsigned short)upnp->WebSocketPortNumber,0,upnp->UDN,"",upnp->UUID,"",upnp->NotifyCycleTime);
            cnt = sendto(response_socket[i], b, packetlength, 0,
            (struct sockaddr *) &dest_addr, sizeof(dest_addr));
        }
    }
    if(STLength>=42 && memcmp(ST,"urn:schemas-upnp-org:device:MediaRenderer:",42)==0 && atoi(ST+42)<=1)
    {
        for(i=0;i<upnp->AddressListLength;++i)
        {
            RTKMR_BuildSsdpResponsePacket(b,&packetlength,upnp->AddressList[i],(unsigned short)upnp->WebSocketPortNumber,0,upnp->UDN,"::urn:schemas-upnp-org:device:MediaRenderer:1",ST,"",upnp->NotifyCycleTime);
            cnt = sendto(response_socket[i], b, packetlength, 0,
            (struct sockaddr *) &dest_addr, sizeof(dest_addr));
        }
    }
/*
    if(STLength>=41 && memcmp(ST,"urn:schemas-upnp-org:service:AVTransport:",41)==0 && atoi(ST+41)<=1)
    {
        for(i=0;i<upnp->AddressListLength;++i)
        {
            RTKMR_BuildSsdpResponsePacket(b,&packetlength,upnp->AddressList[i],(unsigned short)upnp->WebSocketPortNumber,0,upnp->UDN,"::urn:schemas-upnp-org:service:AVTransport:1",ST,"",upnp->NotifyCycleTime);
            cnt = sendto(response_socket[i], b, packetlength, 0,
            (struct sockaddr *) &dest_addr, sizeof(dest_addr));
        }
    }
*/
/*
    if(STLength>=47 && memcmp(ST,"urn:schemas-upnp-org:service:ConnectionManager:",47)==0 && atoi(ST+47)<=1)
    {
        for(i=0;i<upnp->AddressListLength;++i)
        {
            RTKMR_BuildSsdpResponsePacket(b,&packetlength,upnp->AddressList[i],(unsigned short)upnp->WebSocketPortNumber,0,upnp->UDN,"::urn:schemas-upnp-org:service:ConnectionManager:1",ST,"",upnp->NotifyCycleTime);
            cnt = sendto(response_socket[i], b, packetlength, 0,
            (struct sockaddr *) &dest_addr, sizeof(dest_addr));
        }
    }
*/
/*
    if(STLength>=46 && memcmp(ST,"urn:schemas-upnp-org:service:RenderingControl:",46)==0 && atoi(ST+46)<=1)
    {
        for(i=0;i<upnp->AddressListLength;++i)
        {
            RTKMR_BuildSsdpResponsePacket(b,&packetlength,upnp->AddressList[i],(unsigned short)upnp->WebSocketPortNumber,0,upnp->UDN,"::urn:schemas-upnp-org:service:RenderingControl:1",ST,"",upnp->NotifyCycleTime);
            cnt = sendto(response_socket[i], b, packetlength, 0,
            (struct sockaddr *) &dest_addr, sizeof(dest_addr));
        }
    }
*/
    
    
    for(i=0;i<upnp->AddressListLength;++i)
    {
        #if defined(WIN32) || defined(_WIN32_WCE)
        closesocket(response_socket[i]);
        #else
        close(response_socket[i]);
        #endif
    }
    free(response_socket);
    free(mss->ST);
    free(mss);
    free(b);
}
void RTKMR_ProcessMSEARCH(struct RTKMR_DataObject *upnp, struct packetheader *packet)
{
    char* ST = NULL;
    int STLength = 0;
    struct packetheader_field_node *node;
    int MANOK = 0;
    unsigned long MXVal;
    int MXOK = 0;
    int MX;
    struct MSEARCH_state *mss = NULL;
    
    if(memcmp(packet->DirectiveObj,"*",1)==0)
    {
        if(memcmp(packet->Version,"1.1",3)==0)
        {
            node = packet->FirstField;
            while(node!=NULL)
            {
                if(node->FieldLength==2 && strncasecmp(node->Field,"ST",2)==0)
                {
                    //
                    // This is what is being searched for
                    //
                    ST = (char*)malloc(1+node->FieldDataLength);
                    memcpy(ST,node->FieldData,node->FieldDataLength);
                    ST[node->FieldDataLength] = 0;
                    STLength = node->FieldDataLength;
                }
                else if(node->FieldLength==3 && strncasecmp(node->Field,"MAN",3)==0 && memcmp(node->FieldData,"\"ssdp:discover\"",15)==0)
                {
                    //
                    // This is a required header field
                    // 
                    MANOK = 1;
                }
                else if(node->FieldLength==2 && strncasecmp(node->Field,"MX",2)==0 && ILibGetULong(node->FieldData,node->FieldDataLength,&MXVal)==0)
                {
                    //
                    // If the timeout value specified is greater than 10 seconds, just force it
                    // down to 10 seconds
                    //
                    MXOK = 1;
                    MXVal = MXVal>10?10:MXVal;
                }
                node = node->NextField;
            }
            if(MANOK!=0 && MXOK!=0)
            {
                if(MXVal==0)
                {
                    MX = 0;
                }
                else
                {
                    //
                    // The timeout value should be a random number between 0 and the 
                    // specified value
                    //
                    MX = (int)(0 + ((unsigned short)rand() % MXVal));
                }
                mss = (struct MSEARCH_state*)malloc(sizeof(struct MSEARCH_state));
                mss->ST = ST;
                mss->STLength = STLength;
                mss->upnp = upnp;
                memset((char *)&(mss->dest_addr), 0, sizeof(mss->dest_addr));
                mss->dest_addr.sin_family = AF_INET;
                mss->dest_addr.sin_addr = packet->Source->sin_addr;
                mss->dest_addr.sin_port = packet->Source->sin_port;
                
                //
                // Register for a timed callback, so we can respond later
                //
                ILibLifeTime_Add(upnp->WebServerTimer,mss,MX,&RTKMR_PostMX_MSEARCH,&RTKMR_PostMX_Destroy);
            }
            else
            {
                free(ST);
            }
        }
    }
}
void RTKMR_Dispatch_AVTransport_GetDeviceCapabilities(char *buffer, int offset, int bufferLength, struct ILibWebServer_Session *ReaderObject)
{
    unsigned long TempULong;
    int OK = 0;
    char *p_InstanceID = NULL;
    int p_InstanceIDLength = 0;
    unsigned int _InstanceID = 0;
    struct ILibXMLNode *xnode = ILibParseXML(buffer,offset,bufferLength);
    struct ILibXMLNode *root = xnode;
    if(ILibProcessXMLNodeList(root)!=0)
    {
        /* The XML is not well formed! */
        ILibDestructXMLNodeList(root);
        RTKMR_Response_Error(ReaderObject,501,"Invalid XML");
        return;
    }
    while(xnode!=NULL)
    {
        if(xnode->StartTag!=0 && xnode->NameLength==8 && memcmp(xnode->Name,"Envelope",8)==0)
        {
            // Envelope
            xnode = xnode->Next;
            while(xnode!=NULL)
            {
                if(xnode->StartTag!=0 && xnode->NameLength==4 && memcmp(xnode->Name,"Body",4)==0)
                {
                    // Body
                    xnode = xnode->Next;
                    while(xnode!=NULL)
                    {
                        if(xnode->StartTag!=0 && xnode->NameLength==21 && memcmp(xnode->Name,"GetDeviceCapabilities",21)==0)
                        {
                            // Inside the interesting part of the SOAP
                            xnode = xnode->Next;
                            while(xnode!=NULL)
                            {
                                if(xnode->NameLength==10 && memcmp(xnode->Name,"InstanceID",10)==0)
                                {
                                    p_InstanceIDLength = ILibReadInnerXML(xnode,&p_InstanceID);
                                    OK |= 1;
                                }
                                if(xnode->Peer==NULL)
                                {
                                    xnode = xnode->Parent;
                                    break;
                                }
                                else
                                {
                                    xnode = xnode->Peer;
                                }
                            }
                        }
                        if(xnode->Peer==NULL)
                        {
                            xnode = xnode->Parent;
                            break;
                        }
                        else
                        {
                            xnode = xnode->Peer;
                        }
                    }
                }
                if(xnode->Peer==NULL)
                {
                    xnode = xnode->Parent;
                    break;
                }
                else
                {
                    xnode = xnode->Peer;
                }
            }
        }
        xnode = xnode->Peer;
    }
    ILibDestructXMLNodeList(root);
    if (OK != 1)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    
    /* Type Checking */
    OK = ILibGetULong(p_InstanceID,p_InstanceIDLength, &TempULong);
    if(OK!=0)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    _InstanceID = (unsigned int)TempULong;
    RTKMR_FP_AVTransport_GetDeviceCapabilities((void*)ReaderObject,_InstanceID);
}

void RTKMR_Dispatch_AVTransport_GetMediaInfo(char *buffer, int offset, int bufferLength, struct ILibWebServer_Session *ReaderObject)
{
    unsigned long TempULong;
    int OK = 0;
    char *p_InstanceID = NULL;
    int p_InstanceIDLength = 0;
    unsigned int _InstanceID = 0;
    struct ILibXMLNode *xnode = ILibParseXML(buffer,offset,bufferLength);
    struct ILibXMLNode *root = xnode;
    if(ILibProcessXMLNodeList(root)!=0)
    {
        /* The XML is not well formed! */
        ILibDestructXMLNodeList(root);
        RTKMR_Response_Error(ReaderObject,501,"Invalid XML");
        return;
    }
    while(xnode!=NULL)
    {
        if(xnode->StartTag!=0 && xnode->NameLength==8 && memcmp(xnode->Name,"Envelope",8)==0)
        {
            // Envelope
            xnode = xnode->Next;
            while(xnode!=NULL)
            {
                if(xnode->StartTag!=0 && xnode->NameLength==4 && memcmp(xnode->Name,"Body",4)==0)
                {
                    // Body
                    xnode = xnode->Next;
                    while(xnode!=NULL)
                    {
                        if(xnode->StartTag!=0 && xnode->NameLength==12 && memcmp(xnode->Name,"GetMediaInfo",12)==0)
                        {
                            // Inside the interesting part of the SOAP
                            xnode = xnode->Next;
                            while(xnode!=NULL)
                            {
                                if(xnode->NameLength==10 && memcmp(xnode->Name,"InstanceID",10)==0)
                                {
                                    p_InstanceIDLength = ILibReadInnerXML(xnode,&p_InstanceID);
                                    OK |= 1;
                                }
                                if(xnode->Peer==NULL)
                                {
                                    xnode = xnode->Parent;
                                    break;
                                }
                                else
                                {
                                    xnode = xnode->Peer;
                                }
                            }
                        }
                        if(xnode->Peer==NULL)
                        {
                            xnode = xnode->Parent;
                            break;
                        }
                        else
                        {
                            xnode = xnode->Peer;
                        }
                    }
                }
                if(xnode->Peer==NULL)
                {
                    xnode = xnode->Parent;
                    break;
                }
                else
                {
                    xnode = xnode->Peer;
                }
            }
        }
        xnode = xnode->Peer;
    }
    ILibDestructXMLNodeList(root);
    if (OK != 1)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    
    /* Type Checking */
    OK = ILibGetULong(p_InstanceID,p_InstanceIDLength, &TempULong);
    if(OK!=0)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    _InstanceID = (unsigned int)TempULong;
    RTKMR_FP_AVTransport_GetMediaInfo((void*)ReaderObject,_InstanceID);
}

void RTKMR_Dispatch_AVTransport_GetPositionInfo(char *buffer, int offset, int bufferLength, struct ILibWebServer_Session *ReaderObject)
{
    unsigned long TempULong;
    int OK = 0;
    char *p_InstanceID = NULL;
    int p_InstanceIDLength = 0;
    unsigned int _InstanceID = 0;
    struct ILibXMLNode *xnode = ILibParseXML(buffer,offset,bufferLength);
    struct ILibXMLNode *root = xnode;
    if(ILibProcessXMLNodeList(root)!=0)
    {
        /* The XML is not well formed! */
        ILibDestructXMLNodeList(root);
        RTKMR_Response_Error(ReaderObject,501,"Invalid XML");
        return;
    }
    while(xnode!=NULL)
    {
        if(xnode->StartTag!=0 && xnode->NameLength==8 && memcmp(xnode->Name,"Envelope",8)==0)
        {
            // Envelope
            xnode = xnode->Next;
            while(xnode!=NULL)
            {
                if(xnode->StartTag!=0 && xnode->NameLength==4 && memcmp(xnode->Name,"Body",4)==0)
                {
                    // Body
                    xnode = xnode->Next;
                    while(xnode!=NULL)
                    {
                        if(xnode->StartTag!=0 && xnode->NameLength==15 && memcmp(xnode->Name,"GetPositionInfo",15)==0)
                        {
                            // Inside the interesting part of the SOAP
                            xnode = xnode->Next;
                            while(xnode!=NULL)
                            {
                                if(xnode->NameLength==10 && memcmp(xnode->Name,"InstanceID",10)==0)
                                {
                                    p_InstanceIDLength = ILibReadInnerXML(xnode,&p_InstanceID);
                                    OK |= 1;
                                }
                                if(xnode->Peer==NULL)
                                {
                                    xnode = xnode->Parent;
                                    break;
                                }
                                else
                                {
                                    xnode = xnode->Peer;
                                }
                            }
                        }
                        if(xnode->Peer==NULL)
                        {
                            xnode = xnode->Parent;
                            break;
                        }
                        else
                        {
                            xnode = xnode->Peer;
                        }
                    }
                }
                if(xnode->Peer==NULL)
                {
                    xnode = xnode->Parent;
                    break;
                }
                else
                {
                    xnode = xnode->Peer;
                }
            }
        }
        xnode = xnode->Peer;
    }
    ILibDestructXMLNodeList(root);
    if (OK != 1)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    
    /* Type Checking */
    OK = ILibGetULong(p_InstanceID,p_InstanceIDLength, &TempULong);
    if(OK!=0)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    _InstanceID = (unsigned int)TempULong;
    RTKMR_FP_AVTransport_GetPositionInfo((void*)ReaderObject,_InstanceID);
}

void RTKMR_Dispatch_AVTransport_GetTransportInfo(char *buffer, int offset, int bufferLength, struct ILibWebServer_Session *ReaderObject)
{
    unsigned long TempULong;
    int OK = 0;
    char *p_InstanceID = NULL;
    int p_InstanceIDLength = 0;
    unsigned int _InstanceID = 0;
    struct ILibXMLNode *xnode = ILibParseXML(buffer,offset,bufferLength);
    struct ILibXMLNode *root = xnode;
    if(ILibProcessXMLNodeList(root)!=0)
    {
        /* The XML is not well formed! */
        ILibDestructXMLNodeList(root);
        RTKMR_Response_Error(ReaderObject,501,"Invalid XML");
        return;
    }
    while(xnode!=NULL)
    {
        if(xnode->StartTag!=0 && xnode->NameLength==8 && memcmp(xnode->Name,"Envelope",8)==0)
        {
            // Envelope
            xnode = xnode->Next;
            while(xnode!=NULL)
            {
                if(xnode->StartTag!=0 && xnode->NameLength==4 && memcmp(xnode->Name,"Body",4)==0)
                {
                    // Body
                    xnode = xnode->Next;
                    while(xnode!=NULL)
                    {
                        if(xnode->StartTag!=0 && xnode->NameLength==16 && memcmp(xnode->Name,"GetTransportInfo",16)==0)
                        {
                            // Inside the interesting part of the SOAP
                            xnode = xnode->Next;
                            while(xnode!=NULL)
                            {
                                if(xnode->NameLength==10 && memcmp(xnode->Name,"InstanceID",10)==0)
                                {
                                    p_InstanceIDLength = ILibReadInnerXML(xnode,&p_InstanceID);
                                    OK |= 1;
                                }
                                if(xnode->Peer==NULL)
                                {
                                    xnode = xnode->Parent;
                                    break;
                                }
                                else
                                {
                                    xnode = xnode->Peer;
                                }
                            }
                        }
                        if(xnode->Peer==NULL)
                        {
                            xnode = xnode->Parent;
                            break;
                        }
                        else
                        {
                            xnode = xnode->Peer;
                        }
                    }
                }
                if(xnode->Peer==NULL)
                {
                    xnode = xnode->Parent;
                    break;
                }
                else
                {
                    xnode = xnode->Peer;
                }
            }
        }
        xnode = xnode->Peer;
    }
    ILibDestructXMLNodeList(root);
    if (OK != 1)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    
    /* Type Checking */
    OK = ILibGetULong(p_InstanceID,p_InstanceIDLength, &TempULong);
    if(OK!=0)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    _InstanceID = (unsigned int)TempULong;
    RTKMR_FP_AVTransport_GetTransportInfo((void*)ReaderObject,_InstanceID);
}

void RTKMR_Dispatch_AVTransport_GetTransportSettings(char *buffer, int offset, int bufferLength, struct ILibWebServer_Session *ReaderObject)
{
    unsigned long TempULong;
    int OK = 0;
    char *p_InstanceID = NULL;
    int p_InstanceIDLength = 0;
    unsigned int _InstanceID = 0;
    struct ILibXMLNode *xnode = ILibParseXML(buffer,offset,bufferLength);
    struct ILibXMLNode *root = xnode;
    if(ILibProcessXMLNodeList(root)!=0)
    {
        /* The XML is not well formed! */
        ILibDestructXMLNodeList(root);
        RTKMR_Response_Error(ReaderObject,501,"Invalid XML");
        return;
    }
    while(xnode!=NULL)
    {
        if(xnode->StartTag!=0 && xnode->NameLength==8 && memcmp(xnode->Name,"Envelope",8)==0)
        {
            // Envelope
            xnode = xnode->Next;
            while(xnode!=NULL)
            {
                if(xnode->StartTag!=0 && xnode->NameLength==4 && memcmp(xnode->Name,"Body",4)==0)
                {
                    // Body
                    xnode = xnode->Next;
                    while(xnode!=NULL)
                    {
                        if(xnode->StartTag!=0 && xnode->NameLength==20 && memcmp(xnode->Name,"GetTransportSettings",20)==0)
                        {
                            // Inside the interesting part of the SOAP
                            xnode = xnode->Next;
                            while(xnode!=NULL)
                            {
                                if(xnode->NameLength==10 && memcmp(xnode->Name,"InstanceID",10)==0)
                                {
                                    p_InstanceIDLength = ILibReadInnerXML(xnode,&p_InstanceID);
                                    OK |= 1;
                                }
                                if(xnode->Peer==NULL)
                                {
                                    xnode = xnode->Parent;
                                    break;
                                }
                                else
                                {
                                    xnode = xnode->Peer;
                                }
                            }
                        }
                        if(xnode->Peer==NULL)
                        {
                            xnode = xnode->Parent;
                            break;
                        }
                        else
                        {
                            xnode = xnode->Peer;
                        }
                    }
                }
                if(xnode->Peer==NULL)
                {
                    xnode = xnode->Parent;
                    break;
                }
                else
                {
                    xnode = xnode->Peer;
                }
            }
        }
        xnode = xnode->Peer;
    }
    ILibDestructXMLNodeList(root);
    if (OK != 1)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    
    /* Type Checking */
    OK = ILibGetULong(p_InstanceID,p_InstanceIDLength, &TempULong);
    if(OK!=0)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    _InstanceID = (unsigned int)TempULong;
    RTKMR_FP_AVTransport_GetTransportSettings((void*)ReaderObject,_InstanceID);
}

void RTKMR_Dispatch_AVTransport_Next(char *buffer, int offset, int bufferLength, struct ILibWebServer_Session *ReaderObject)
{
    unsigned long TempULong;
    int OK = 0;
    char *p_InstanceID = NULL;
    int p_InstanceIDLength = 0;
    unsigned int _InstanceID = 0;
    struct ILibXMLNode *xnode = ILibParseXML(buffer,offset,bufferLength);
    struct ILibXMLNode *root = xnode;
    if(ILibProcessXMLNodeList(root)!=0)
    {
        /* The XML is not well formed! */
        ILibDestructXMLNodeList(root);
        RTKMR_Response_Error(ReaderObject,501,"Invalid XML");
        return;
    }
    while(xnode!=NULL)
    {
        if(xnode->StartTag!=0 && xnode->NameLength==8 && memcmp(xnode->Name,"Envelope",8)==0)
        {
            // Envelope
            xnode = xnode->Next;
            while(xnode!=NULL)
            {
                if(xnode->StartTag!=0 && xnode->NameLength==4 && memcmp(xnode->Name,"Body",4)==0)
                {
                    // Body
                    xnode = xnode->Next;
                    while(xnode!=NULL)
                    {
                        if(xnode->StartTag!=0 && xnode->NameLength==4 && memcmp(xnode->Name,"Next",4)==0)
                        {
                            // Inside the interesting part of the SOAP
                            xnode = xnode->Next;
                            while(xnode!=NULL)
                            {
                                if(xnode->NameLength==10 && memcmp(xnode->Name,"InstanceID",10)==0)
                                {
                                    p_InstanceIDLength = ILibReadInnerXML(xnode,&p_InstanceID);
                                    OK |= 1;
                                }
                                if(xnode->Peer==NULL)
                                {
                                    xnode = xnode->Parent;
                                    break;
                                }
                                else
                                {
                                    xnode = xnode->Peer;
                                }
                            }
                        }
                        if(xnode->Peer==NULL)
                        {
                            xnode = xnode->Parent;
                            break;
                        }
                        else
                        {
                            xnode = xnode->Peer;
                        }
                    }
                }
                if(xnode->Peer==NULL)
                {
                    xnode = xnode->Parent;
                    break;
                }
                else
                {
                    xnode = xnode->Peer;
                }
            }
        }
        xnode = xnode->Peer;
    }
    ILibDestructXMLNodeList(root);
    if (OK != 1)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    
    /* Type Checking */
    OK = ILibGetULong(p_InstanceID,p_InstanceIDLength, &TempULong);
    if(OK!=0)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    _InstanceID = (unsigned int)TempULong;
    RTKMR_FP_AVTransport_Next((void*)ReaderObject,_InstanceID);
}

void RTKMR_Dispatch_AVTransport_Pause(char *buffer, int offset, int bufferLength, struct ILibWebServer_Session *ReaderObject)
{
    unsigned long TempULong;
    int OK = 0;
    char *p_InstanceID = NULL;
    int p_InstanceIDLength = 0;
    unsigned int _InstanceID = 0;
    struct ILibXMLNode *xnode = ILibParseXML(buffer,offset,bufferLength);
    struct ILibXMLNode *root = xnode;
    if(ILibProcessXMLNodeList(root)!=0)
    {
        /* The XML is not well formed! */
        ILibDestructXMLNodeList(root);
        RTKMR_Response_Error(ReaderObject,501,"Invalid XML");
        return;
    }
    while(xnode!=NULL)
    {
        if(xnode->StartTag!=0 && xnode->NameLength==8 && memcmp(xnode->Name,"Envelope",8)==0)
        {
            // Envelope
            xnode = xnode->Next;
            while(xnode!=NULL)
            {
                if(xnode->StartTag!=0 && xnode->NameLength==4 && memcmp(xnode->Name,"Body",4)==0)
                {
                    // Body
                    xnode = xnode->Next;
                    while(xnode!=NULL)
                    {
                        if(xnode->StartTag!=0 && xnode->NameLength==5 && memcmp(xnode->Name,"Pause",5)==0)
                        {
                            // Inside the interesting part of the SOAP
                            xnode = xnode->Next;
                            while(xnode!=NULL)
                            {
                                if(xnode->NameLength==10 && memcmp(xnode->Name,"InstanceID",10)==0)
                                {
                                    p_InstanceIDLength = ILibReadInnerXML(xnode,&p_InstanceID);
                                    OK |= 1;
                                }
                                if(xnode->Peer==NULL)
                                {
                                    xnode = xnode->Parent;
                                    break;
                                }
                                else
                                {
                                    xnode = xnode->Peer;
                                }
                            }
                        }
                        if(xnode->Peer==NULL)
                        {
                            xnode = xnode->Parent;
                            break;
                        }
                        else
                        {
                            xnode = xnode->Peer;
                        }
                    }
                }
                if(xnode->Peer==NULL)
                {
                    xnode = xnode->Parent;
                    break;
                }
                else
                {
                    xnode = xnode->Peer;
                }
            }
        }
        xnode = xnode->Peer;
    }
    ILibDestructXMLNodeList(root);
    if (OK != 1)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    
    /* Type Checking */
    OK = ILibGetULong(p_InstanceID,p_InstanceIDLength, &TempULong);
    if(OK!=0)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    _InstanceID = (unsigned int)TempULong;
    RTKMR_FP_AVTransport_Pause((void*)ReaderObject,_InstanceID);
}

void RTKMR_Dispatch_AVTransport_Play(char *buffer, int offset, int bufferLength, struct ILibWebServer_Session *ReaderObject)
{
    unsigned long TempULong;
    int OK = 0;
    char *p_InstanceID = NULL;
    int p_InstanceIDLength = 0;
    unsigned int _InstanceID = 0;
    char *p_Speed = NULL;
    int p_SpeedLength = 0;
    char* _Speed = "";
    int _SpeedLength;
    struct ILibXMLNode *xnode = ILibParseXML(buffer,offset,bufferLength);
    struct ILibXMLNode *root = xnode;
    if(ILibProcessXMLNodeList(root)!=0)
    {
        /* The XML is not well formed! */
        ILibDestructXMLNodeList(root);
        RTKMR_Response_Error(ReaderObject,501,"Invalid XML");
        return;
    }
    while(xnode!=NULL)
    {
        if(xnode->StartTag!=0 && xnode->NameLength==8 && memcmp(xnode->Name,"Envelope",8)==0)
        {
            // Envelope
            xnode = xnode->Next;
            while(xnode!=NULL)
            {
                if(xnode->StartTag!=0 && xnode->NameLength==4 && memcmp(xnode->Name,"Body",4)==0)
                {
                    // Body
                    xnode = xnode->Next;
                    while(xnode!=NULL)
                    {
                        if(xnode->StartTag!=0 && xnode->NameLength==4 && memcmp(xnode->Name,"Play",4)==0)
                        {
                            // Inside the interesting part of the SOAP
                            xnode = xnode->Next;
                            while(xnode!=NULL)
                            {
                                if(xnode->NameLength==10 && memcmp(xnode->Name,"InstanceID",10)==0)
                                {
                                    p_InstanceIDLength = ILibReadInnerXML(xnode,&p_InstanceID);
                                    OK |= 1;
                                }
                                else if(xnode->NameLength==5 && memcmp(xnode->Name,"Speed",5)==0)
                                {
                                    p_SpeedLength = ILibReadInnerXML(xnode,&p_Speed);
                                    p_Speed[p_SpeedLength]=0;
                                    OK |= 2;
                                }
                                if(xnode->Peer==NULL)
                                {
                                    xnode = xnode->Parent;
                                    break;
                                }
                                else
                                {
                                    xnode = xnode->Peer;
                                }
                            }
                        }
                        if(xnode->Peer==NULL)
                        {
                            xnode = xnode->Parent;
                            break;
                        }
                        else
                        {
                            xnode = xnode->Peer;
                        }
                    }
                }
                if(xnode->Peer==NULL)
                {
                    xnode = xnode->Parent;
                    break;
                }
                else
                {
                    xnode = xnode->Peer;
                }
            }
        }
        xnode = xnode->Peer;
    }
    ILibDestructXMLNodeList(root);
    if (OK != 3)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    
    /* Type Checking */
    OK = ILibGetULong(p_InstanceID,p_InstanceIDLength, &TempULong);
    if(OK!=0)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    _InstanceID = (unsigned int)TempULong;
    _SpeedLength = ILibInPlaceXmlUnEscape(p_Speed);
    _Speed = p_Speed;
    if(memcmp(_Speed, "1\0",2) != 0
    && memcmp(_Speed, " vendor-defined \0",17) != 0
    )
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    RTKMR_FP_AVTransport_Play((void*)ReaderObject,_InstanceID,_Speed);
}

void RTKMR_Dispatch_AVTransport_Previous(char *buffer, int offset, int bufferLength, struct ILibWebServer_Session *ReaderObject)
{
    unsigned long TempULong;
    int OK = 0;
    char *p_InstanceID = NULL;
    int p_InstanceIDLength = 0;
    unsigned int _InstanceID = 0;
    struct ILibXMLNode *xnode = ILibParseXML(buffer,offset,bufferLength);
    struct ILibXMLNode *root = xnode;
    if(ILibProcessXMLNodeList(root)!=0)
    {
        /* The XML is not well formed! */
        ILibDestructXMLNodeList(root);
        RTKMR_Response_Error(ReaderObject,501,"Invalid XML");
        return;
    }
    while(xnode!=NULL)
    {
        if(xnode->StartTag!=0 && xnode->NameLength==8 && memcmp(xnode->Name,"Envelope",8)==0)
        {
            // Envelope
            xnode = xnode->Next;
            while(xnode!=NULL)
            {
                if(xnode->StartTag!=0 && xnode->NameLength==4 && memcmp(xnode->Name,"Body",4)==0)
                {
                    // Body
                    xnode = xnode->Next;
                    while(xnode!=NULL)
                    {
                        if(xnode->StartTag!=0 && xnode->NameLength==8 && memcmp(xnode->Name,"Previous",8)==0)
                        {
                            // Inside the interesting part of the SOAP
                            xnode = xnode->Next;
                            while(xnode!=NULL)
                            {
                                if(xnode->NameLength==10 && memcmp(xnode->Name,"InstanceID",10)==0)
                                {
                                    p_InstanceIDLength = ILibReadInnerXML(xnode,&p_InstanceID);
                                    OK |= 1;
                                }
                                if(xnode->Peer==NULL)
                                {
                                    xnode = xnode->Parent;
                                    break;
                                }
                                else
                                {
                                    xnode = xnode->Peer;
                                }
                            }
                        }
                        if(xnode->Peer==NULL)
                        {
                            xnode = xnode->Parent;
                            break;
                        }
                        else
                        {
                            xnode = xnode->Peer;
                        }
                    }
                }
                if(xnode->Peer==NULL)
                {
                    xnode = xnode->Parent;
                    break;
                }
                else
                {
                    xnode = xnode->Peer;
                }
            }
        }
        xnode = xnode->Peer;
    }
    ILibDestructXMLNodeList(root);
    if (OK != 1)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    
    /* Type Checking */
    OK = ILibGetULong(p_InstanceID,p_InstanceIDLength, &TempULong);
    if(OK!=0)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    _InstanceID = (unsigned int)TempULong;
    RTKMR_FP_AVTransport_Previous((void*)ReaderObject,_InstanceID);
}

void RTKMR_Dispatch_AVTransport_Seek(char *buffer, int offset, int bufferLength, struct ILibWebServer_Session *ReaderObject)
{
    unsigned long TempULong;
    int OK = 0;
    char *p_InstanceID = NULL;
    int p_InstanceIDLength = 0;
    unsigned int _InstanceID = 0;
    char *p_Unit = NULL;
    int p_UnitLength = 0;
    char* _Unit = "";
    int _UnitLength;
    char *p_Target = NULL;
    int p_TargetLength = 0;
    char* _Target = "";
    int _TargetLength;
    struct ILibXMLNode *xnode = ILibParseXML(buffer,offset,bufferLength);
    struct ILibXMLNode *root = xnode;
    if(ILibProcessXMLNodeList(root)!=0)
    {
        /* The XML is not well formed! */
        ILibDestructXMLNodeList(root);
        RTKMR_Response_Error(ReaderObject,501,"Invalid XML");
        return;
    }
    while(xnode!=NULL)
    {
        if(xnode->StartTag!=0 && xnode->NameLength==8 && memcmp(xnode->Name,"Envelope",8)==0)
        {
            // Envelope
            xnode = xnode->Next;
            while(xnode!=NULL)
            {
                if(xnode->StartTag!=0 && xnode->NameLength==4 && memcmp(xnode->Name,"Body",4)==0)
                {
                    // Body
                    xnode = xnode->Next;
                    while(xnode!=NULL)
                    {
                        if(xnode->StartTag!=0 && xnode->NameLength==4 && memcmp(xnode->Name,"Seek",4)==0)
                        {
                            // Inside the interesting part of the SOAP
                            xnode = xnode->Next;
                            while(xnode!=NULL)
                            {
                                if(xnode->NameLength==10 && memcmp(xnode->Name,"InstanceID",10)==0)
                                {
                                    p_InstanceIDLength = ILibReadInnerXML(xnode,&p_InstanceID);
                                    OK |= 1;
                                }
                                else if(xnode->NameLength==4 && memcmp(xnode->Name,"Unit",4)==0)
                                {
                                    p_UnitLength = ILibReadInnerXML(xnode,&p_Unit);
                                    p_Unit[p_UnitLength]=0;
                                    OK |= 2;
                                }
                                else if(xnode->NameLength==6 && memcmp(xnode->Name,"Target",6)==0)
                                {
                                    p_TargetLength = ILibReadInnerXML(xnode,&p_Target);
                                    p_Target[p_TargetLength]=0;
                                    OK |= 4;
                                }
                                if(xnode->Peer==NULL)
                                {
                                    xnode = xnode->Parent;
                                    break;
                                }
                                else
                                {
                                    xnode = xnode->Peer;
                                }
                            }
                        }
                        if(xnode->Peer==NULL)
                        {
                            xnode = xnode->Parent;
                            break;
                        }
                        else
                        {
                            xnode = xnode->Peer;
                        }
                    }
                }
                if(xnode->Peer==NULL)
                {
                    xnode = xnode->Parent;
                    break;
                }
                else
                {
                    xnode = xnode->Peer;
                }
            }
        }
        xnode = xnode->Peer;
    }
    ILibDestructXMLNodeList(root);
    if (OK != 7)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    
    /* Type Checking */
    OK = ILibGetULong(p_InstanceID,p_InstanceIDLength, &TempULong);
    if(OK!=0)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    _InstanceID = (unsigned int)TempULong;
    _UnitLength = ILibInPlaceXmlUnEscape(p_Unit);
    _Unit = p_Unit;
    if(memcmp(_Unit, "TRACK_NR\0",9) != 0
    )
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    _TargetLength = ILibInPlaceXmlUnEscape(p_Target);
    _Target = p_Target;
    RTKMR_FP_AVTransport_Seek((void*)ReaderObject,_InstanceID,_Unit,_Target);
}

void RTKMR_Dispatch_AVTransport_SetAVTransportURI(char *buffer, int offset, int bufferLength, struct ILibWebServer_Session *ReaderObject)
{
    unsigned long TempULong;
    int OK = 0;
    char *p_InstanceID = NULL;
    int p_InstanceIDLength = 0;
    unsigned int _InstanceID = 0;
    char *p_CurrentURI = NULL;
    int p_CurrentURILength = 0;
    char* _CurrentURI = "";
    int _CurrentURILength;
    char *p_CurrentURIMetaData = NULL;
    int p_CurrentURIMetaDataLength = 0;
    char* _CurrentURIMetaData = "";
    int _CurrentURIMetaDataLength;
    struct ILibXMLNode *xnode = ILibParseXML(buffer,offset,bufferLength);
    struct ILibXMLNode *root = xnode;
    if(ILibProcessXMLNodeList(root)!=0)
    {
        /* The XML is not well formed! */
        ILibDestructXMLNodeList(root);
        RTKMR_Response_Error(ReaderObject,501,"Invalid XML");
        return;
    }
    while(xnode!=NULL)
    {
        if(xnode->StartTag!=0 && xnode->NameLength==8 && memcmp(xnode->Name,"Envelope",8)==0)
        {
            // Envelope
            xnode = xnode->Next;
            while(xnode!=NULL)
            {
                if(xnode->StartTag!=0 && xnode->NameLength==4 && memcmp(xnode->Name,"Body",4)==0)
                {
                    // Body
                    xnode = xnode->Next;
                    while(xnode!=NULL)
                    {
                        if(xnode->StartTag!=0 && xnode->NameLength==17 && memcmp(xnode->Name,"SetAVTransportURI",17)==0)
                        {
                            // Inside the interesting part of the SOAP
                            xnode = xnode->Next;
                            while(xnode!=NULL)
                            {
                                if(xnode->NameLength==10 && memcmp(xnode->Name,"InstanceID",10)==0)
                                {
                                    p_InstanceIDLength = ILibReadInnerXML(xnode,&p_InstanceID);
                                    OK |= 1;
                                }
                                else if(xnode->NameLength==10 && memcmp(xnode->Name,"CurrentURI",10)==0)
                                {
                                    p_CurrentURILength = ILibReadInnerXML(xnode,&p_CurrentURI);
                                    p_CurrentURI[p_CurrentURILength]=0;
                                    OK |= 2;
                                }
                                else if(xnode->NameLength==18 && memcmp(xnode->Name,"CurrentURIMetaData",18)==0)
                                {
                                    p_CurrentURIMetaDataLength = ILibReadInnerXML(xnode,&p_CurrentURIMetaData);
                                    p_CurrentURIMetaData[p_CurrentURIMetaDataLength]=0;
                                    OK |= 4;
                                }
                                if(xnode->Peer==NULL)
                                {
                                    xnode = xnode->Parent;
                                    break;
                                }
                                else
                                {
                                    xnode = xnode->Peer;
                                }
                            }
                        }
                        if(xnode->Peer==NULL)
                        {
                            xnode = xnode->Parent;
                            break;
                        }
                        else
                        {
                            xnode = xnode->Peer;
                        }
                    }
                }
                if(xnode->Peer==NULL)
                {
                    xnode = xnode->Parent;
                    break;
                }
                else
                {
                    xnode = xnode->Peer;
                }
            }
        }
        xnode = xnode->Peer;
    }
    ILibDestructXMLNodeList(root);
    if (OK != 7)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    
    /* Type Checking */
    OK = ILibGetULong(p_InstanceID,p_InstanceIDLength, &TempULong);
    if(OK!=0)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    _InstanceID = (unsigned int)TempULong;
    _CurrentURILength = ILibInPlaceXmlUnEscape(p_CurrentURI);
    _CurrentURI = p_CurrentURI;
    _CurrentURIMetaDataLength = ILibInPlaceXmlUnEscape(p_CurrentURIMetaData);
    _CurrentURIMetaData = p_CurrentURIMetaData;
    RTKMR_FP_AVTransport_SetAVTransportURI((void*)ReaderObject,_InstanceID,_CurrentURI,_CurrentURIMetaData);
}

void RTKMR_Dispatch_AVTransport_SetPlayMode(char *buffer, int offset, int bufferLength, struct ILibWebServer_Session *ReaderObject)
{
    unsigned long TempULong;
    int OK = 0;
    char *p_InstanceID = NULL;
    int p_InstanceIDLength = 0;
    unsigned int _InstanceID = 0;
    char *p_NewPlayMode = NULL;
    int p_NewPlayModeLength = 0;
    char* _NewPlayMode = "";
    int _NewPlayModeLength;
    struct ILibXMLNode *xnode = ILibParseXML(buffer,offset,bufferLength);
    struct ILibXMLNode *root = xnode;
    if(ILibProcessXMLNodeList(root)!=0)
    {
        /* The XML is not well formed! */
        ILibDestructXMLNodeList(root);
        RTKMR_Response_Error(ReaderObject,501,"Invalid XML");
        return;
    }
    while(xnode!=NULL)
    {
        if(xnode->StartTag!=0 && xnode->NameLength==8 && memcmp(xnode->Name,"Envelope",8)==0)
        {
            // Envelope
            xnode = xnode->Next;
            while(xnode!=NULL)
            {
                if(xnode->StartTag!=0 && xnode->NameLength==4 && memcmp(xnode->Name,"Body",4)==0)
                {
                    // Body
                    xnode = xnode->Next;
                    while(xnode!=NULL)
                    {
                        if(xnode->StartTag!=0 && xnode->NameLength==11 && memcmp(xnode->Name,"SetPlayMode",11)==0)
                        {
                            // Inside the interesting part of the SOAP
                            xnode = xnode->Next;
                            while(xnode!=NULL)
                            {
                                if(xnode->NameLength==10 && memcmp(xnode->Name,"InstanceID",10)==0)
                                {
                                    p_InstanceIDLength = ILibReadInnerXML(xnode,&p_InstanceID);
                                    OK |= 1;
                                }
                                else if(xnode->NameLength==11 && memcmp(xnode->Name,"NewPlayMode",11)==0)
                                {
                                    p_NewPlayModeLength = ILibReadInnerXML(xnode,&p_NewPlayMode);
                                    p_NewPlayMode[p_NewPlayModeLength]=0;
                                    OK |= 2;
                                }
                                if(xnode->Peer==NULL)
                                {
                                    xnode = xnode->Parent;
                                    break;
                                }
                                else
                                {
                                    xnode = xnode->Peer;
                                }
                            }
                        }
                        if(xnode->Peer==NULL)
                        {
                            xnode = xnode->Parent;
                            break;
                        }
                        else
                        {
                            xnode = xnode->Peer;
                        }
                    }
                }
                if(xnode->Peer==NULL)
                {
                    xnode = xnode->Parent;
                    break;
                }
                else
                {
                    xnode = xnode->Peer;
                }
            }
        }
        xnode = xnode->Peer;
    }
    ILibDestructXMLNodeList(root);
    if (OK != 3)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    
    /* Type Checking */
    OK = ILibGetULong(p_InstanceID,p_InstanceIDLength, &TempULong);
    if(OK!=0)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    _InstanceID = (unsigned int)TempULong;
    _NewPlayModeLength = ILibInPlaceXmlUnEscape(p_NewPlayMode);
    _NewPlayMode = p_NewPlayMode;
    if(memcmp(_NewPlayMode, "NORMAL\0",7) != 0
    && memcmp(_NewPlayMode, "REPEAT_ALL\0",11) != 0
    )
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    RTKMR_FP_AVTransport_SetPlayMode((void*)ReaderObject,_InstanceID,_NewPlayMode);
}

void RTKMR_Dispatch_AVTransport_Stop(char *buffer, int offset, int bufferLength, struct ILibWebServer_Session *ReaderObject)
{
    unsigned long TempULong;
    int OK = 0;
    char *p_InstanceID = NULL;
    int p_InstanceIDLength = 0;
    unsigned int _InstanceID = 0;
    struct ILibXMLNode *xnode = ILibParseXML(buffer,offset,bufferLength);
    struct ILibXMLNode *root = xnode;
    if(ILibProcessXMLNodeList(root)!=0)
    {
        /* The XML is not well formed! */
        ILibDestructXMLNodeList(root);
        RTKMR_Response_Error(ReaderObject,501,"Invalid XML");
        return;
    }
    while(xnode!=NULL)
    {
        if(xnode->StartTag!=0 && xnode->NameLength==8 && memcmp(xnode->Name,"Envelope",8)==0)
        {
            // Envelope
            xnode = xnode->Next;
            while(xnode!=NULL)
            {
                if(xnode->StartTag!=0 && xnode->NameLength==4 && memcmp(xnode->Name,"Body",4)==0)
                {
                    // Body
                    xnode = xnode->Next;
                    while(xnode!=NULL)
                    {
                        if(xnode->StartTag!=0 && xnode->NameLength==4 && memcmp(xnode->Name,"Stop",4)==0)
                        {
                            // Inside the interesting part of the SOAP
                            xnode = xnode->Next;
                            while(xnode!=NULL)
                            {
                                if(xnode->NameLength==10 && memcmp(xnode->Name,"InstanceID",10)==0)
                                {
                                    p_InstanceIDLength = ILibReadInnerXML(xnode,&p_InstanceID);
                                    OK |= 1;
                                }
                                if(xnode->Peer==NULL)
                                {
                                    xnode = xnode->Parent;
                                    break;
                                }
                                else
                                {
                                    xnode = xnode->Peer;
                                }
                            }
                        }
                        if(xnode->Peer==NULL)
                        {
                            xnode = xnode->Parent;
                            break;
                        }
                        else
                        {
                            xnode = xnode->Peer;
                        }
                    }
                }
                if(xnode->Peer==NULL)
                {
                    xnode = xnode->Parent;
                    break;
                }
                else
                {
                    xnode = xnode->Peer;
                }
            }
        }
        xnode = xnode->Peer;
    }
    ILibDestructXMLNodeList(root);
    if (OK != 1)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    
    /* Type Checking */
    OK = ILibGetULong(p_InstanceID,p_InstanceIDLength, &TempULong);
    if(OK!=0)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    _InstanceID = (unsigned int)TempULong;
    RTKMR_FP_AVTransport_Stop((void*)ReaderObject,_InstanceID);
}

void RTKMR_Dispatch_ConnectionManager_ConnectionComplete(char *buffer, int offset, int bufferLength, struct ILibWebServer_Session *ReaderObject)
{
    long TempLong;
    int OK = 0;
    char *p_ConnectionID = NULL;
    int p_ConnectionIDLength = 0;
    int _ConnectionID = 0;
    struct ILibXMLNode *xnode = ILibParseXML(buffer,offset,bufferLength);
    struct ILibXMLNode *root = xnode;
    if(ILibProcessXMLNodeList(root)!=0)
    {
        /* The XML is not well formed! */
        ILibDestructXMLNodeList(root);
        RTKMR_Response_Error(ReaderObject,501,"Invalid XML");
        return;
    }
    while(xnode!=NULL)
    {
        if(xnode->StartTag!=0 && xnode->NameLength==8 && memcmp(xnode->Name,"Envelope",8)==0)
        {
            // Envelope
            xnode = xnode->Next;
            while(xnode!=NULL)
            {
                if(xnode->StartTag!=0 && xnode->NameLength==4 && memcmp(xnode->Name,"Body",4)==0)
                {
                    // Body
                    xnode = xnode->Next;
                    while(xnode!=NULL)
                    {
                        if(xnode->StartTag!=0 && xnode->NameLength==18 && memcmp(xnode->Name,"ConnectionComplete",18)==0)
                        {
                            // Inside the interesting part of the SOAP
                            xnode = xnode->Next;
                            while(xnode!=NULL)
                            {
                                if(xnode->NameLength==12 && memcmp(xnode->Name,"ConnectionID",12)==0)
                                {
                                    p_ConnectionIDLength = ILibReadInnerXML(xnode,&p_ConnectionID);
                                    OK |= 1;
                                }
                                if(xnode->Peer==NULL)
                                {
                                    xnode = xnode->Parent;
                                    break;
                                }
                                else
                                {
                                    xnode = xnode->Peer;
                                }
                            }
                        }
                        if(xnode->Peer==NULL)
                        {
                            xnode = xnode->Parent;
                            break;
                        }
                        else
                        {
                            xnode = xnode->Peer;
                        }
                    }
                }
                if(xnode->Peer==NULL)
                {
                    xnode = xnode->Parent;
                    break;
                }
                else
                {
                    xnode = xnode->Peer;
                }
            }
        }
        xnode = xnode->Peer;
    }
    ILibDestructXMLNodeList(root);
    if (OK != 1)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    
    /* Type Checking */
    OK = ILibGetLong(p_ConnectionID,p_ConnectionIDLength, &TempLong);
    if(OK!=0)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    _ConnectionID = (int)TempLong;
    RTKMR_FP_ConnectionManager_ConnectionComplete((void*)ReaderObject,_ConnectionID);
}

#define RTKMR_Dispatch_ConnectionManager_GetCurrentConnectionIDs(buffer,offset,bufferLength, session)\
{\
    RTKMR_FP_ConnectionManager_GetCurrentConnectionIDs((void*)session);\
}

void RTKMR_Dispatch_ConnectionManager_GetCurrentConnectionInfo(char *buffer, int offset, int bufferLength, struct ILibWebServer_Session *ReaderObject)
{
    long TempLong;
    int OK = 0;
    char *p_ConnectionID = NULL;
    int p_ConnectionIDLength = 0;
    int _ConnectionID = 0;
    struct ILibXMLNode *xnode = ILibParseXML(buffer,offset,bufferLength);
    struct ILibXMLNode *root = xnode;
    if(ILibProcessXMLNodeList(root)!=0)
    {
        /* The XML is not well formed! */
        ILibDestructXMLNodeList(root);
        RTKMR_Response_Error(ReaderObject,501,"Invalid XML");
        return;
    }
    while(xnode!=NULL)
    {
        if(xnode->StartTag!=0 && xnode->NameLength==8 && memcmp(xnode->Name,"Envelope",8)==0)
        {
            // Envelope
            xnode = xnode->Next;
            while(xnode!=NULL)
            {
                if(xnode->StartTag!=0 && xnode->NameLength==4 && memcmp(xnode->Name,"Body",4)==0)
                {
                    // Body
                    xnode = xnode->Next;
                    while(xnode!=NULL)
                    {
                        if(xnode->StartTag!=0 && xnode->NameLength==24 && memcmp(xnode->Name,"GetCurrentConnectionInfo",24)==0)
                        {
                            // Inside the interesting part of the SOAP
                            xnode = xnode->Next;
                            while(xnode!=NULL)
                            {
                                if(xnode->NameLength==12 && memcmp(xnode->Name,"ConnectionID",12)==0)
                                {
                                    p_ConnectionIDLength = ILibReadInnerXML(xnode,&p_ConnectionID);
                                    OK |= 1;
                                }
                                if(xnode->Peer==NULL)
                                {
                                    xnode = xnode->Parent;
                                    break;
                                }
                                else
                                {
                                    xnode = xnode->Peer;
                                }
                            }
                        }
                        if(xnode->Peer==NULL)
                        {
                            xnode = xnode->Parent;
                            break;
                        }
                        else
                        {
                            xnode = xnode->Peer;
                        }
                    }
                }
                if(xnode->Peer==NULL)
                {
                    xnode = xnode->Parent;
                    break;
                }
                else
                {
                    xnode = xnode->Peer;
                }
            }
        }
        xnode = xnode->Peer;
    }
    ILibDestructXMLNodeList(root);
    if (OK != 1)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    
    /* Type Checking */
    OK = ILibGetLong(p_ConnectionID,p_ConnectionIDLength, &TempLong);
    if(OK!=0)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    _ConnectionID = (int)TempLong;
    RTKMR_FP_ConnectionManager_GetCurrentConnectionInfo((void*)ReaderObject,_ConnectionID);
}

#define RTKMR_Dispatch_ConnectionManager_GetProtocolInfo(buffer,offset,bufferLength, session)\
{\
    RTKMR_FP_ConnectionManager_GetProtocolInfo((void*)session);\
}

void RTKMR_Dispatch_ConnectionManager_PrepareForConnection(char *buffer, int offset, int bufferLength, struct ILibWebServer_Session *ReaderObject)
{
    long TempLong;
    int OK = 0;
    char *p_RemoteProtocolInfo = NULL;
    int p_RemoteProtocolInfoLength = 0;
    char* _RemoteProtocolInfo = "";
    int _RemoteProtocolInfoLength;
    char *p_PeerConnectionManager = NULL;
    int p_PeerConnectionManagerLength = 0;
    char* _PeerConnectionManager = "";
    int _PeerConnectionManagerLength;
    char *p_PeerConnectionID = NULL;
    int p_PeerConnectionIDLength = 0;
    int _PeerConnectionID = 0;
    char *p_Direction = NULL;
    int p_DirectionLength = 0;
    char* _Direction = "";
    int _DirectionLength;
    struct ILibXMLNode *xnode = ILibParseXML(buffer,offset,bufferLength);
    struct ILibXMLNode *root = xnode;
    if(ILibProcessXMLNodeList(root)!=0)
    {
        /* The XML is not well formed! */
        ILibDestructXMLNodeList(root);
        RTKMR_Response_Error(ReaderObject,501,"Invalid XML");
        return;
    }
    while(xnode!=NULL)
    {
        if(xnode->StartTag!=0 && xnode->NameLength==8 && memcmp(xnode->Name,"Envelope",8)==0)
        {
            // Envelope
            xnode = xnode->Next;
            while(xnode!=NULL)
            {
                if(xnode->StartTag!=0 && xnode->NameLength==4 && memcmp(xnode->Name,"Body",4)==0)
                {
                    // Body
                    xnode = xnode->Next;
                    while(xnode!=NULL)
                    {
                        if(xnode->StartTag!=0 && xnode->NameLength==20 && memcmp(xnode->Name,"PrepareForConnection",20)==0)
                        {
                            // Inside the interesting part of the SOAP
                            xnode = xnode->Next;
                            while(xnode!=NULL)
                            {
                                if(xnode->NameLength==18 && memcmp(xnode->Name,"RemoteProtocolInfo",18)==0)
                                {
                                    p_RemoteProtocolInfoLength = ILibReadInnerXML(xnode,&p_RemoteProtocolInfo);
                                    p_RemoteProtocolInfo[p_RemoteProtocolInfoLength]=0;
                                    OK |= 1;
                                }
                                else if(xnode->NameLength==21 && memcmp(xnode->Name,"PeerConnectionManager",21)==0)
                                {
                                    p_PeerConnectionManagerLength = ILibReadInnerXML(xnode,&p_PeerConnectionManager);
                                    p_PeerConnectionManager[p_PeerConnectionManagerLength]=0;
                                    OK |= 2;
                                }
                                else if(xnode->NameLength==16 && memcmp(xnode->Name,"PeerConnectionID",16)==0)
                                {
                                    p_PeerConnectionIDLength = ILibReadInnerXML(xnode,&p_PeerConnectionID);
                                    OK |= 4;
                                }
                                else if(xnode->NameLength==9 && memcmp(xnode->Name,"Direction",9)==0)
                                {
                                    p_DirectionLength = ILibReadInnerXML(xnode,&p_Direction);
                                    p_Direction[p_DirectionLength]=0;
                                    OK |= 8;
                                }
                                if(xnode->Peer==NULL)
                                {
                                    xnode = xnode->Parent;
                                    break;
                                }
                                else
                                {
                                    xnode = xnode->Peer;
                                }
                            }
                        }
                        if(xnode->Peer==NULL)
                        {
                            xnode = xnode->Parent;
                            break;
                        }
                        else
                        {
                            xnode = xnode->Peer;
                        }
                    }
                }
                if(xnode->Peer==NULL)
                {
                    xnode = xnode->Parent;
                    break;
                }
                else
                {
                    xnode = xnode->Peer;
                }
            }
        }
        xnode = xnode->Peer;
    }
    ILibDestructXMLNodeList(root);
    if (OK != 15)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    
    /* Type Checking */
    _RemoteProtocolInfoLength = ILibInPlaceXmlUnEscape(p_RemoteProtocolInfo);
    _RemoteProtocolInfo = p_RemoteProtocolInfo;
    _PeerConnectionManagerLength = ILibInPlaceXmlUnEscape(p_PeerConnectionManager);
    _PeerConnectionManager = p_PeerConnectionManager;
    OK = ILibGetLong(p_PeerConnectionID,p_PeerConnectionIDLength, &TempLong);
    if(OK!=0)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    _PeerConnectionID = (int)TempLong;
    _DirectionLength = ILibInPlaceXmlUnEscape(p_Direction);
    _Direction = p_Direction;
    if(memcmp(_Direction, "Input\0",6) != 0
    && memcmp(_Direction, "Output\0",7) != 0
    )
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    RTKMR_FP_ConnectionManager_PrepareForConnection((void*)ReaderObject,_RemoteProtocolInfo,_PeerConnectionManager,_PeerConnectionID,_Direction);
}

void RTKMR_Dispatch_RenderingControl_GetMute(char *buffer, int offset, int bufferLength, struct ILibWebServer_Session *ReaderObject)
{
    unsigned long TempULong;
    int OK = 0;
    char *p_InstanceID = NULL;
    int p_InstanceIDLength = 0;
    unsigned int _InstanceID = 0;
    char *p_Channel = NULL;
    int p_ChannelLength = 0;
    char* _Channel = "";
    int _ChannelLength;
    struct ILibXMLNode *xnode = ILibParseXML(buffer,offset,bufferLength);
    struct ILibXMLNode *root = xnode;
    if(ILibProcessXMLNodeList(root)!=0)
    {
        /* The XML is not well formed! */
        ILibDestructXMLNodeList(root);
        RTKMR_Response_Error(ReaderObject,501,"Invalid XML");
        return;
    }
    while(xnode!=NULL)
    {
        if(xnode->StartTag!=0 && xnode->NameLength==8 && memcmp(xnode->Name,"Envelope",8)==0)
        {
            // Envelope
            xnode = xnode->Next;
            while(xnode!=NULL)
            {
                if(xnode->StartTag!=0 && xnode->NameLength==4 && memcmp(xnode->Name,"Body",4)==0)
                {
                    // Body
                    xnode = xnode->Next;
                    while(xnode!=NULL)
                    {
                        if(xnode->StartTag!=0 && xnode->NameLength==7 && memcmp(xnode->Name,"GetMute",7)==0)
                        {
                            // Inside the interesting part of the SOAP
                            xnode = xnode->Next;
                            while(xnode!=NULL)
                            {
                                if(xnode->NameLength==10 && memcmp(xnode->Name,"InstanceID",10)==0)
                                {
                                    p_InstanceIDLength = ILibReadInnerXML(xnode,&p_InstanceID);
                                    OK |= 1;
                                }
                                else if(xnode->NameLength==7 && memcmp(xnode->Name,"Channel",7)==0)
                                {
                                    p_ChannelLength = ILibReadInnerXML(xnode,&p_Channel);
                                    p_Channel[p_ChannelLength]=0;
                                    OK |= 2;
                                }
                                if(xnode->Peer==NULL)
                                {
                                    xnode = xnode->Parent;
                                    break;
                                }
                                else
                                {
                                    xnode = xnode->Peer;
                                }
                            }
                        }
                        if(xnode->Peer==NULL)
                        {
                            xnode = xnode->Parent;
                            break;
                        }
                        else
                        {
                            xnode = xnode->Peer;
                        }
                    }
                }
                if(xnode->Peer==NULL)
                {
                    xnode = xnode->Parent;
                    break;
                }
                else
                {
                    xnode = xnode->Peer;
                }
            }
        }
        xnode = xnode->Peer;
    }
    ILibDestructXMLNodeList(root);
    if (OK != 3)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    
    /* Type Checking */
    OK = ILibGetULong(p_InstanceID,p_InstanceIDLength, &TempULong);
    if(OK!=0)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    _InstanceID = (unsigned int)TempULong;
    _ChannelLength = ILibInPlaceXmlUnEscape(p_Channel);
    _Channel = p_Channel;
    if(memcmp(_Channel, "Master\0",7) != 0
    )
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    RTKMR_FP_RenderingControl_GetMute((void*)ReaderObject,_InstanceID,_Channel);
}

void RTKMR_Dispatch_RenderingControl_GetVolume(char *buffer, int offset, int bufferLength, struct ILibWebServer_Session *ReaderObject)
{
    unsigned long TempULong;
    int OK = 0;
    char *p_InstanceID = NULL;
    int p_InstanceIDLength = 0;
    unsigned int _InstanceID = 0;
    char *p_Channel = NULL;
    int p_ChannelLength = 0;
    char* _Channel = "";
    int _ChannelLength;
    struct ILibXMLNode *xnode = ILibParseXML(buffer,offset,bufferLength);
    struct ILibXMLNode *root = xnode;
    if(ILibProcessXMLNodeList(root)!=0)
    {
        /* The XML is not well formed! */
        ILibDestructXMLNodeList(root);
        RTKMR_Response_Error(ReaderObject,501,"Invalid XML");
        return;
    }
    while(xnode!=NULL)
    {
        if(xnode->StartTag!=0 && xnode->NameLength==8 && memcmp(xnode->Name,"Envelope",8)==0)
        {
            // Envelope
            xnode = xnode->Next;
            while(xnode!=NULL)
            {
                if(xnode->StartTag!=0 && xnode->NameLength==4 && memcmp(xnode->Name,"Body",4)==0)
                {
                    // Body
                    xnode = xnode->Next;
                    while(xnode!=NULL)
                    {
                        if(xnode->StartTag!=0 && xnode->NameLength==9 && memcmp(xnode->Name,"GetVolume",9)==0)
                        {
                            // Inside the interesting part of the SOAP
                            xnode = xnode->Next;
                            while(xnode!=NULL)
                            {
                                if(xnode->NameLength==10 && memcmp(xnode->Name,"InstanceID",10)==0)
                                {
                                    p_InstanceIDLength = ILibReadInnerXML(xnode,&p_InstanceID);
                                    OK |= 1;
                                }
                                else if(xnode->NameLength==7 && memcmp(xnode->Name,"Channel",7)==0)
                                {
                                    p_ChannelLength = ILibReadInnerXML(xnode,&p_Channel);
                                    p_Channel[p_ChannelLength]=0;
                                    OK |= 2;
                                }
                                if(xnode->Peer==NULL)
                                {
                                    xnode = xnode->Parent;
                                    break;
                                }
                                else
                                {
                                    xnode = xnode->Peer;
                                }
                            }
                        }
                        if(xnode->Peer==NULL)
                        {
                            xnode = xnode->Parent;
                            break;
                        }
                        else
                        {
                            xnode = xnode->Peer;
                        }
                    }
                }
                if(xnode->Peer==NULL)
                {
                    xnode = xnode->Parent;
                    break;
                }
                else
                {
                    xnode = xnode->Peer;
                }
            }
        }
        xnode = xnode->Peer;
    }
    ILibDestructXMLNodeList(root);
    if (OK != 3)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    
    /* Type Checking */
    OK = ILibGetULong(p_InstanceID,p_InstanceIDLength, &TempULong);
    if(OK!=0)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    _InstanceID = (unsigned int)TempULong;
    _ChannelLength = ILibInPlaceXmlUnEscape(p_Channel);
    _Channel = p_Channel;
    if(memcmp(_Channel, "Master\0",7) != 0
    )
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    RTKMR_FP_RenderingControl_GetVolume((void*)ReaderObject,_InstanceID,_Channel);
}

void RTKMR_Dispatch_RenderingControl_GetVolumeDB(char *buffer, int offset, int bufferLength, struct ILibWebServer_Session *ReaderObject)
{
    unsigned long TempULong;
    int OK = 0;
    char *p_InstanceID = NULL;
    int p_InstanceIDLength = 0;
    unsigned int _InstanceID = 0;
    char *p_Channel = NULL;
    int p_ChannelLength = 0;
    char* _Channel = "";
    int _ChannelLength;
    struct ILibXMLNode *xnode = ILibParseXML(buffer,offset,bufferLength);
    struct ILibXMLNode *root = xnode;
    if(ILibProcessXMLNodeList(root)!=0)
    {
        /* The XML is not well formed! */
        ILibDestructXMLNodeList(root);
        RTKMR_Response_Error(ReaderObject,501,"Invalid XML");
        return;
    }
    while(xnode!=NULL)
    {
        if(xnode->StartTag!=0 && xnode->NameLength==8 && memcmp(xnode->Name,"Envelope",8)==0)
        {
            // Envelope
            xnode = xnode->Next;
            while(xnode!=NULL)
            {
                if(xnode->StartTag!=0 && xnode->NameLength==4 && memcmp(xnode->Name,"Body",4)==0)
                {
                    // Body
                    xnode = xnode->Next;
                    while(xnode!=NULL)
                    {
                        if(xnode->StartTag!=0 && xnode->NameLength==11 && memcmp(xnode->Name,"GetVolumeDB",11)==0)
                        {
                            // Inside the interesting part of the SOAP
                            xnode = xnode->Next;
                            while(xnode!=NULL)
                            {
                                if(xnode->NameLength==10 && memcmp(xnode->Name,"InstanceID",10)==0)
                                {
                                    p_InstanceIDLength = ILibReadInnerXML(xnode,&p_InstanceID);
                                    OK |= 1;
                                }
                                else if(xnode->NameLength==7 && memcmp(xnode->Name,"Channel",7)==0)
                                {
                                    p_ChannelLength = ILibReadInnerXML(xnode,&p_Channel);
                                    p_Channel[p_ChannelLength]=0;
                                    OK |= 2;
                                }
                                if(xnode->Peer==NULL)
                                {
                                    xnode = xnode->Parent;
                                    break;
                                }
                                else
                                {
                                    xnode = xnode->Peer;
                                }
                            }
                        }
                        if(xnode->Peer==NULL)
                        {
                            xnode = xnode->Parent;
                            break;
                        }
                        else
                        {
                            xnode = xnode->Peer;
                        }
                    }
                }
                if(xnode->Peer==NULL)
                {
                    xnode = xnode->Parent;
                    break;
                }
                else
                {
                    xnode = xnode->Peer;
                }
            }
        }
        xnode = xnode->Peer;
    }
    ILibDestructXMLNodeList(root);
    if (OK != 3)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    
    /* Type Checking */
    OK = ILibGetULong(p_InstanceID,p_InstanceIDLength, &TempULong);
    if(OK!=0)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    _InstanceID = (unsigned int)TempULong;
    _ChannelLength = ILibInPlaceXmlUnEscape(p_Channel);
    _Channel = p_Channel;
    if(memcmp(_Channel, "Master\0",7) != 0
    )
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    RTKMR_FP_RenderingControl_GetVolumeDB((void*)ReaderObject,_InstanceID,_Channel);
}

void RTKMR_Dispatch_RenderingControl_GetVolumeDBRange(char *buffer, int offset, int bufferLength, struct ILibWebServer_Session *ReaderObject)
{
    unsigned long TempULong;
    int OK = 0;
    char *p_InstanceID = NULL;
    int p_InstanceIDLength = 0;
    unsigned int _InstanceID = 0;
    char *p_Channel = NULL;
    int p_ChannelLength = 0;
    char* _Channel = "";
    int _ChannelLength;
    struct ILibXMLNode *xnode = ILibParseXML(buffer,offset,bufferLength);
    struct ILibXMLNode *root = xnode;
    if(ILibProcessXMLNodeList(root)!=0)
    {
        /* The XML is not well formed! */
        ILibDestructXMLNodeList(root);
        RTKMR_Response_Error(ReaderObject,501,"Invalid XML");
        return;
    }
    while(xnode!=NULL)
    {
        if(xnode->StartTag!=0 && xnode->NameLength==8 && memcmp(xnode->Name,"Envelope",8)==0)
        {
            // Envelope
            xnode = xnode->Next;
            while(xnode!=NULL)
            {
                if(xnode->StartTag!=0 && xnode->NameLength==4 && memcmp(xnode->Name,"Body",4)==0)
                {
                    // Body
                    xnode = xnode->Next;
                    while(xnode!=NULL)
                    {
                        if(xnode->StartTag!=0 && xnode->NameLength==16 && memcmp(xnode->Name,"GetVolumeDBRange",16)==0)
                        {
                            // Inside the interesting part of the SOAP
                            xnode = xnode->Next;
                            while(xnode!=NULL)
                            {
                                if(xnode->NameLength==10 && memcmp(xnode->Name,"InstanceID",10)==0)
                                {
                                    p_InstanceIDLength = ILibReadInnerXML(xnode,&p_InstanceID);
                                    OK |= 1;
                                }
                                else if(xnode->NameLength==7 && memcmp(xnode->Name,"Channel",7)==0)
                                {
                                    p_ChannelLength = ILibReadInnerXML(xnode,&p_Channel);
                                    p_Channel[p_ChannelLength]=0;
                                    OK |= 2;
                                }
                                if(xnode->Peer==NULL)
                                {
                                    xnode = xnode->Parent;
                                    break;
                                }
                                else
                                {
                                    xnode = xnode->Peer;
                                }
                            }
                        }
                        if(xnode->Peer==NULL)
                        {
                            xnode = xnode->Parent;
                            break;
                        }
                        else
                        {
                            xnode = xnode->Peer;
                        }
                    }
                }
                if(xnode->Peer==NULL)
                {
                    xnode = xnode->Parent;
                    break;
                }
                else
                {
                    xnode = xnode->Peer;
                }
            }
        }
        xnode = xnode->Peer;
    }
    ILibDestructXMLNodeList(root);
    if (OK != 3)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    
    /* Type Checking */
    OK = ILibGetULong(p_InstanceID,p_InstanceIDLength, &TempULong);
    if(OK!=0)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    _InstanceID = (unsigned int)TempULong;
    _ChannelLength = ILibInPlaceXmlUnEscape(p_Channel);
    _Channel = p_Channel;
    if(memcmp(_Channel, "Master\0",7) != 0
    )
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    RTKMR_FP_RenderingControl_GetVolumeDBRange((void*)ReaderObject,_InstanceID,_Channel);
}

void RTKMR_Dispatch_RenderingControl_ListPresets(char *buffer, int offset, int bufferLength, struct ILibWebServer_Session *ReaderObject)
{
    unsigned long TempULong;
    int OK = 0;
    char *p_InstanceID = NULL;
    int p_InstanceIDLength = 0;
    unsigned int _InstanceID = 0;
    struct ILibXMLNode *xnode = ILibParseXML(buffer,offset,bufferLength);
    struct ILibXMLNode *root = xnode;
    if(ILibProcessXMLNodeList(root)!=0)
    {
        /* The XML is not well formed! */
        ILibDestructXMLNodeList(root);
        RTKMR_Response_Error(ReaderObject,501,"Invalid XML");
        return;
    }
    while(xnode!=NULL)
    {
        if(xnode->StartTag!=0 && xnode->NameLength==8 && memcmp(xnode->Name,"Envelope",8)==0)
        {
            // Envelope
            xnode = xnode->Next;
            while(xnode!=NULL)
            {
                if(xnode->StartTag!=0 && xnode->NameLength==4 && memcmp(xnode->Name,"Body",4)==0)
                {
                    // Body
                    xnode = xnode->Next;
                    while(xnode!=NULL)
                    {
                        if(xnode->StartTag!=0 && xnode->NameLength==11 && memcmp(xnode->Name,"ListPresets",11)==0)
                        {
                            // Inside the interesting part of the SOAP
                            xnode = xnode->Next;
                            while(xnode!=NULL)
                            {
                                if(xnode->NameLength==10 && memcmp(xnode->Name,"InstanceID",10)==0)
                                {
                                    p_InstanceIDLength = ILibReadInnerXML(xnode,&p_InstanceID);
                                    OK |= 1;
                                }
                                if(xnode->Peer==NULL)
                                {
                                    xnode = xnode->Parent;
                                    break;
                                }
                                else
                                {
                                    xnode = xnode->Peer;
                                }
                            }
                        }
                        if(xnode->Peer==NULL)
                        {
                            xnode = xnode->Parent;
                            break;
                        }
                        else
                        {
                            xnode = xnode->Peer;
                        }
                    }
                }
                if(xnode->Peer==NULL)
                {
                    xnode = xnode->Parent;
                    break;
                }
                else
                {
                    xnode = xnode->Peer;
                }
            }
        }
        xnode = xnode->Peer;
    }
    ILibDestructXMLNodeList(root);
    if (OK != 1)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    
    /* Type Checking */
    OK = ILibGetULong(p_InstanceID,p_InstanceIDLength, &TempULong);
    if(OK!=0)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    _InstanceID = (unsigned int)TempULong;
    RTKMR_FP_RenderingControl_ListPresets((void*)ReaderObject,_InstanceID);
}

void RTKMR_Dispatch_RenderingControl_SelectPreset(char *buffer, int offset, int bufferLength, struct ILibWebServer_Session *ReaderObject)
{
    unsigned long TempULong;
    int OK = 0;
    char *p_InstanceID = NULL;
    int p_InstanceIDLength = 0;
    unsigned int _InstanceID = 0;
    char *p_PresetName = NULL;
    int p_PresetNameLength = 0;
    char* _PresetName = "";
    int _PresetNameLength;
    struct ILibXMLNode *xnode = ILibParseXML(buffer,offset,bufferLength);
    struct ILibXMLNode *root = xnode;
    if(ILibProcessXMLNodeList(root)!=0)
    {
        /* The XML is not well formed! */
        ILibDestructXMLNodeList(root);
        RTKMR_Response_Error(ReaderObject,501,"Invalid XML");
        return;
    }
    while(xnode!=NULL)
    {
        if(xnode->StartTag!=0 && xnode->NameLength==8 && memcmp(xnode->Name,"Envelope",8)==0)
        {
            // Envelope
            xnode = xnode->Next;
            while(xnode!=NULL)
            {
                if(xnode->StartTag!=0 && xnode->NameLength==4 && memcmp(xnode->Name,"Body",4)==0)
                {
                    // Body
                    xnode = xnode->Next;
                    while(xnode!=NULL)
                    {
                        if(xnode->StartTag!=0 && xnode->NameLength==12 && memcmp(xnode->Name,"SelectPreset",12)==0)
                        {
                            // Inside the interesting part of the SOAP
                            xnode = xnode->Next;
                            while(xnode!=NULL)
                            {
                                if(xnode->NameLength==10 && memcmp(xnode->Name,"InstanceID",10)==0)
                                {
                                    p_InstanceIDLength = ILibReadInnerXML(xnode,&p_InstanceID);
                                    OK |= 1;
                                }
                                else if(xnode->NameLength==10 && memcmp(xnode->Name,"PresetName",10)==0)
                                {
                                    p_PresetNameLength = ILibReadInnerXML(xnode,&p_PresetName);
                                    p_PresetName[p_PresetNameLength]=0;
                                    OK |= 2;
                                }
                                if(xnode->Peer==NULL)
                                {
                                    xnode = xnode->Parent;
                                    break;
                                }
                                else
                                {
                                    xnode = xnode->Peer;
                                }
                            }
                        }
                        if(xnode->Peer==NULL)
                        {
                            xnode = xnode->Parent;
                            break;
                        }
                        else
                        {
                            xnode = xnode->Peer;
                        }
                    }
                }
                if(xnode->Peer==NULL)
                {
                    xnode = xnode->Parent;
                    break;
                }
                else
                {
                    xnode = xnode->Peer;
                }
            }
        }
        xnode = xnode->Peer;
    }
    ILibDestructXMLNodeList(root);
    if (OK != 3)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    
    /* Type Checking */
    OK = ILibGetULong(p_InstanceID,p_InstanceIDLength, &TempULong);
    if(OK!=0)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    _InstanceID = (unsigned int)TempULong;
    _PresetNameLength = ILibInPlaceXmlUnEscape(p_PresetName);
    _PresetName = p_PresetName;
    if(memcmp(_PresetName, "FactoryDefaults\0",16) != 0
    )
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    RTKMR_FP_RenderingControl_SelectPreset((void*)ReaderObject,_InstanceID,_PresetName);
}

void RTKMR_Dispatch_RenderingControl_SetMute(char *buffer, int offset, int bufferLength, struct ILibWebServer_Session *ReaderObject)
{
    unsigned long TempULong;
    int OK = 0;
    char *p_InstanceID = NULL;
    int p_InstanceIDLength = 0;
    unsigned int _InstanceID = 0;
    char *p_Channel = NULL;
    int p_ChannelLength = 0;
    char* _Channel = "";
    int _ChannelLength;
    char *p_DesiredMute = NULL;
    int p_DesiredMuteLength = 0;
    int _DesiredMute = 0;
    struct ILibXMLNode *xnode = ILibParseXML(buffer,offset,bufferLength);
    struct ILibXMLNode *root = xnode;
    if(ILibProcessXMLNodeList(root)!=0)
    {
        /* The XML is not well formed! */
        ILibDestructXMLNodeList(root);
        RTKMR_Response_Error(ReaderObject,501,"Invalid XML");
        return;
    }
    while(xnode!=NULL)
    {
        if(xnode->StartTag!=0 && xnode->NameLength==8 && memcmp(xnode->Name,"Envelope",8)==0)
        {
            // Envelope
            xnode = xnode->Next;
            while(xnode!=NULL)
            {
                if(xnode->StartTag!=0 && xnode->NameLength==4 && memcmp(xnode->Name,"Body",4)==0)
                {
                    // Body
                    xnode = xnode->Next;
                    while(xnode!=NULL)
                    {
                        if(xnode->StartTag!=0 && xnode->NameLength==7 && memcmp(xnode->Name,"SetMute",7)==0)
                        {
                            // Inside the interesting part of the SOAP
                            xnode = xnode->Next;
                            while(xnode!=NULL)
                            {
                                if(xnode->NameLength==10 && memcmp(xnode->Name,"InstanceID",10)==0)
                                {
                                    p_InstanceIDLength = ILibReadInnerXML(xnode,&p_InstanceID);
                                    OK |= 1;
                                }
                                else if(xnode->NameLength==7 && memcmp(xnode->Name,"Channel",7)==0)
                                {
                                    p_ChannelLength = ILibReadInnerXML(xnode,&p_Channel);
                                    p_Channel[p_ChannelLength]=0;
                                    OK |= 2;
                                }
                                else if(xnode->NameLength==11 && memcmp(xnode->Name,"DesiredMute",11)==0)
                                {
                                    p_DesiredMuteLength = ILibReadInnerXML(xnode,&p_DesiredMute);
                                    OK |= 4;
                                }
                                if(xnode->Peer==NULL)
                                {
                                    xnode = xnode->Parent;
                                    break;
                                }
                                else
                                {
                                    xnode = xnode->Peer;
                                }
                            }
                        }
                        if(xnode->Peer==NULL)
                        {
                            xnode = xnode->Parent;
                            break;
                        }
                        else
                        {
                            xnode = xnode->Peer;
                        }
                    }
                }
                if(xnode->Peer==NULL)
                {
                    xnode = xnode->Parent;
                    break;
                }
                else
                {
                    xnode = xnode->Peer;
                }
            }
        }
        xnode = xnode->Peer;
    }
    ILibDestructXMLNodeList(root);
    if (OK != 7)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    
    /* Type Checking */
    OK = ILibGetULong(p_InstanceID,p_InstanceIDLength, &TempULong);
    if(OK!=0)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    _InstanceID = (unsigned int)TempULong;
    _ChannelLength = ILibInPlaceXmlUnEscape(p_Channel);
    _Channel = p_Channel;
    if(memcmp(_Channel, "Master\0",7) != 0
    )
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    OK=0;
    if(p_DesiredMuteLength==4)
    {
        if(strncasecmp(p_DesiredMute,"true",4)==0)
        {
            OK = 1;
            _DesiredMute = 1;
        }
    }
    if(p_DesiredMuteLength==5)
    {
        if(strncasecmp(p_DesiredMute,"false",5)==0)
        {
            OK = 1;
            _DesiredMute = 0;
        }
    }
    if(p_DesiredMuteLength==1)
    {
        if(memcmp(p_DesiredMute,"0",1)==0)
        {
            OK = 1;
            _DesiredMute = 0;
        }
        if(memcmp(p_DesiredMute,"1",1)==0)
        {
            OK = 1;
            _DesiredMute = 1;
        }
    }
    if(OK==0)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    RTKMR_FP_RenderingControl_SetMute((void*)ReaderObject,_InstanceID,_Channel,_DesiredMute);
}

void RTKMR_Dispatch_RenderingControl_SetVolume(char *buffer, int offset, int bufferLength, struct ILibWebServer_Session *ReaderObject)
{
    unsigned long TempULong;
    int OK = 0;
    char *p_InstanceID = NULL;
    int p_InstanceIDLength = 0;
    unsigned int _InstanceID = 0;
    char *p_Channel = NULL;
    int p_ChannelLength = 0;
    char* _Channel = "";
    int _ChannelLength;
    char *p_DesiredVolume = NULL;
    int p_DesiredVolumeLength = 0;
    unsigned short _DesiredVolume = 0;
    struct ILibXMLNode *xnode = ILibParseXML(buffer,offset,bufferLength);
    struct ILibXMLNode *root = xnode;
    if(ILibProcessXMLNodeList(root)!=0)
    {
        /* The XML is not well formed! */
        ILibDestructXMLNodeList(root);
        RTKMR_Response_Error(ReaderObject,501,"Invalid XML");
        return;
    }
    while(xnode!=NULL)
    {
        if(xnode->StartTag!=0 && xnode->NameLength==8 && memcmp(xnode->Name,"Envelope",8)==0)
        {
            // Envelope
            xnode = xnode->Next;
            while(xnode!=NULL)
            {
                if(xnode->StartTag!=0 && xnode->NameLength==4 && memcmp(xnode->Name,"Body",4)==0)
                {
                    // Body
                    xnode = xnode->Next;
                    while(xnode!=NULL)
                    {
                        if(xnode->StartTag!=0 && xnode->NameLength==9 && memcmp(xnode->Name,"SetVolume",9)==0)
                        {
                            // Inside the interesting part of the SOAP
                            xnode = xnode->Next;
                            while(xnode!=NULL)
                            {
                                if(xnode->NameLength==10 && memcmp(xnode->Name,"InstanceID",10)==0)
                                {
                                    p_InstanceIDLength = ILibReadInnerXML(xnode,&p_InstanceID);
                                    OK |= 1;
                                }
                                else if(xnode->NameLength==7 && memcmp(xnode->Name,"Channel",7)==0)
                                {
                                    p_ChannelLength = ILibReadInnerXML(xnode,&p_Channel);
                                    p_Channel[p_ChannelLength]=0;
                                    OK |= 2;
                                }
                                else if(xnode->NameLength==13 && memcmp(xnode->Name,"DesiredVolume",13)==0)
                                {
                                    p_DesiredVolumeLength = ILibReadInnerXML(xnode,&p_DesiredVolume);
                                    OK |= 4;
                                }
                                if(xnode->Peer==NULL)
                                {
                                    xnode = xnode->Parent;
                                    break;
                                }
                                else
                                {
                                    xnode = xnode->Peer;
                                }
                            }
                        }
                        if(xnode->Peer==NULL)
                        {
                            xnode = xnode->Parent;
                            break;
                        }
                        else
                        {
                            xnode = xnode->Peer;
                        }
                    }
                }
                if(xnode->Peer==NULL)
                {
                    xnode = xnode->Parent;
                    break;
                }
                else
                {
                    xnode = xnode->Peer;
                }
            }
        }
        xnode = xnode->Peer;
    }
    ILibDestructXMLNodeList(root);
    if (OK != 7)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    
    /* Type Checking */
    OK = ILibGetULong(p_InstanceID,p_InstanceIDLength, &TempULong);
    if(OK!=0)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    _InstanceID = (unsigned int)TempULong;
    _ChannelLength = ILibInPlaceXmlUnEscape(p_Channel);
    _Channel = p_Channel;
    if(memcmp(_Channel, "Master\0",7) != 0
    )
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    OK = ILibGetULong(p_DesiredVolume,p_DesiredVolumeLength, &TempULong);
    if(OK!=0)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    else
    {
        if(!(TempULong>=(unsigned long)0x0 && TempULong<=(unsigned long)0xFFFF))
        {
            RTKMR_Response_Error(ReaderObject,402,"Illegal value");
            return;
        }
        _DesiredVolume = (unsigned short)TempULong;
    }
    RTKMR_FP_RenderingControl_SetVolume((void*)ReaderObject,_InstanceID,_Channel,_DesiredVolume);
}

void RTKMR_Dispatch_RenderingControl_SetVolumeDB(char *buffer, int offset, int bufferLength, struct ILibWebServer_Session *ReaderObject)
{
    long TempLong;
    unsigned long TempULong;
    int OK = 0;
    char *p_InstanceID = NULL;
    int p_InstanceIDLength = 0;
    unsigned int _InstanceID = 0;
    char *p_Channel = NULL;
    int p_ChannelLength = 0;
    char* _Channel = "";
    int _ChannelLength;
    char *p_DesiredVolume = NULL;
    int p_DesiredVolumeLength = 0;
    short _DesiredVolume = 0;
    struct ILibXMLNode *xnode = ILibParseXML(buffer,offset,bufferLength);
    struct ILibXMLNode *root = xnode;
    if(ILibProcessXMLNodeList(root)!=0)
    {
        /* The XML is not well formed! */
        ILibDestructXMLNodeList(root);
        RTKMR_Response_Error(ReaderObject,501,"Invalid XML");
        return;
    }
    while(xnode!=NULL)
    {
        if(xnode->StartTag!=0 && xnode->NameLength==8 && memcmp(xnode->Name,"Envelope",8)==0)
        {
            // Envelope
            xnode = xnode->Next;
            while(xnode!=NULL)
            {
                if(xnode->StartTag!=0 && xnode->NameLength==4 && memcmp(xnode->Name,"Body",4)==0)
                {
                    // Body
                    xnode = xnode->Next;
                    while(xnode!=NULL)
                    {
                        if(xnode->StartTag!=0 && xnode->NameLength==11 && memcmp(xnode->Name,"SetVolumeDB",11)==0)
                        {
                            // Inside the interesting part of the SOAP
                            xnode = xnode->Next;
                            while(xnode!=NULL)
                            {
                                if(xnode->NameLength==10 && memcmp(xnode->Name,"InstanceID",10)==0)
                                {
                                    p_InstanceIDLength = ILibReadInnerXML(xnode,&p_InstanceID);
                                    OK |= 1;
                                }
                                else if(xnode->NameLength==7 && memcmp(xnode->Name,"Channel",7)==0)
                                {
                                    p_ChannelLength = ILibReadInnerXML(xnode,&p_Channel);
                                    p_Channel[p_ChannelLength]=0;
                                    OK |= 2;
                                }
                                else if(xnode->NameLength==13 && memcmp(xnode->Name,"DesiredVolume",13)==0)
                                {
                                    p_DesiredVolumeLength = ILibReadInnerXML(xnode,&p_DesiredVolume);
                                    OK |= 4;
                                }
                                if(xnode->Peer==NULL)
                                {
                                    xnode = xnode->Parent;
                                    break;
                                }
                                else
                                {
                                    xnode = xnode->Peer;
                                }
                            }
                        }
                        if(xnode->Peer==NULL)
                        {
                            xnode = xnode->Parent;
                            break;
                        }
                        else
                        {
                            xnode = xnode->Peer;
                        }
                    }
                }
                if(xnode->Peer==NULL)
                {
                    xnode = xnode->Parent;
                    break;
                }
                else
                {
                    xnode = xnode->Peer;
                }
            }
        }
        xnode = xnode->Peer;
    }
    ILibDestructXMLNodeList(root);
    if (OK != 7)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    
    /* Type Checking */
    OK = ILibGetULong(p_InstanceID,p_InstanceIDLength, &TempULong);
    if(OK!=0)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    _InstanceID = (unsigned int)TempULong;
    _ChannelLength = ILibInPlaceXmlUnEscape(p_Channel);
    _Channel = p_Channel;
    if(memcmp(_Channel, "Master\0",7) != 0
    )
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    OK = ILibGetLong(p_DesiredVolume,p_DesiredVolumeLength, &TempLong);
    if(OK!=0)
    {
        RTKMR_Response_Error(ReaderObject,402,"Illegal value");
        return;
    }
    else
    {
        if(!(TempLong>=(long)0xFFFF8000 && TempLong<=(long)0x7FFF))
        {
            RTKMR_Response_Error(ReaderObject,402,"Illegal value");
            return;
        }
        _DesiredVolume = (short)TempLong;
    }
    RTKMR_FP_RenderingControl_SetVolumeDB((void*)ReaderObject,_InstanceID,_Channel,_DesiredVolume);
}


int RTKMR_ProcessPOST(struct ILibWebServer_Session *session, struct packetheader* header, char *bodyBuffer, int offset, int bodyBufferLength)
{
    struct packetheader_field_node *f = header->FirstField;
    char* HOST;
    char* SOAPACTION = NULL;
    int SOAPACTIONLength = 0;
    struct parser_result *r,*r2;
    struct parser_result_field *prf;
    
    int RetVal = 0;
    
    //
    // Iterate through all the HTTP Headers
    //
    while(f!=NULL)
    {
        if(f->FieldLength==4 && strncasecmp(f->Field,"HOST",4)==0)
        {
            HOST = f->FieldData;
        }
        else if(f->FieldLength==10 && strncasecmp(f->Field,"SOAPACTION",10)==0)
        {
            r = ILibParseString(f->FieldData,0,f->FieldDataLength,"#",1);
            SOAPACTION = r->LastResult->data;
            SOAPACTIONLength = r->LastResult->datalength-1;
            ILibDestructParserResults(r);
        }
        else if(f->FieldLength==10 && strncasecmp(f->Field,"USER-AGENT",10)==0)
        {
            // Check UPnP version of the Control Point which invoked us
            r = ILibParseString(f->FieldData,0,f->FieldDataLength," ",1);
            prf = r->FirstResult;
            while(prf!=NULL)
            {
                if(prf->datalength>5 && memcmp(prf->data,"UPnP/",5)==0)
                {
                    r2 = ILibParseString(prf->data+5,0,prf->datalength-5,".",1);
                    r2->FirstResult->data[r2->FirstResult->datalength]=0;
                    r2->LastResult->data[r2->LastResult->datalength]=0;
                    if(atoi(r2->FirstResult->data)==1 && atoi(r2->LastResult->data)>0)
                    {
                        session->Reserved9=1;
                    }
                    ILibDestructParserResults(r2);
                }
                prf = prf->NextResult;
            }
            ILibDestructParserResults(r);
        }
        f = f->NextField;
    }
    
    if(header->DirectiveObjLength==20 && memcmp((header->DirectiveObj)+1,"AVTransport/control",19)==0)
    {
        if(SOAPACTIONLength==21 && memcmp(SOAPACTION,"GetDeviceCapabilities",21)==0)
        {
            RTKMR_Dispatch_AVTransport_GetDeviceCapabilities(bodyBuffer, offset, bodyBufferLength, session);
        }
        else if(SOAPACTIONLength==12 && memcmp(SOAPACTION,"GetMediaInfo",12)==0)
        {
            RTKMR_Dispatch_AVTransport_GetMediaInfo(bodyBuffer, offset, bodyBufferLength, session);
        }
        else if(SOAPACTIONLength==15 && memcmp(SOAPACTION,"GetPositionInfo",15)==0)
        {
            RTKMR_Dispatch_AVTransport_GetPositionInfo(bodyBuffer, offset, bodyBufferLength, session);
        }
        else if(SOAPACTIONLength==16 && memcmp(SOAPACTION,"GetTransportInfo",16)==0)
        {
            RTKMR_Dispatch_AVTransport_GetTransportInfo(bodyBuffer, offset, bodyBufferLength, session);
        }
        else if(SOAPACTIONLength==20 && memcmp(SOAPACTION,"GetTransportSettings",20)==0)
        {
            RTKMR_Dispatch_AVTransport_GetTransportSettings(bodyBuffer, offset, bodyBufferLength, session);
        }
        else if(SOAPACTIONLength==4 && memcmp(SOAPACTION,"Next",4)==0)
        {
            RTKMR_Dispatch_AVTransport_Next(bodyBuffer, offset, bodyBufferLength, session);
        }
        else if(SOAPACTIONLength==5 && memcmp(SOAPACTION,"Pause",5)==0)
        {
            RTKMR_Dispatch_AVTransport_Pause(bodyBuffer, offset, bodyBufferLength, session);
        }
        else if(SOAPACTIONLength==4 && memcmp(SOAPACTION,"Play",4)==0)
        {
            RTKMR_Dispatch_AVTransport_Play(bodyBuffer, offset, bodyBufferLength, session);
        }
        else if(SOAPACTIONLength==8 && memcmp(SOAPACTION,"Previous",8)==0)
        {
            RTKMR_Dispatch_AVTransport_Previous(bodyBuffer, offset, bodyBufferLength, session);
        }
        else if(SOAPACTIONLength==4 && memcmp(SOAPACTION,"Seek",4)==0)
        {
            RTKMR_Dispatch_AVTransport_Seek(bodyBuffer, offset, bodyBufferLength, session);
        }
        else if(SOAPACTIONLength==17 && memcmp(SOAPACTION,"SetAVTransportURI",17)==0)
        {
            RTKMR_Dispatch_AVTransport_SetAVTransportURI(bodyBuffer, offset, bodyBufferLength, session);
        }
        else if(SOAPACTIONLength==11 && memcmp(SOAPACTION,"SetPlayMode",11)==0)
        {
            RTKMR_Dispatch_AVTransport_SetPlayMode(bodyBuffer, offset, bodyBufferLength, session);
        }
        else if(SOAPACTIONLength==4 && memcmp(SOAPACTION,"Stop",4)==0)
        {
            RTKMR_Dispatch_AVTransport_Stop(bodyBuffer, offset, bodyBufferLength, session);
        }
        else
        {
            RetVal=1;
        }
    }
    else if(header->DirectiveObjLength==26 && memcmp((header->DirectiveObj)+1,"ConnectionManager/control",25)==0)
    {
        if(SOAPACTIONLength==18 && memcmp(SOAPACTION,"ConnectionComplete",18)==0)
        {
            RTKMR_Dispatch_ConnectionManager_ConnectionComplete(bodyBuffer, offset, bodyBufferLength, session);
        }
        else if(SOAPACTIONLength==23 && memcmp(SOAPACTION,"GetCurrentConnectionIDs",23)==0)
        {
            RTKMR_Dispatch_ConnectionManager_GetCurrentConnectionIDs(bodyBuffer, offset, bodyBufferLength, session);
        }
        else if(SOAPACTIONLength==24 && memcmp(SOAPACTION,"GetCurrentConnectionInfo",24)==0)
        {
            RTKMR_Dispatch_ConnectionManager_GetCurrentConnectionInfo(bodyBuffer, offset, bodyBufferLength, session);
        }
        else if(SOAPACTIONLength==15 && memcmp(SOAPACTION,"GetProtocolInfo",15)==0)
        {
            RTKMR_Dispatch_ConnectionManager_GetProtocolInfo(bodyBuffer, offset, bodyBufferLength, session);
        }
        else if(SOAPACTIONLength==20 && memcmp(SOAPACTION,"PrepareForConnection",20)==0)
        {
            RTKMR_Dispatch_ConnectionManager_PrepareForConnection(bodyBuffer, offset, bodyBufferLength, session);
        }
        else
        {
            RetVal=1;
        }
    }
    else if(header->DirectiveObjLength==25 && memcmp((header->DirectiveObj)+1,"RenderingControl/control",24)==0)
    {
        if(SOAPACTIONLength==7 && memcmp(SOAPACTION,"GetMute",7)==0)
        {
            RTKMR_Dispatch_RenderingControl_GetMute(bodyBuffer, offset, bodyBufferLength, session);
        }
        else if(SOAPACTIONLength==9 && memcmp(SOAPACTION,"GetVolume",9)==0)
        {
            RTKMR_Dispatch_RenderingControl_GetVolume(bodyBuffer, offset, bodyBufferLength, session);
        }
        else if(SOAPACTIONLength==11 && memcmp(SOAPACTION,"GetVolumeDB",11)==0)
        {
            RTKMR_Dispatch_RenderingControl_GetVolumeDB(bodyBuffer, offset, bodyBufferLength, session);
        }
        else if(SOAPACTIONLength==16 && memcmp(SOAPACTION,"GetVolumeDBRange",16)==0)
        {
            RTKMR_Dispatch_RenderingControl_GetVolumeDBRange(bodyBuffer, offset, bodyBufferLength, session);
        }
        else if(SOAPACTIONLength==11 && memcmp(SOAPACTION,"ListPresets",11)==0)
        {
            RTKMR_Dispatch_RenderingControl_ListPresets(bodyBuffer, offset, bodyBufferLength, session);
        }
        else if(SOAPACTIONLength==12 && memcmp(SOAPACTION,"SelectPreset",12)==0)
        {
            RTKMR_Dispatch_RenderingControl_SelectPreset(bodyBuffer, offset, bodyBufferLength, session);
        }
        else if(SOAPACTIONLength==7 && memcmp(SOAPACTION,"SetMute",7)==0)
        {
            RTKMR_Dispatch_RenderingControl_SetMute(bodyBuffer, offset, bodyBufferLength, session);
        }
        else if(SOAPACTIONLength==9 && memcmp(SOAPACTION,"SetVolume",9)==0)
        {
            RTKMR_Dispatch_RenderingControl_SetVolume(bodyBuffer, offset, bodyBufferLength, session);
        }
        else if(SOAPACTIONLength==11 && memcmp(SOAPACTION,"SetVolumeDB",11)==0)
        {
            RTKMR_Dispatch_RenderingControl_SetVolumeDB(bodyBuffer, offset, bodyBufferLength, session);
        }
        else
        {
            RetVal=1;
        }
    }
    else
    {
        RetVal=1;
    }
    
    
    return(RetVal);
}
struct SubscriberInfo* RTKMR_RemoveSubscriberInfo(struct SubscriberInfo **Head, int *TotalSubscribers,char* SID, int SIDLength)
{
    struct SubscriberInfo *info = *Head;
    while(info!=NULL)
    {
        if(info->SIDLength==SIDLength && memcmp(info->SID,SID,SIDLength)==0)
        {
            if ( info->Previous )
            info->Previous->Next = info->Next;
            else
            *Head = info->Next;
            if ( info->Next )
            info->Next->Previous = info->Previous;
            break;
        }
        info = info->Next;
        
    }
    if(info!=NULL)
    {
        info->Previous = NULL;
        info->Next = NULL;
        --(*TotalSubscribers);
    }
    return(info);
}

#define RTKMR_DestructSubscriberInfo(info)\
{\
    free(info->Path);\
    free(info->SID);\
    free(info);\
}

#define RTKMR_DestructEventObject(EvObject)\
{\
    free(EvObject->PacketBody);\
    free(EvObject);\
}

#define RTKMR_DestructEventDataObject(EvData)\
{\
    free(EvData);\
}
void RTKMR_ExpireSubscriberInfo(struct RTKMR_DataObject *d, struct SubscriberInfo *info)
{
    struct SubscriberInfo *t = info;
    while(t->Previous!=NULL)
    {
        t = t->Previous;
    }
    if(d->HeadSubscriberPtr_AVTransport==t)
    {
        --(d->NumberOfSubscribers_AVTransport);
    }
    else if(d->HeadSubscriberPtr_ConnectionManager==t)
    {
        --(d->NumberOfSubscribers_ConnectionManager);
    }
    else if(d->HeadSubscriberPtr_RenderingControl==t)
    {
        --(d->NumberOfSubscribers_RenderingControl);
    }
    
    
    if(info->Previous!=NULL)
    {
        // This is not the Head
        info->Previous->Next = info->Next;
        if(info->Next!=NULL)
        {
            info->Next->Previous = info->Previous;
        }
    }
    else
    {
        // This is the Head
        if(d->HeadSubscriberPtr_AVTransport==info)
        {
            d->HeadSubscriberPtr_AVTransport = info->Next;
            if(info->Next!=NULL)
            {
                info->Next->Previous = NULL;
            }
        }
        else if(d->HeadSubscriberPtr_ConnectionManager==info)
        {
            d->HeadSubscriberPtr_ConnectionManager = info->Next;
            if(info->Next!=NULL)
            {
                info->Next->Previous = NULL;
            }
        }
        else if(d->HeadSubscriberPtr_RenderingControl==info)
        {
            d->HeadSubscriberPtr_RenderingControl = info->Next;
            if(info->Next!=NULL)
            {
                info->Next->Previous = NULL;
            }
        }
        else 
        {
            // Error
            return;
        }
        
    }
    --info->RefCount;
    if(info->RefCount==0)
    {
        RTKMR_DestructSubscriberInfo(info);
    }
}

int RTKMR_SubscriptionExpired(struct SubscriberInfo *info)
{
    int RetVal = 0;
    #if defined(WIN32) || defined(_WIN32_WCE)
    if(info->RenewByTime < GetTickCount()/1000) {RetVal = -1;}
    #else
    struct timeval tv;
    gettimeofday(&tv,NULL);
    if((info->RenewByTime).tv_sec < tv.tv_sec) {RetVal = -1;}
    #endif
    return(RetVal);
}

void RTKMR_GetInitialEventBody_AVTransport(struct RTKMR_DataObject *RTKMR_Object,char ** body, int *bodylength)
{
    int TempLength;
    TempLength = (int)(25+(int)strlen(RTKMR_Object->AVTransport_LastChange));
    *body = (char*)malloc(sizeof(char)*TempLength);
    *bodylength = sprintf(*body,"LastChange>%s</LastChange",RTKMR_Object->AVTransport_LastChange);
}
void RTKMR_GetInitialEventBody_ConnectionManager(struct RTKMR_DataObject *RTKMR_Object,char ** body, int *bodylength)
{
    int TempLength;
    TempLength = (int)(177+(int)strlen(RTKMR_Object->ConnectionManager_SourceProtocolInfo)+(int)strlen(RTKMR_Object->ConnectionManager_SinkProtocolInfo)+(int)strlen(RTKMR_Object->ConnectionManager_CurrentConnectionIDs));
    *body = (char*)malloc(sizeof(char)*TempLength);
    *bodylength = sprintf(*body,"SourceProtocolInfo>%s</SourceProtocolInfo></e:property><e:property><SinkProtocolInfo>%s</SinkProtocolInfo></e:property><e:property><CurrentConnectionIDs>%s</CurrentConnectionIDs",RTKMR_Object->ConnectionManager_SourceProtocolInfo,RTKMR_Object->ConnectionManager_SinkProtocolInfo,RTKMR_Object->ConnectionManager_CurrentConnectionIDs);
}
void RTKMR_GetInitialEventBody_RenderingControl(struct RTKMR_DataObject *RTKMR_Object,char ** body, int *bodylength)
{
    int TempLength;
    TempLength = (int)(25+(int)strlen(RTKMR_Object->RenderingControl_LastChange));
    *body = (char*)malloc(sizeof(char)*TempLength);
    *bodylength = sprintf(*body,"LastChange>%s</LastChange",RTKMR_Object->RenderingControl_LastChange);
}


void RTKMR_ProcessUNSUBSCRIBE(struct packetheader *header, struct ILibWebServer_Session *session)
{
    char* SID = NULL;
    int SIDLength = 0;
    struct SubscriberInfo *Info;
    struct packetheader_field_node *f;
    char* packet = (char*)malloc(sizeof(char)*50);
    int packetlength;
    
    //
    // Iterate through all the HTTP headers
    //
    f = header->FirstField;
    while(f!=NULL)
    {
        if(f->FieldLength==3)
        {
            if(strncasecmp(f->Field,"SID",3)==0)
            {
                //
                // Get the Subscription ID
                //
                SID = f->FieldData;
                SIDLength = f->FieldDataLength;
            }
        }
        f = f->NextField;
    }
    sem_wait(&(((struct RTKMR_DataObject*)session->User)->EventLock));
    if(header->DirectiveObjLength==18 && memcmp(header->DirectiveObj + 1,"AVTransport/event",17)==0)
    {
        Info = RTKMR_RemoveSubscriberInfo(&(((struct RTKMR_DataObject*)session->User)->HeadSubscriberPtr_AVTransport),&(((struct RTKMR_DataObject*)session->User)->NumberOfSubscribers_AVTransport),SID,SIDLength);
        if(Info!=NULL)
        {
            --Info->RefCount;
            if(Info->RefCount==0)
            {
                RTKMR_DestructSubscriberInfo(Info);
            }
            packetlength = sprintf(packet,"HTTP/1.1 %d %s\r\nContent-Length: 0\r\n\r\n",200,"OK");
            ILibWebServer_Send_Raw(session,packet,packetlength,0,1);
        }
        else
        {
            packetlength = sprintf(packet,"HTTP/1.1 %d %s\r\nContent-Length: 0\r\n\r\n",412,"Invalid SID");
            ILibWebServer_Send_Raw(session,packet,packetlength,0,1);
        }
    }
    else if(header->DirectiveObjLength==24 && memcmp(header->DirectiveObj + 1,"ConnectionManager/event",23)==0)
    {
        Info = RTKMR_RemoveSubscriberInfo(&(((struct RTKMR_DataObject*)session->User)->HeadSubscriberPtr_ConnectionManager),&(((struct RTKMR_DataObject*)session->User)->NumberOfSubscribers_ConnectionManager),SID,SIDLength);
        if(Info!=NULL)
        {
            --Info->RefCount;
            if(Info->RefCount==0)
            {
                RTKMR_DestructSubscriberInfo(Info);
            }
            packetlength = sprintf(packet,"HTTP/1.1 %d %s\r\nContent-Length: 0\r\n\r\n",200,"OK");
            ILibWebServer_Send_Raw(session,packet,packetlength,0,1);
        }
        else
        {
            packetlength = sprintf(packet,"HTTP/1.1 %d %s\r\nContent-Length: 0\r\n\r\n",412,"Invalid SID");
            ILibWebServer_Send_Raw(session,packet,packetlength,0,1);
        }
    }
    else if(header->DirectiveObjLength==23 && memcmp(header->DirectiveObj + 1,"RenderingControl/event",22)==0)
    {
        Info = RTKMR_RemoveSubscriberInfo(&(((struct RTKMR_DataObject*)session->User)->HeadSubscriberPtr_RenderingControl),&(((struct RTKMR_DataObject*)session->User)->NumberOfSubscribers_RenderingControl),SID,SIDLength);
        if(Info!=NULL)
        {
            --Info->RefCount;
            if(Info->RefCount==0)
            {
                RTKMR_DestructSubscriberInfo(Info);
            }
            packetlength = sprintf(packet,"HTTP/1.1 %d %s\r\nContent-Length: 0\r\n\r\n",200,"OK");
            ILibWebServer_Send_Raw(session,packet,packetlength,0,1);
        }
        else
        {
            packetlength = sprintf(packet,"HTTP/1.1 %d %s\r\nContent-Length: 0\r\n\r\n",412,"Invalid SID");
            ILibWebServer_Send_Raw(session,packet,packetlength,0,1);
        }
    }
    
    sem_post(&(((struct RTKMR_DataObject*)session->User)->EventLock));
}
void RTKMR_TryToSubscribe(char* ServiceName, long Timeout, char* URL, int URLLength,struct ILibWebServer_Session *session)
{
    int *TotalSubscribers = NULL;
    struct SubscriberInfo **HeadPtr = NULL;
    struct SubscriberInfo *NewSubscriber,*TempSubscriber;
    int SIDNumber,rnumber;
    char *SID;
    char *TempString;
    int TempStringLength;
    char *TempString2;
    long TempLong;
    char *packet;
    int packetlength;
    char* path;
    
    char* escapedURI;
    int escapedURILength;
    
    char *packetbody = NULL;
    int packetbodyLength;
    
    struct parser_result *p;
    struct parser_result *p2;
    
    struct RTKMR_DataObject *dataObject = (struct RTKMR_DataObject*)session->User;
    
    if(strncmp(ServiceName,"AVTransport",11)==0)
    {
        TotalSubscribers = &(dataObject->NumberOfSubscribers_AVTransport);
        HeadPtr = &(dataObject->HeadSubscriberPtr_AVTransport);
    }
    if(strncmp(ServiceName,"ConnectionManager",17)==0)
    {
        TotalSubscribers = &(dataObject->NumberOfSubscribers_ConnectionManager);
        HeadPtr = &(dataObject->HeadSubscriberPtr_ConnectionManager);
    }
    if(strncmp(ServiceName,"RenderingControl",16)==0)
    {
        TotalSubscribers = &(dataObject->NumberOfSubscribers_RenderingControl);
        HeadPtr = &(dataObject->HeadSubscriberPtr_RenderingControl);
    }
    
    
    if(*HeadPtr!=NULL)
    {
        NewSubscriber = *HeadPtr;
        while(NewSubscriber!=NULL)
        {
            if(RTKMR_SubscriptionExpired(NewSubscriber)!=0)
            {
                TempSubscriber = NewSubscriber->Next;
                NewSubscriber = RTKMR_RemoveSubscriberInfo(HeadPtr,TotalSubscribers,NewSubscriber->SID,NewSubscriber->SIDLength);
                RTKMR_DestructSubscriberInfo(NewSubscriber);
                NewSubscriber = TempSubscriber;
            }
            else
            {
                NewSubscriber = NewSubscriber->Next;
            }
        }
    }
    //
    // The Maximum number of subscribers can be bounded
    //
    if(*TotalSubscribers<10)
    {
        NewSubscriber = (struct SubscriberInfo*)malloc(sizeof(struct SubscriberInfo));
        memset(NewSubscriber,0,sizeof(struct SubscriberInfo));
        
        
        //
        // The SID must be globally unique, so lets generate it using
        // a bunch of random hex characters
        //
        SID = (char*)malloc(43);
        memset(SID,0,38);
        sprintf(SID,"uuid:");
        for(SIDNumber=5;SIDNumber<=12;++SIDNumber)
        {
            rnumber = rand()%16;
            sprintf(SID+SIDNumber,"%x",rnumber);
        }
        sprintf(SID+SIDNumber,"-");
        for(SIDNumber=14;SIDNumber<=17;++SIDNumber)
        {
            rnumber = rand()%16;
            sprintf(SID+SIDNumber,"%x",rnumber);
        }
        sprintf(SID+SIDNumber,"-");
        for(SIDNumber=19;SIDNumber<=22;++SIDNumber)
        {
            rnumber = rand()%16;
            sprintf(SID+SIDNumber,"%x",rnumber);
        }
        sprintf(SID+SIDNumber,"-");
        for(SIDNumber=24;SIDNumber<=27;++SIDNumber)
        {
            rnumber = rand()%16;
            sprintf(SID+SIDNumber,"%x",rnumber);
        }
        sprintf(SID+SIDNumber,"-");
        for(SIDNumber=29;SIDNumber<=40;++SIDNumber)
        {
            rnumber = rand()%16;
            sprintf(SID+SIDNumber,"%x",rnumber);
        }
        
        p = ILibParseString(URL,0,URLLength,"://",3);
        if(p->NumResults==1)
        {
            ILibWebServer_Send_Raw(session,"HTTP/1.1 412 Precondition Failed\r\nContent-Length: 0\r\n\r\n",55,1,1);
            ILibDestructParserResults(p);
            return;
        }
        TempString = p->LastResult->data;
        TempStringLength = p->LastResult->datalength;
        ILibDestructParserResults(p);
        p = ILibParseString(TempString,0,TempStringLength,"/",1);
        p2 = ILibParseString(p->FirstResult->data,0,p->FirstResult->datalength,":",1);
        TempString2 = (char*)malloc(1+sizeof(char)*p2->FirstResult->datalength);
        memcpy(TempString2,p2->FirstResult->data,p2->FirstResult->datalength);
        TempString2[p2->FirstResult->datalength] = '\0';
        NewSubscriber->Address = inet_addr(TempString2);
        if(p2->NumResults==1)
        {
            NewSubscriber->Port = 80;
            path = (char*)malloc(1+TempStringLength - p2->FirstResult->datalength -1);
            memcpy(path,TempString + p2->FirstResult->datalength,TempStringLength - p2->FirstResult->datalength -1);
            path[TempStringLength - p2->FirstResult->datalength - 1] = '\0';
            NewSubscriber->Path = path;
            NewSubscriber->PathLength = (int)strlen(path);
        }
        else
        {
            ILibGetLong(p2->LastResult->data,p2->LastResult->datalength,&TempLong);
            NewSubscriber->Port = (unsigned short)TempLong;
            if(TempStringLength==p->FirstResult->datalength)
            {
                path = (char*)malloc(2);
                memcpy(path,"/",1);
                path[1] = '\0';
            }
            else
            {
                path = (char*)malloc(1+TempStringLength - p->FirstResult->datalength -1);
                memcpy(path,TempString + p->FirstResult->datalength,TempStringLength - p->FirstResult->datalength -1);
                path[TempStringLength - p->FirstResult->datalength -1] = '\0';
            }
            NewSubscriber->Path = path;
            NewSubscriber->PathLength = (int)strlen(path);
        }
        ILibDestructParserResults(p);
        ILibDestructParserResults(p2);
        free(TempString2);
        
        
        escapedURI = (char*)malloc(ILibHTTPEscapeLength(NewSubscriber->Path));
        escapedURILength = ILibHTTPEscape(escapedURI,NewSubscriber->Path);
        
        free(NewSubscriber->Path);
        NewSubscriber->Path = escapedURI;
        NewSubscriber->PathLength = escapedURILength;
        
        
        NewSubscriber->RefCount = 1;
        NewSubscriber->Disposing = 0;
        NewSubscriber->Previous = NULL;
        NewSubscriber->SID = SID;
        NewSubscriber->SIDLength = (int)strlen(SID);
        NewSubscriber->SEQ = 0;
        
        //
        // Determine what the subscription renewal cycle is
        //
        #if defined(WIN32) || defined(_WIN32_WCE)
        NewSubscriber->RenewByTime = (GetTickCount() / 1000) + Timeout;
        #else
        gettimeofday(&(NewSubscriber->RenewByTime),NULL);
        (NewSubscriber->RenewByTime).tv_sec += (int)Timeout;
        #endif
        NewSubscriber->Next = *HeadPtr;
        if(*HeadPtr!=NULL) {(*HeadPtr)->Previous = NewSubscriber;}
        *HeadPtr = NewSubscriber;
        ++(*TotalSubscribers);
        LVL3DEBUG(printf("\r\n\r\nSubscribed [%s] %d.%d.%d.%d:%d FOR %d Duration\r\n",NewSubscriber->SID,(NewSubscriber->Address)&0xFF,(NewSubscriber->Address>>8)&0xFF,(NewSubscriber->Address>>16)&0xFF,(NewSubscriber->Address>>24)&0xFF,NewSubscriber->Port,Timeout);)
        #if defined(WIN32) || defined(_WIN32_WCE)	
        LVL3DEBUG(printf("TIMESTAMP: %d <%d>\r\n\r\n",(NewSubscriber->RenewByTime)-Timeout,NewSubscriber);)
        #else
        LVL3DEBUG(printf("TIMESTAMP: %d <%d>\r\n\r\n",(NewSubscriber->RenewByTime).tv_sec-Timeout,NewSubscriber);)
        #endif
        packet = (char*)malloc(134 + (int)strlen(SID) + (int)strlen(RTKMR_PLATFORM) + 4);
        packetlength = sprintf(packet,"HTTP/1.1 200 OK\r\nSERVER: %s, UPnP/1.0, Intel MicroStack/1.0.1868\r\nSID: %s\r\nTIMEOUT: Second-%ld\r\nContent-Length: 0\r\n\r\n",RTKMR_PLATFORM,SID,Timeout);
        if(strcmp(ServiceName,"AVTransport")==0)
        {
            RTKMR_GetInitialEventBody_AVTransport(dataObject,&packetbody,&packetbodyLength);
        }
        else if(strcmp(ServiceName,"ConnectionManager")==0)
        {
            RTKMR_GetInitialEventBody_ConnectionManager(dataObject,&packetbody,&packetbodyLength);
        }
        else if(strcmp(ServiceName,"RenderingControl")==0)
        {
            RTKMR_GetInitialEventBody_RenderingControl(dataObject,&packetbody,&packetbodyLength);
        }
        
        if (packetbody != NULL)	    {
            ILibWebServer_Send_Raw(session,packet,packetlength,0,1);
            
            RTKMR_SendEvent_Body(dataObject,packetbody,packetbodyLength,NewSubscriber);
            free(packetbody);
        } 
    }
    else
    {
        /* Too many subscribers */
        ILibWebServer_Send_Raw(session,"HTTP/1.1 412 Too Many Subscribers\r\nContent-Length: 0\r\n\r\n",56,1,1);
    }
}
void RTKMR_SubscribeEvents(char* path,int pathlength,char* Timeout,int TimeoutLength,char* URL,int URLLength,struct ILibWebServer_Session* session)
{
    long TimeoutVal;
    char* buffer = (char*)malloc(1+sizeof(char)*pathlength);
    
    ILibGetLong(Timeout,TimeoutLength,&TimeoutVal);
    memcpy(buffer,path,pathlength);
    buffer[pathlength] = '\0';
    free(buffer);
    if(TimeoutVal>7200) {TimeoutVal=7200;}
    
    if(pathlength==18 && memcmp(path+1,"AVTransport/event",17)==0)
    {
        RTKMR_TryToSubscribe("AVTransport",TimeoutVal,URL,URLLength,session);
    }
    else if(pathlength==24 && memcmp(path+1,"ConnectionManager/event",23)==0)
    {
        RTKMR_TryToSubscribe("ConnectionManager",TimeoutVal,URL,URLLength,session);
    }
    else if(pathlength==23 && memcmp(path+1,"RenderingControl/event",22)==0)
    {
        RTKMR_TryToSubscribe("RenderingControl",TimeoutVal,URL,URLLength,session);
    }
    else
    {
        ILibWebServer_Send_Raw(session,"HTTP/1.1 412 Invalid Service Name\r\nContent-Length: 0\r\n\r\n",56,1,1);
    }
    
}
void RTKMR_RenewEvents(char* path,int pathlength,char *_SID,int SIDLength, char* Timeout, int TimeoutLength, struct ILibWebServer_Session *ReaderObject)
{
    struct SubscriberInfo *info = NULL;
    long TimeoutVal;
    #if !defined(WIN32) && !defined(_WIN32_WCE)
    struct timeval tv;
    #endif
    char* packet;
    int packetlength;
    char* SID = (char*)malloc(SIDLength+1);
    memcpy(SID,_SID,SIDLength);
    SID[SIDLength] ='\0';
    #if defined(WIN32) || defined(_WIN32_WCE)
    LVL3DEBUG(printf("\r\n\r\nTIMESTAMP: %d\r\n",GetTickCount()/1000);)
    #else
    LVL3DEBUG(gettimeofday(&tv,NULL);)
    LVL3DEBUG(printf("\r\n\r\nTIMESTAMP: %d\r\n",tv.tv_sec);)
    #endif
    LVL3DEBUG(printf("SUBSCRIBER [%s] attempting to Renew Events for %s Duration [",SID,Timeout);)
    
    if(pathlength==18 && memcmp(path+1,"AVTransport/event",17)==0)
    {
        info = ((struct RTKMR_DataObject*)ReaderObject->User)->HeadSubscriberPtr_AVTransport;
    }
    else if(pathlength==24 && memcmp(path+1,"ConnectionManager/event",23)==0)
    {
        info = ((struct RTKMR_DataObject*)ReaderObject->User)->HeadSubscriberPtr_ConnectionManager;
    }
    else if(pathlength==23 && memcmp(path+1,"RenderingControl/event",22)==0)
    {
        info = ((struct RTKMR_DataObject*)ReaderObject->User)->HeadSubscriberPtr_RenderingControl;
    }
    
    
    //
    // Find this SID in the subscriber list, and recalculate
    // the expiration timeout
    //
    while(info!=NULL && strcmp(info->SID,SID)!=0)
    {
        info = info->Next;
    }
    if(info!=NULL)
    {
        ILibGetLong(Timeout,TimeoutLength,&TimeoutVal);
        #if defined(WIN32) || defined(_WIN32_WCE)
        info->RenewByTime = TimeoutVal + (GetTickCount() / 1000);
        #else
        gettimeofday(&tv,NULL);
        (info->RenewByTime).tv_sec = tv.tv_sec + TimeoutVal;
        #endif
        packet = (char*)malloc(134 + (int)strlen(SID) + 4);
        packetlength = sprintf(packet,"HTTP/1.1 200 OK\r\nSERVER: %s, UPnP/1.0, Intel MicroStack/1.0.1868\r\nSID: %s\r\nTIMEOUT: Second-%ld\r\nContent-Length: 0\r\n\r\n",RTKMR_PLATFORM,SID,TimeoutVal);
        ILibWebServer_Send_Raw(ReaderObject,packet,packetlength,0,1);
        LVL3DEBUG(printf("OK] {%d} <%d>\r\n\r\n",TimeoutVal,info);)
    }
    else
    {
        LVL3DEBUG(printf("FAILED]\r\n\r\n");)
        ILibWebServer_Send_Raw(ReaderObject,"HTTP/1.1 412 Precondition Failed\r\nContent-Length: 0\r\n\r\n",55,1,1);
    }
    free(SID);
}
void RTKMR_ProcessSUBSCRIBE(struct packetheader *header, struct ILibWebServer_Session *session)
{
    char* SID = NULL;
    int SIDLength = 0;
    char* Timeout = NULL;
    int TimeoutLength = 0;
    char* URL = NULL;
    int URLLength = 0;
    struct parser_result *p;
    
    struct packetheader_field_node *f;
    
    //
    // Iterate through all the HTTP Headers
    //
    f = header->FirstField;
    while(f!=NULL)
    {
        if(f->FieldLength==3 && strncasecmp(f->Field,"SID",3)==0)
        {
            //
            // Get the Subscription ID
            //
            SID = f->FieldData;
            SIDLength = f->FieldDataLength;
        }
        else if(f->FieldLength==8 && strncasecmp(f->Field,"Callback",8)==0)
        {
            //
            // Get the Callback URL
            //
            URL = f->FieldData;
            URLLength = f->FieldDataLength;
        }
        else if(f->FieldLength==7 && strncasecmp(f->Field,"Timeout",7)==0)
        {
            //
            // Get the requested timeout value
            //
            Timeout = f->FieldData;
            TimeoutLength = f->FieldDataLength;
        }
        
        f = f->NextField;
    }
    if(Timeout==NULL)
    {
        //
        // It a timeout wasn't specified, force it to a specific value
        //
        Timeout = "7200";
        TimeoutLength = 4;
    }
    else
    {
        p = ILibParseString(Timeout,0,TimeoutLength,"-",1);
        if(p->NumResults==2)
        {
            Timeout = p->LastResult->data;
            TimeoutLength = p->LastResult->datalength;
            if(TimeoutLength==8 && strncasecmp(Timeout,"INFINITE",8)==0)
            {
                //
                // Infinite timeouts will cause problems, so we don't allow it
                //
                Timeout = "7200";
                TimeoutLength = 4;
            }
        }
        else
        {
            Timeout = "7200";
            TimeoutLength = 4;
        }
        ILibDestructParserResults(p);
    }
    if(SID==NULL)
    {
        //
        // If not SID was specified, this is a subscription request
        //
        
        /* Subscribe */
        RTKMR_SubscribeEvents(header->DirectiveObj,header->DirectiveObjLength,Timeout,TimeoutLength,URL,URLLength,session);
    }
    else
    {
        //
        // If a SID was specified, it is a renewal request for an existing subscription
        //
        
        /* Renew */
        RTKMR_RenewEvents(header->DirectiveObj,header->DirectiveObjLength,SID,SIDLength,Timeout,TimeoutLength,session);
    }
}
void RTKMR_ProcessHTTPPacket(struct ILibWebServer_Session *session, struct packetheader* header, char *bodyBuffer, int offset, int bodyBufferLength)

{
    struct RTKMR_DataObject *dataObject = (struct RTKMR_DataObject*)session->User;
    #if defined(WIN32) || defined(_WIN32_WCE)
    char *responseHeader = "\r\nCONTENT-TYPE:  text/xml; charset=\"utf-8\"\r\nServer: WINDOWS, UPnP/1.0, Intel MicroStack/1.0.1868";
    #else
    char *responseHeader = "\r\nCONTENT-TYPE:  text/xml; charset=\"utf-8\"\r\nServer: POSIX, UPnP/1.0, Intel MicroStack/1.0.1868";
    #endif
    char *errorTemplate = "HTTP/1.1 %d %s\r\nServer: %s, UPnP/1.0, Intel MicroStack/1.0.1868\r\nContent-Length: 0\r\n\r\n";
    char *errorPacket;
    int errorPacketLength;
    char *buffer;
    
    LVL3DEBUG(errorPacketLength=ILibGetRawPacket(header,&errorPacket);)
    LVL3DEBUG(printf("%s\r\n",errorPacket);)
    LVL3DEBUG(free(errorPacket);)			
    
    
    if(header->DirectiveLength==4 && memcmp(header->Directive,"HEAD",4)==0)
    {
        if(header->DirectiveObjLength==1 && memcmp(header->DirectiveObj,"/",1)==0)
        {
            //
            // A HEAD request for the device description document.
            // We stream the document back, so we don't return content length or anything
            // because the actual response won't have it either
            //
            ILibWebServer_StreamHeader_Raw(session,200,"OK",responseHeader,1);
            ILibWebServer_StreamBody(session,NULL,0,ILibAsyncSocket_MemoryOwnership_STATIC,1);
        }
#if 0
        else if(header->DirectiveObjLength==21 && memcmp((header->DirectiveObj)+1,"AVTransport/scpd.xml",20)==0)
        {
            ILibWebServer_StreamHeader_Raw(session,200,"OK",responseHeader,1);
            ILibWebServer_StreamBody(session,NULL,0,ILibAsyncSocket_MemoryOwnership_STATIC,1);
        }
        else if(header->DirectiveObjLength==27 && memcmp((header->DirectiveObj)+1,"ConnectionManager/scpd.xml",26)==0)
        {
            ILibWebServer_StreamHeader_Raw(session,200,"OK",responseHeader,1);
            ILibWebServer_StreamBody(session,NULL,0,ILibAsyncSocket_MemoryOwnership_STATIC,1);
        }
        else if(header->DirectiveObjLength==26 && memcmp((header->DirectiveObj)+1,"RenderingControl/scpd.xml",25)==0)
        {
            ILibWebServer_StreamHeader_Raw(session,200,"OK",responseHeader,1);
            ILibWebServer_StreamBody(session,NULL,0,ILibAsyncSocket_MemoryOwnership_STATIC,1);
        }
#endif
        
        else
        {
            //
            // A HEAD request for something we don't have
            //
            errorPacket = (char*)malloc(128);
            errorPacketLength = sprintf(errorPacket,errorTemplate,404,"File Not Found",RTKMR_PLATFORM);
            ILibWebServer_Send_Raw(session,errorPacket,errorPacketLength,0,1);
        }
    }
    else if(header->DirectiveLength==3 && memcmp(header->Directive,"GET",3)==0)
    {
        if(header->DirectiveObjLength==1 && memcmp(header->DirectiveObj,"/",1)==0)
        {
            //
            // A GET Request for the device description document, so lets stream
            // it back to the client
            //
            ILibWebServer_StreamHeader_Raw(session,200,"OK",responseHeader,1);
            ILibWebServer_StreamBody(session,dataObject->DeviceDescription,dataObject->DeviceDescriptionLength,1,1);
        }
#if 0
        else if(header->DirectiveObjLength==21 && memcmp((header->DirectiveObj)+1,"AVTransport/scpd.xml",20)==0)
        {
            buffer = ILibDecompressString((char*)RTKMR_AVTransportDescription,RTKMR_AVTransportDescriptionLength,RTKMR_AVTransportDescriptionLengthUX);
            ILibWebServer_Send_Raw(session,buffer,RTKMR_AVTransportDescriptionLengthUX,0,1);
        }
        else if(header->DirectiveObjLength==27 && memcmp((header->DirectiveObj)+1,"ConnectionManager/scpd.xml",26)==0)
        {
            buffer = ILibDecompressString((char*)RTKMR_ConnectionManagerDescription,RTKMR_ConnectionManagerDescriptionLength,RTKMR_ConnectionManagerDescriptionLengthUX);
            ILibWebServer_Send_Raw(session,buffer,RTKMR_ConnectionManagerDescriptionLengthUX,0,1);
        }
        else if(header->DirectiveObjLength==26 && memcmp((header->DirectiveObj)+1,"RenderingControl/scpd.xml",25)==0)
        {
            buffer = ILibDecompressString((char*)RTKMR_RenderingControlDescription,RTKMR_RenderingControlDescriptionLength,RTKMR_RenderingControlDescriptionLengthUX);
            ILibWebServer_Send_Raw(session,buffer,RTKMR_RenderingControlDescriptionLengthUX,0,1);
        }
#endif
        
        else
        {
            //
            // A GET Request for something we don't have
            //
            errorPacket = (char*)malloc(128);
            errorPacketLength = sprintf(errorPacket,errorTemplate,404,"File Not Found",RTKMR_PLATFORM);
            ILibWebServer_Send_Raw(session,errorPacket,errorPacketLength,0,1);
        }
    }
    else if(header->DirectiveLength==4 && memcmp(header->Directive,"POST",4)==0)
    {
        //
        // Defer Control to the POST Handler
        //
        if(RTKMR_ProcessPOST(session,header,bodyBuffer,offset,bodyBufferLength)!=0)
        {
            //
            // A POST for an action that doesn't exist
            //
            RTKMR_Response_Error(session,401,"Invalid Action");
        }
    }
    else if(header->DirectiveLength==9 && memcmp(header->Directive,"SUBSCRIBE",9)==0)
    {
        //
        // Subscription Handler
        //
        RTKMR_ProcessSUBSCRIBE(header,session);
    }
    else if(header->DirectiveLength==11 && memcmp(header->Directive,"UNSUBSCRIBE",11)==0)
    {
        //
        // UnSubscribe Handler
        //
        RTKMR_ProcessUNSUBSCRIBE(header,session);
    }
    else
    {
        //
        // The client tried something we didn't expect/support
        //
        errorPacket = (char*)malloc(128);
        errorPacketLength = sprintf(errorPacket,errorTemplate,400,"Bad Request",RTKMR_PLATFORM);
        ILibWebServer_Send_Raw(session,errorPacket,errorPacketLength,1,1);
    }
}
void RTKMR_FragmentedSendNotify_Destroy(void *data);
void RTKMR_MasterPreSelect(void* object,fd_set *socketset, fd_set *writeset, fd_set *errorset, int* blocktime)
{
    int i;
    struct RTKMR_DataObject *RTKMR_Object = (struct RTKMR_DataObject*)object;
    
    int ra = 1;
    struct sockaddr_in addr;
    struct ip_mreq mreq;
    unsigned char TTL = 4;
    struct RTKMR_FragmentNotifyStruct *f;
    int timeout;
    
    if(RTKMR_Object->InitialNotify==0)
    {
        //
        // The initial "HELLO" packets were not sent yet, so lets send them
        //
        RTKMR_Object->InitialNotify = -1;
        //
        // In case we were interrupted, we need to flush out the caches of
        // all the control points by sending a "byebye" first, to insure
        // control points don't ignore our "hello" packets thinking they are just
        // periodic re-advertisements.
        //
        RTKMR_SendByeBye(RTKMR_Object);
        
        //
        // Iterate through all the packet types
        //
        for(i=1;i<=6;++i)
        {
            f = (struct RTKMR_FragmentNotifyStruct*)malloc(sizeof(struct RTKMR_FragmentNotifyStruct));
            f->packetNumber=i;
            f->upnp = RTKMR_Object;
            //
            // We need to inject some delay in these packets to space them out,
            // otherwise we could overflow the inbound buffer of the recipient, causing them
            // to lose packets. And UPnP/1.0 control points are not as robust as UPnP/1.1 control points,
            // so they need all the help they can get ;)
            //
            timeout = (int)(0 + ((unsigned short)rand() % (500)));
            do
            {
                f->upnp->InitialNotify = rand();
            }while(f->upnp->InitialNotify==0);
            //
            // Register for the timed callback, to actually send the packet
            //
            ILibLifeTime_AddEx(f->upnp->WebServerTimer,f,timeout,&RTKMR_FragmentedSendNotify,&RTKMR_FragmentedSendNotify_Destroy);
        }
    }
    if(RTKMR_Object->UpdateFlag!=0)
    {
        //
        // Somebody told us that we should recheck our IP Address table,
        // as one of them may have changed
        //
        RTKMR_Object->UpdateFlag = 0;
        
        /* Clear Sockets */
        
        
        //
        // Iterate through all the currently bound IP addresses
        // and release the sockets
        //
        for(i=0;i<RTKMR_Object->AddressListLength;++i)
        {
            #if defined(WIN32) || defined(_WIN32_WCE)
            closesocket(RTKMR_Object->NOTIFY_SEND_socks[i]);
            #else
#if 1 ////def OLD_SOCKET
            close(RTKMR_Object->NOTIFY_SEND_socks[i]);
#endif 
            #endif
        }
        free(RTKMR_Object->NOTIFY_SEND_socks);
        
        /* Set up socket */
        free(RTKMR_Object->AddressList);
        //
        // Fetch a current list of ip addresses
        //
        RTKMR_Object->AddressListLength = ILibGetLocalIPAddressList(&(RTKMR_Object->AddressList));
        #if defined(WIN32) || defined(_WIN32_WCE)	
        RTKMR_Object->NOTIFY_SEND_socks = (SOCKET*)malloc(sizeof(int)*(RTKMR_Object->AddressListLength));
        #else
        RTKMR_Object->NOTIFY_SEND_socks = (int*)malloc(sizeof(int)*(RTKMR_Object->AddressListLength));
        #endif
        
        //
        // Now that we have a new list of IP addresses, re-initialise everything
        //
        for(i=0;i<RTKMR_Object->AddressListLength;++i)
        {
#if 1 ////def OLD_SOCKET
            RTKMR_Object->NOTIFY_SEND_socks[i] = socket(AF_INET, SOCK_DGRAM, 0);
#else
            RTKMR_Object->NOTIFY_SEND_socks = RTKMR_Object->NOTIFY_RECEIVE_sock;
#endif
            memset((char *)&(addr), 0, sizeof(addr));
            addr.sin_family = AF_INET;
            addr.sin_addr.s_addr = RTKMR_Object->AddressList[i];
            addr.sin_port = (unsigned short)htons(UPNP_PORT);
#if 1 ////def OLD_SOCKET
            if (setsockopt(RTKMR_Object->NOTIFY_SEND_socks[i], SOL_SOCKET, SO_REUSEADDR,(char*)&ra, sizeof(ra)) == 0)
#endif 
            {
                if (setsockopt(RTKMR_Object->NOTIFY_SEND_socks[i], IPPROTO_IP, IP_MULTICAST_TTL,(char*)&TTL, sizeof(TTL)) < 0)
                {
                    // Ignore the case if setting the Multicast-TTL fails
                }
#if 1 ////def OLD_SOCKET
                if (bind(RTKMR_Object->NOTIFY_SEND_socks[i], (struct sockaddr *) &(addr), sizeof(addr)) == 0)
#endif 
                {
                    mreq.imr_multiaddr.s_addr = inet_addr(UPNP_GROUP);
                    mreq.imr_interface.s_addr = RTKMR_Object->AddressList[i];
                    if (setsockopt(RTKMR_Object->NOTIFY_RECEIVE_sock, IPPROTO_IP, IP_ADD_MEMBERSHIP,(char*)&mreq, sizeof(mreq)) < 0)
                    {
                        // Does not matter if it fails, just ignore
                    }
                }
            }
        }
        
        //
        // Iterate through all the packet types, and re-broadcast
        //
        for(i=1;i<=6;++i)
        {
            f = (struct RTKMR_FragmentNotifyStruct*)malloc(sizeof(struct RTKMR_FragmentNotifyStruct));
            f->packetNumber=i;
            f->upnp = RTKMR_Object;
            //
            // Inject some random delay, to spread these packets out, to help prevent
            // the inbound buffer of the recipient from overflowing, causing dropped packets.
            //
            timeout = (int)(0 + ((unsigned short)rand() % (500)));
            ILibLifeTime_AddEx(f->upnp->WebServerTimer,f,timeout,&RTKMR_FragmentedSendNotify,&RTKMR_FragmentedSendNotify_Destroy);
        }
    }
    FD_SET(RTKMR_Object->NOTIFY_RECEIVE_sock,socketset);
    
}

void RTKMR_MasterPostSelect(void* object,int slct, fd_set *socketset, fd_set *writeset, fd_set *errorset)
{
    #if defined(WIN32) || defined(_WIN32_WCE)
    unsigned long flags=0;
    #endif
    int cnt = 0;
    
    struct packetheader *packet;
    struct RTKMR_DataObject *RTKMR_Object = (struct RTKMR_DataObject*)object;
    
    if(slct>0)
    {
        
        
        //
        // Check to see if we got any Multicast SEARCH requests
        //
        if(FD_ISSET(RTKMR_Object->NOTIFY_RECEIVE_sock,socketset)!=0)
        {	
            cnt = recvfrom(RTKMR_Object->NOTIFY_RECEIVE_sock, RTKMR_Object->message, sizeof(RTKMR_Object->message), 0,
            (struct sockaddr *) &(RTKMR_Object->addr), &(RTKMR_Object->addrlen));
            if (cnt < 0)
            {
                printf("recvfrom");
                exit(1);
            }
            else if (cnt == 0)
            {
                /* Socket Closed? */
            }
            packet = ILibParsePacketHeader(RTKMR_Object->message,0,cnt);
            if(packet!=NULL)
            {
                packet->Source = (struct sockaddr_in*)&(RTKMR_Object->addr);
                packet->ReceivingAddress = 0;
                if(packet->StatusCode==-1 && memcmp(packet->Directive,"M-SEARCH",8)==0)
                {
                    //
                    // Process the search request with our Multicast M-SEARCH Handler
                    //
                    RTKMR_ProcessMSEARCH(RTKMR_Object, packet);
                }
                ILibDestructPacket(packet);
            }
        }
        
    }
}
void RTKMR_FragmentedSendNotify_Destroy(void *data)
{
    free(data);
}
void RTKMR_FragmentedSendNotify(void *data)
{
    struct RTKMR_FragmentNotifyStruct *FNS = (struct RTKMR_FragmentNotifyStruct*)data;
    int timeout;
    int packetlength;
    char* packet = (char*)malloc(5000);
    int i,i2;
    struct sockaddr_in addr;
    int addrlen;
    struct in_addr interface_addr;
    
    memset((char *)&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = inet_addr(UPNP_GROUP);
    addr.sin_port = (unsigned short)htons(UPNP_PORT);
    addrlen = sizeof(addr);
    
    memset((char *)&interface_addr, 0, sizeof(interface_addr));
    
    for(i=0;i<FNS->upnp->AddressListLength;++i)
    {
        interface_addr.s_addr = FNS->upnp->AddressList[i];
        if (setsockopt(FNS->upnp->NOTIFY_SEND_socks[i], IPPROTO_IP, IP_MULTICAST_IF,(char*)&interface_addr, sizeof(interface_addr)) == 0)
        {
            for (i2=0;i2<2;i2++)
            {
                switch(FNS->packetNumber)
                {
                    case 1:
                    RTKMR_BuildSsdpNotifyPacket(packet,&packetlength,FNS->upnp->AddressList[i],(unsigned short)FNS->upnp->WebSocketPortNumber,0,FNS->upnp->UDN,"::upnp:rootdevice","upnp:rootdevice","",FNS->upnp->NotifyCycleTime);
                    sendto(FNS->upnp->NOTIFY_SEND_socks[i], packet, packetlength, 0, (struct sockaddr *) &addr, addrlen);
                    break;
                    case 2:
                    RTKMR_BuildSsdpNotifyPacket(packet,&packetlength,FNS->upnp->AddressList[i],(unsigned short)FNS->upnp->WebSocketPortNumber,0,FNS->upnp->UDN,"","uuid:",FNS->upnp->UDN,FNS->upnp->NotifyCycleTime);
                    sendto(FNS->upnp->NOTIFY_SEND_socks[i], packet, packetlength, 0, (struct sockaddr *) &addr, addrlen);
                    break;
                    case 3:
                    RTKMR_BuildSsdpNotifyPacket(packet,&packetlength,FNS->upnp->AddressList[i],(unsigned short)FNS->upnp->WebSocketPortNumber,0,FNS->upnp->UDN,"::urn:schemas-upnp-org:device:MediaRenderer:1","urn:schemas-upnp-org:device:MediaRenderer:1","",FNS->upnp->NotifyCycleTime);
                    sendto(FNS->upnp->NOTIFY_SEND_socks[i], packet, packetlength, 0, (struct sockaddr *) &addr, addrlen);
                    break;
#if 0
                    case 4:
                    RTKMR_BuildSsdpNotifyPacket(packet,&packetlength,FNS->upnp->AddressList[i],(unsigned short)FNS->upnp->WebSocketPortNumber,0,FNS->upnp->UDN,"::urn:schemas-upnp-org:service:AVTransport:1","urn:schemas-upnp-org:service:AVTransport:1","",FNS->upnp->NotifyCycleTime);
                    sendto(FNS->upnp->NOTIFY_SEND_socks[i], packet, packetlength, 0, (struct sockaddr *) &addr, addrlen);
                    break;
#endif
#if 0
                    case 5:
                    RTKMR_BuildSsdpNotifyPacket(packet,&packetlength,FNS->upnp->AddressList[i],(unsigned short)FNS->upnp->WebSocketPortNumber,0,FNS->upnp->UDN,"::urn:schemas-upnp-org:service:ConnectionManager:1","urn:schemas-upnp-org:service:ConnectionManager:1","",FNS->upnp->NotifyCycleTime);
                    sendto(FNS->upnp->NOTIFY_SEND_socks[i], packet, packetlength, 0, (struct sockaddr *) &addr, addrlen);
                    break;
#endif
#if 0
                    case 6:
                    RTKMR_BuildSsdpNotifyPacket(packet,&packetlength,FNS->upnp->AddressList[i],(unsigned short)FNS->upnp->WebSocketPortNumber,0,FNS->upnp->UDN,"::urn:schemas-upnp-org:service:RenderingControl:1","urn:schemas-upnp-org:service:RenderingControl:1","",FNS->upnp->NotifyCycleTime);
                    sendto(FNS->upnp->NOTIFY_SEND_socks[i], packet, packetlength, 0, (struct sockaddr *) &addr, addrlen);
                    break;
#endif
                    
                }
            }
        }
    }
    free(packet);
    timeout = (int)((FNS->upnp->NotifyCycleTime/4) + ((unsigned short)rand() % (FNS->upnp->NotifyCycleTime/2 - FNS->upnp->NotifyCycleTime/4)));
    ILibLifeTime_Add(FNS->upnp->WebServerTimer,FNS,timeout,&RTKMR_FragmentedSendNotify,&RTKMR_FragmentedSendNotify_Destroy);
}
void RTKMR_SendNotify(const struct RTKMR_DataObject *upnp)
{
    int packetlength;
    char* packet = (char*)malloc(5000);
    int i,i2;
    struct sockaddr_in addr;
    int addrlen;
    struct in_addr interface_addr;
    
    memset((char *)&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = inet_addr(UPNP_GROUP);
    addr.sin_port = (unsigned short)htons(UPNP_PORT);
    addrlen = sizeof(addr);
    
    memset((char *)&interface_addr, 0, sizeof(interface_addr));
    
    for(i=0;i<upnp->AddressListLength;++i)
    {
        interface_addr.s_addr = upnp->AddressList[i];
        #if !defined(_WIN32_WCE) || (defined(_WIN32_WCE) && _WIN32_WCE>=4)
        if (setsockopt(upnp->NOTIFY_SEND_socks[i], IPPROTO_IP, IP_MULTICAST_IF,(char*)&interface_addr, sizeof(interface_addr)) == 0)
        {
            #endif
            for (i2=0;i2<2;i2++)
            {
                RTKMR_BuildSsdpNotifyPacket(packet,&packetlength,upnp->AddressList[i],(unsigned short)upnp->WebSocketPortNumber,0,upnp->UDN,"::upnp:rootdevice","upnp:rootdevice","",upnp->NotifyCycleTime);
                if (sendto(upnp->NOTIFY_SEND_socks[i], packet, packetlength, 0, (struct sockaddr *) &addr, addrlen) < 0) {exit(1);}
                RTKMR_BuildSsdpNotifyPacket(packet,&packetlength,upnp->AddressList[i],(unsigned short)upnp->WebSocketPortNumber,0,upnp->UDN,"","uuid:",upnp->UDN,upnp->NotifyCycleTime);
                if (sendto(upnp->NOTIFY_SEND_socks[i], packet, packetlength, 0, (struct sockaddr *) &addr, addrlen) < 0) {exit(1);}
                RTKMR_BuildSsdpNotifyPacket(packet,&packetlength,upnp->AddressList[i],(unsigned short)upnp->WebSocketPortNumber,0,upnp->UDN,"::urn:schemas-upnp-org:device:MediaRenderer:1","urn:schemas-upnp-org:device:MediaRenderer:1","",upnp->NotifyCycleTime);
#if 0
                if (sendto(upnp->NOTIFY_SEND_socks[i], packet, packetlength, 0, (struct sockaddr *) &addr, addrlen) < 0) {exit(1);}
                RTKMR_BuildSsdpNotifyPacket(packet,&packetlength,upnp->AddressList[i],(unsigned short)upnp->WebSocketPortNumber,0,upnp->UDN,"::urn:schemas-upnp-org:service:AVTransport:1","urn:schemas-upnp-org:service:AVTransport:1","",upnp->NotifyCycleTime);
                if (sendto(upnp->NOTIFY_SEND_socks[i], packet, packetlength, 0, (struct sockaddr *) &addr, addrlen) < 0) {exit(1);}
                RTKMR_BuildSsdpNotifyPacket(packet,&packetlength,upnp->AddressList[i],(unsigned short)upnp->WebSocketPortNumber,0,upnp->UDN,"::urn:schemas-upnp-org:service:ConnectionManager:1","urn:schemas-upnp-org:service:ConnectionManager:1","",upnp->NotifyCycleTime);
                if (sendto(upnp->NOTIFY_SEND_socks[i], packet, packetlength, 0, (struct sockaddr *) &addr, addrlen) < 0) {exit(1);}
                RTKMR_BuildSsdpNotifyPacket(packet,&packetlength,upnp->AddressList[i],(unsigned short)upnp->WebSocketPortNumber,0,upnp->UDN,"::urn:schemas-upnp-org:service:RenderingControl:1","urn:schemas-upnp-org:service:RenderingControl:1","",upnp->NotifyCycleTime);
                if (sendto(upnp->NOTIFY_SEND_socks[i], packet, packetlength, 0, (struct sockaddr *) &addr, addrlen) < 0) {exit(1);}
#endif
                
            }
            #if !defined(_WIN32_WCE) || (defined(_WIN32_WCE) && _WIN32_WCE>=4)
        }
        #endif
    }
    free(packet);
}

#define RTKMR_BuildSsdpByeByePacket(outpacket,outlength,USN,USNex,NT,NTex,DeviceID)\
{\
    if(DeviceID==0)\
    {\
        *outlength = sprintf(outpacket,"NOTIFY * HTTP/1.1\r\nHOST: 239.255.255.250:1900\r\nNTS: ssdp:byebye\r\nUSN: uuid:%s%s\r\nNT: %s%s\r\nContent-Length: 0\r\n\r\n",USN,USNex,NT,NTex);\
    }\
    else\
    {\
        if(memcmp(NT,"uuid:",5)==0)\
        {\
            *outlength = sprintf(outpacket,"NOTIFY * HTTP/1.1\r\nHOST: 239.255.255.250:1900\r\nNTS: ssdp:byebye\r\nUSN: uuid:%s_%d%s\r\nNT: %s%s_%d\r\nContent-Length: 0\r\n\r\n",USN,DeviceID,USNex,NT,NTex,DeviceID);\
        }\
        else\
        {\
            *outlength = sprintf(outpacket,"NOTIFY * HTTP/1.1\r\nHOST: 239.255.255.250:1900\r\nNTS: ssdp:byebye\r\nUSN: uuid:%s_%d%s\r\nNT: %s%s\r\nContent-Length: 0\r\n\r\n",USN,DeviceID,USNex,NT,NTex);\
        }\
    }\
}


void RTKMR_SendByeBye(const struct RTKMR_DataObject *upnp)
{
    ////int TempVal=0;
    int packetlength;
    char* packet = (char*)malloc(5000);
    int i, i2;
    struct sockaddr_in addr;
    int addrlen;
    struct in_addr interface_addr;
    
    memset((char *)&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = inet_addr(UPNP_GROUP);
    addr.sin_port = (unsigned short)htons(UPNP_PORT);
    addrlen = sizeof(addr);
    
    memset((char *)&interface_addr, 0, sizeof(interface_addr));
    
    for(i=0;i<upnp->AddressListLength;++i)
    {
        
        interface_addr.s_addr = upnp->AddressList[i];
        #if !defined(_WIN32_WCE) || (defined(_WIN32_WCE) && _WIN32_WCE>=4)
        if (setsockopt(upnp->NOTIFY_SEND_socks[i], IPPROTO_IP, IP_MULTICAST_IF,(char*)&interface_addr, sizeof(interface_addr)) == 0)
        {
            #endif		
            
            for (i2=0;i2<2;i2++)
            {
                RTKMR_BuildSsdpByeByePacket(packet,&packetlength,upnp->UDN,"::upnp:rootdevice","upnp:rootdevice","",0);
                if (sendto(upnp->NOTIFY_SEND_socks[i], packet, packetlength, 0, (struct sockaddr *) &addr, addrlen) < 0) exit(1);
                RTKMR_BuildSsdpByeByePacket(packet,&packetlength,upnp->UDN,"","uuid:",upnp->UDN,0);
                if (sendto(upnp->NOTIFY_SEND_socks[i], packet, packetlength, 0, (struct sockaddr *) &addr, addrlen) < 0) exit(1);
                RTKMR_BuildSsdpByeByePacket(packet,&packetlength,upnp->UDN,"::urn:schemas-upnp-org:device:MediaRenderer:1","urn:schemas-upnp-org:device:MediaRenderer:1","",0);
                if (sendto(upnp->NOTIFY_SEND_socks[i], packet, packetlength, 0, (struct sockaddr *) &addr, addrlen) < 0) exit(1);
#if 0
                RTKMR_BuildSsdpByeByePacket(packet,&packetlength,upnp->UDN,"::urn:schemas-upnp-org:service:AVTransport:1","urn:schemas-upnp-org:service:AVTransport:1","",0);
                if (sendto(upnp->NOTIFY_SEND_socks[i], packet, packetlength, 0, (struct sockaddr *) &addr, addrlen) < 0) exit(1);
                RTKMR_BuildSsdpByeByePacket(packet,&packetlength,upnp->UDN,"::urn:schemas-upnp-org:service:ConnectionManager:1","urn:schemas-upnp-org:service:ConnectionManager:1","",0);
                if (sendto(upnp->NOTIFY_SEND_socks[i], packet, packetlength, 0, (struct sockaddr *) &addr, addrlen) < 0) exit(1);
                RTKMR_BuildSsdpByeByePacket(packet,&packetlength,upnp->UDN,"::urn:schemas-upnp-org:service:RenderingControl:1","urn:schemas-upnp-org:service:RenderingControl:1","",0);
                if (sendto(upnp->NOTIFY_SEND_socks[i], packet, packetlength, 0, (struct sockaddr *) &addr, addrlen) < 0) exit(1);
#endif
                
            }
            #if !defined(_WIN32_WCE) || (defined(_WIN32_WCE) && _WIN32_WCE>=4)
        }
        #endif
    }
    free(packet);
}

/*! \fn RTKMR_Response_Error(const RTKMR_SessionToken RTKMR_Token, const int ErrorCode, const char* ErrorMsg)
\brief Responds to the client invocation with a SOAP Fault
\param RTKMR_Token UPnP token
\param ErrorCode Fault Code
\param ErrorMsg Error Detail
*/
void RTKMR_Response_Error(const RTKMR_SessionToken RTKMR_Token, const int ErrorCode, const char* ErrorMsg)
{
    char* body;
    int bodylength;
    char* head;
    int headlength;
    body = (char*)malloc(395 + (int)strlen(ErrorMsg));
    bodylength = sprintf(body,"<s:Envelope\r\n xmlns:s=\"http://schemas.xmlsoap.org/soap/envelope/\" s:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\"><s:Body><s:Fault><faultcode>s:Client</faultcode><faultstring>UPnPError</faultstring><detail><UPnPError xmlns=\"urn:schemas-upnp-org:control-1-0\"><errorCode>%d</errorCode><errorDescription>%s</errorDescription></UPnPError></detail></s:Fault></s:Body></s:Envelope>",ErrorCode,ErrorMsg);
    head = (char*)malloc(59);
    headlength = sprintf(head,"HTTP/1.1 500 Internal\r\nContent-Length: %d\r\n\r\n",bodylength);
    ILibWebServer_Send_Raw((struct ILibWebServer_Session*)RTKMR_Token,head,headlength,0,0);
    ILibWebServer_Send_Raw((struct ILibWebServer_Session*)RTKMR_Token,body,bodylength,0,1);
}

/*! \fn RTKMR_GetLocalInterfaceToHost(const RTKMR_SessionToken RTKMR_Token)
\brief When a UPnP request is dispatched, this method determines which ip address actually received this request
\param RTKMR_Token UPnP token
\returns IP Address
*/
int RTKMR_GetLocalInterfaceToHost(const RTKMR_SessionToken RTKMR_Token)
{
    return(ILibWebServer_GetLocalInterface((struct ILibWebServer_Session*)RTKMR_Token));
}

void RTKMR_ResponseGeneric(const RTKMR_MicroStackToken RTKMR_Token,const char* ServiceURI,const char* MethodName,const char* Params)
{
    char* packet;
    int packetlength;
    struct ILibWebServer_Session *session = (struct ILibWebServer_Session*)RTKMR_Token;
    int RVAL=0;
    
    packet = (char*)malloc(239+strlen(ServiceURI)+strlen(Params)+(strlen(MethodName)*2));
    packetlength = sprintf(packet,"<?xml version=\"1.0\" encoding=\"utf-8\"?>\r\n<s:Envelope s:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\" xmlns:s=\"http://schemas.xmlsoap.org/soap/envelope/\"><s:Body><u:%sResponse xmlns:u=\"%s\">%s</u:%sResponse></s:Body></s:Envelope>",MethodName,ServiceURI,Params,MethodName);
    LVL3DEBUG(printf("SendBody: %s\r\n",packet);)
    #if defined(WIN32) || defined(_WIN32_WCE)
    RVAL=ILibWebServer_StreamHeader_Raw(session,200,"OK","\r\nEXT:\r\nCONTENT-TYPE: text/xml; charset=\"utf-8\"\r\nSERVER: WINDOWS, UPnP/1.0, Intel MicroStack/1.0.1868",1);
    #else
    RVAL=ILibWebServer_StreamHeader_Raw(session,200,"OK","\r\nEXT:\r\nCONTENT-TYPE: text/xml; charset=\"utf-8\"\r\nSERVER: POSIX, UPnP/1.0, Intel MicroStack/1.0.1868",1);
    #endif
    if(RVAL!=ILibAsyncSocket_SEND_ON_CLOSED_SOCKET_ERROR && RVAL != ILibWebServer_SEND_RESULTED_IN_DISCONNECT)
    {
        RVAL=ILibWebServer_StreamBody(session,packet,packetlength,0,1);
    }
}

/*! \fn RTKMR_Response_AVTransport_GetDeviceCapabilities(const RTKMR_SessionToken RTKMR_Token, const char* unescaped_PlayMedia, const char* unescaped_RecMedia, const char* unescaped_RecQualityModes)
\brief Response Method for AVTransport >> urn:schemas-upnp-org:service:AVTransport:1 >> GetDeviceCapabilities
\param RTKMR_Token MicroStack token
\param unescaped_PlayMedia Value of argument PlayMedia \b     Note: Automatically Escaped
\param unescaped_RecMedia Value of argument RecMedia \b     Note: Automatically Escaped
\param unescaped_RecQualityModes Value of argument RecQualityModes \b     Note: Automatically Escaped
*/
void RTKMR_Response_AVTransport_GetDeviceCapabilities(const RTKMR_SessionToken RTKMR_Token, const char* unescaped_PlayMedia, const char* unescaped_RecMedia, const char* unescaped_RecQualityModes)
{
    char* body;
    char *PlayMedia = (char*)malloc(1+ILibXmlEscapeLength(unescaped_PlayMedia));
    char *RecMedia = (char*)malloc(1+ILibXmlEscapeLength(unescaped_RecMedia));
    char *RecQualityModes = (char*)malloc(1+ILibXmlEscapeLength(unescaped_RecQualityModes));
    
    ILibXmlEscape(PlayMedia,unescaped_PlayMedia);
    ILibXmlEscape(RecMedia,unescaped_RecMedia);
    ILibXmlEscape(RecQualityModes,unescaped_RecQualityModes);
    body = (char*)malloc(80+strlen(PlayMedia)+strlen(RecMedia)+strlen(RecQualityModes));
    sprintf(body,"<PlayMedia>%s</PlayMedia><RecMedia>%s</RecMedia><RecQualityModes>%s</RecQualityModes>",PlayMedia,RecMedia,RecQualityModes);
    RTKMR_ResponseGeneric(RTKMR_Token,"urn:schemas-upnp-org:service:AVTransport:1","GetDeviceCapabilities",body);
    free(body);
    free(PlayMedia);
    free(RecMedia);
    free(RecQualityModes);
}

/*! \fn RTKMR_Response_AVTransport_GetMediaInfo(const RTKMR_SessionToken RTKMR_Token, const unsigned int NrTracks, const char* unescaped_MediaDuration, const char* unescaped_CurrentURI, const char* unescaped_CurrentURIMetaData, const char* unescaped_NextURI, const char* unescaped_NextURIMetaData, const char* unescaped_PlayMedium, const char* unescaped_RecordMedium, const char* unescaped_WriteStatus)
\brief Response Method for AVTransport >> urn:schemas-upnp-org:service:AVTransport:1 >> GetMediaInfo
\param RTKMR_Token MicroStack token
\param NrTracks Value of argument NrTracks
\param unescaped_MediaDuration Value of argument MediaDuration \b     Note: Automatically Escaped
\param unescaped_CurrentURI Value of argument CurrentURI \b     Note: Automatically Escaped
\param unescaped_CurrentURIMetaData Value of argument CurrentURIMetaData \b     Note: Automatically Escaped
\param unescaped_NextURI Value of argument NextURI \b     Note: Automatically Escaped
\param unescaped_NextURIMetaData Value of argument NextURIMetaData \b     Note: Automatically Escaped
\param unescaped_PlayMedium Value of argument PlayMedium \b     Note: Automatically Escaped
\param unescaped_RecordMedium Value of argument RecordMedium \b     Note: Automatically Escaped
\param unescaped_WriteStatus Value of argument WriteStatus \b     Note: Automatically Escaped
*/
void RTKMR_Response_AVTransport_GetMediaInfo(const RTKMR_SessionToken RTKMR_Token, const unsigned int NrTracks, const char* unescaped_MediaDuration, const char* unescaped_CurrentURI, const char* unescaped_CurrentURIMetaData, const char* unescaped_NextURI, const char* unescaped_NextURIMetaData, const char* unescaped_PlayMedium, const char* unescaped_RecordMedium, const char* unescaped_WriteStatus)
{
    char* body;
    char *MediaDuration = (char*)malloc(1+ILibXmlEscapeLength(unescaped_MediaDuration));
    char *CurrentURI = (char*)malloc(1+ILibXmlEscapeLength(unescaped_CurrentURI));
    char *CurrentURIMetaData = (char*)malloc(1+ILibXmlEscapeLength(unescaped_CurrentURIMetaData));
    char *NextURI = (char*)malloc(1+ILibXmlEscapeLength(unescaped_NextURI));
    char *NextURIMetaData = (char*)malloc(1+ILibXmlEscapeLength(unescaped_NextURIMetaData));
    char *PlayMedium = (char*)malloc(1+ILibXmlEscapeLength(unescaped_PlayMedium));
    char *RecordMedium = (char*)malloc(1+ILibXmlEscapeLength(unescaped_RecordMedium));
    char *WriteStatus = (char*)malloc(1+ILibXmlEscapeLength(unescaped_WriteStatus));
    
    ILibXmlEscape(MediaDuration,unescaped_MediaDuration);
    ILibXmlEscape(CurrentURI,unescaped_CurrentURI);
    ILibXmlEscape(CurrentURIMetaData,unescaped_CurrentURIMetaData);
    ILibXmlEscape(NextURI,unescaped_NextURI);
    ILibXmlEscape(NextURIMetaData,unescaped_NextURIMetaData);
    ILibXmlEscape(PlayMedium,unescaped_PlayMedium);
    ILibXmlEscape(RecordMedium,unescaped_RecordMedium);
    ILibXmlEscape(WriteStatus,unescaped_WriteStatus);
    body = (char*)malloc(265+strlen(MediaDuration)+strlen(CurrentURI)+strlen(CurrentURIMetaData)+strlen(NextURI)+strlen(NextURIMetaData)+strlen(PlayMedium)+strlen(RecordMedium)+strlen(WriteStatus));
    sprintf(body,"<NrTracks>%u</NrTracks><MediaDuration>%s</MediaDuration><CurrentURI>%s</CurrentURI><CurrentURIMetaData>%s</CurrentURIMetaData><NextURI>%s</NextURI><NextURIMetaData>%s</NextURIMetaData><PlayMedium>%s</PlayMedium><RecordMedium>%s</RecordMedium><WriteStatus>%s</WriteStatus>",NrTracks,MediaDuration,CurrentURI,CurrentURIMetaData,NextURI,NextURIMetaData,PlayMedium,RecordMedium,WriteStatus);
    RTKMR_ResponseGeneric(RTKMR_Token,"urn:schemas-upnp-org:service:AVTransport:1","GetMediaInfo",body);
    free(body);
    free(MediaDuration);
    free(CurrentURI);
    free(CurrentURIMetaData);
    free(NextURI);
    free(NextURIMetaData);
    free(PlayMedium);
    free(RecordMedium);
    free(WriteStatus);
}

/*! \fn RTKMR_Response_AVTransport_GetPositionInfo(const RTKMR_SessionToken RTKMR_Token, const unsigned int Track, const char* unescaped_TrackDuration, const char* unescaped_TrackMetaData, const char* unescaped_TrackURI, const char* unescaped_RelTime, const char* unescaped_AbsTime, const int RelCount, const int AbsCount)
\brief Response Method for AVTransport >> urn:schemas-upnp-org:service:AVTransport:1 >> GetPositionInfo
\param RTKMR_Token MicroStack token
\param Track Value of argument Track
\param unescaped_TrackDuration Value of argument TrackDuration \b     Note: Automatically Escaped
\param unescaped_TrackMetaData Value of argument TrackMetaData \b     Note: Automatically Escaped
\param unescaped_TrackURI Value of argument TrackURI \b     Note: Automatically Escaped
\param unescaped_RelTime Value of argument RelTime \b     Note: Automatically Escaped
\param unescaped_AbsTime Value of argument AbsTime \b     Note: Automatically Escaped
\param RelCount Value of argument RelCount
\param AbsCount Value of argument AbsCount
*/
void RTKMR_Response_AVTransport_GetPositionInfo(const RTKMR_SessionToken RTKMR_Token, const unsigned int Track, const char* unescaped_TrackDuration, const char* unescaped_TrackMetaData, const char* unescaped_TrackURI, const char* unescaped_RelTime, const char* unescaped_AbsTime, const int RelCount, const int AbsCount)
{
    char* body;
    char *TrackDuration = (char*)malloc(1+ILibXmlEscapeLength(unescaped_TrackDuration));
    char *TrackMetaData = (char*)malloc(1+ILibXmlEscapeLength(unescaped_TrackMetaData));
    char *TrackURI = (char*)malloc(1+ILibXmlEscapeLength(unescaped_TrackURI));
    char *RelTime = (char*)malloc(1+ILibXmlEscapeLength(unescaped_RelTime));
    char *AbsTime = (char*)malloc(1+ILibXmlEscapeLength(unescaped_AbsTime));
    
    ILibXmlEscape(TrackDuration,unescaped_TrackDuration);
    ILibXmlEscape(TrackMetaData,unescaped_TrackMetaData);
    ILibXmlEscape(TrackURI,unescaped_TrackURI);
    ILibXmlEscape(RelTime,unescaped_RelTime);
    ILibXmlEscape(AbsTime,unescaped_AbsTime);
    body = (char*)malloc(212+strlen(TrackDuration)+strlen(TrackMetaData)+strlen(TrackURI)+strlen(RelTime)+strlen(AbsTime));
    sprintf(body,"<Track>%u</Track><TrackDuration>%s</TrackDuration><TrackMetaData>%s</TrackMetaData><TrackURI>%s</TrackURI><RelTime>%s</RelTime><AbsTime>%s</AbsTime><RelCount>%d</RelCount><AbsCount>%d</AbsCount>",Track,TrackDuration,TrackMetaData,TrackURI,RelTime,AbsTime,RelCount,AbsCount);
    RTKMR_ResponseGeneric(RTKMR_Token,"urn:schemas-upnp-org:service:AVTransport:1","GetPositionInfo",body);
    free(body);
    free(TrackDuration);
    free(TrackMetaData);
    free(TrackURI);
    free(RelTime);
    free(AbsTime);
}

/*! \fn RTKMR_Response_AVTransport_GetTransportInfo(const RTKMR_SessionToken RTKMR_Token, const char* unescaped_CurrentTransportState, const char* unescaped_CurrentTransportStatus, const char* unescaped_CurrentSpeed)
\brief Response Method for AVTransport >> urn:schemas-upnp-org:service:AVTransport:1 >> GetTransportInfo
\param RTKMR_Token MicroStack token
\param unescaped_CurrentTransportState Value of argument CurrentTransportState \b     Note: Automatically Escaped
\param unescaped_CurrentTransportStatus Value of argument CurrentTransportStatus \b     Note: Automatically Escaped
\param unescaped_CurrentSpeed Value of argument CurrentSpeed \b     Note: Automatically Escaped
*/
void RTKMR_Response_AVTransport_GetTransportInfo(const RTKMR_SessionToken RTKMR_Token, const char* unescaped_CurrentTransportState, const char* unescaped_CurrentTransportStatus, const char* unescaped_CurrentSpeed)
{
    char* body;
    char *CurrentTransportState = (char*)malloc(1+ILibXmlEscapeLength(unescaped_CurrentTransportState));
    char *CurrentTransportStatus = (char*)malloc(1+ILibXmlEscapeLength(unescaped_CurrentTransportStatus));
    char *CurrentSpeed = (char*)malloc(1+ILibXmlEscapeLength(unescaped_CurrentSpeed));
    
    ILibXmlEscape(CurrentTransportState,unescaped_CurrentTransportState);
    ILibXmlEscape(CurrentTransportStatus,unescaped_CurrentTransportStatus);
    ILibXmlEscape(CurrentSpeed,unescaped_CurrentSpeed);
    body = (char*)malloc(126+strlen(CurrentTransportState)+strlen(CurrentTransportStatus)+strlen(CurrentSpeed));
    sprintf(body,"<CurrentTransportState>%s</CurrentTransportState><CurrentTransportStatus>%s</CurrentTransportStatus><CurrentSpeed>%s</CurrentSpeed>",CurrentTransportState,CurrentTransportStatus,CurrentSpeed);
    RTKMR_ResponseGeneric(RTKMR_Token,"urn:schemas-upnp-org:service:AVTransport:1","GetTransportInfo",body);
    free(body);
    free(CurrentTransportState);
    free(CurrentTransportStatus);
    free(CurrentSpeed);
}

/*! \fn RTKMR_Response_AVTransport_GetTransportSettings(const RTKMR_SessionToken RTKMR_Token, const char* unescaped_PlayMode, const char* unescaped_RecQualityMode)
\brief Response Method for AVTransport >> urn:schemas-upnp-org:service:AVTransport:1 >> GetTransportSettings
\param RTKMR_Token MicroStack token
\param unescaped_PlayMode Value of argument PlayMode \b     Note: Automatically Escaped
\param unescaped_RecQualityMode Value of argument RecQualityMode \b     Note: Automatically Escaped
*/
void RTKMR_Response_AVTransport_GetTransportSettings(const RTKMR_SessionToken RTKMR_Token, const char* unescaped_PlayMode, const char* unescaped_RecQualityMode)
{
    char* body;
    char *PlayMode = (char*)malloc(1+ILibXmlEscapeLength(unescaped_PlayMode));
    char *RecQualityMode = (char*)malloc(1+ILibXmlEscapeLength(unescaped_RecQualityMode));
    
    ILibXmlEscape(PlayMode,unescaped_PlayMode);
    ILibXmlEscape(RecQualityMode,unescaped_RecQualityMode);
    body = (char*)malloc(55+strlen(PlayMode)+strlen(RecQualityMode));
    sprintf(body,"<PlayMode>%s</PlayMode><RecQualityMode>%s</RecQualityMode>",PlayMode,RecQualityMode);
    RTKMR_ResponseGeneric(RTKMR_Token,"urn:schemas-upnp-org:service:AVTransport:1","GetTransportSettings",body);
    free(body);
    free(PlayMode);
    free(RecQualityMode);
}

/*! \fn RTKMR_Response_AVTransport_Next(const RTKMR_SessionToken RTKMR_Token)
\brief Response Method for AVTransport >> urn:schemas-upnp-org:service:AVTransport:1 >> Next
\param RTKMR_Token MicroStack token
*/
void RTKMR_Response_AVTransport_Next(const RTKMR_SessionToken RTKMR_Token)
{
    RTKMR_ResponseGeneric(RTKMR_Token,"urn:schemas-upnp-org:service:AVTransport:1","Next","");
}

/*! \fn RTKMR_Response_AVTransport_Pause(const RTKMR_SessionToken RTKMR_Token)
\brief Response Method for AVTransport >> urn:schemas-upnp-org:service:AVTransport:1 >> Pause
\param RTKMR_Token MicroStack token
*/
void RTKMR_Response_AVTransport_Pause(const RTKMR_SessionToken RTKMR_Token)
{
    RTKMR_ResponseGeneric(RTKMR_Token,"urn:schemas-upnp-org:service:AVTransport:1","Pause","");
}

/*! \fn RTKMR_Response_AVTransport_Play(const RTKMR_SessionToken RTKMR_Token)
\brief Response Method for AVTransport >> urn:schemas-upnp-org:service:AVTransport:1 >> Play
\param RTKMR_Token MicroStack token
*/
void RTKMR_Response_AVTransport_Play(const RTKMR_SessionToken RTKMR_Token)
{
    RTKMR_ResponseGeneric(RTKMR_Token,"urn:schemas-upnp-org:service:AVTransport:1","Play","");
}

/*! \fn RTKMR_Response_AVTransport_Previous(const RTKMR_SessionToken RTKMR_Token)
\brief Response Method for AVTransport >> urn:schemas-upnp-org:service:AVTransport:1 >> Previous
\param RTKMR_Token MicroStack token
*/
void RTKMR_Response_AVTransport_Previous(const RTKMR_SessionToken RTKMR_Token)
{
    RTKMR_ResponseGeneric(RTKMR_Token,"urn:schemas-upnp-org:service:AVTransport:1","Previous","");
}

/*! \fn RTKMR_Response_AVTransport_Seek(const RTKMR_SessionToken RTKMR_Token)
\brief Response Method for AVTransport >> urn:schemas-upnp-org:service:AVTransport:1 >> Seek
\param RTKMR_Token MicroStack token
*/
void RTKMR_Response_AVTransport_Seek(const RTKMR_SessionToken RTKMR_Token)
{
    RTKMR_ResponseGeneric(RTKMR_Token,"urn:schemas-upnp-org:service:AVTransport:1","Seek","");
}

/*! \fn RTKMR_Response_AVTransport_SetAVTransportURI(const RTKMR_SessionToken RTKMR_Token)
\brief Response Method for AVTransport >> urn:schemas-upnp-org:service:AVTransport:1 >> SetAVTransportURI
\param RTKMR_Token MicroStack token
*/
void RTKMR_Response_AVTransport_SetAVTransportURI(const RTKMR_SessionToken RTKMR_Token)
{
    RTKMR_ResponseGeneric(RTKMR_Token,"urn:schemas-upnp-org:service:AVTransport:1","SetAVTransportURI","");
}

/*! \fn RTKMR_Response_AVTransport_SetPlayMode(const RTKMR_SessionToken RTKMR_Token)
\brief Response Method for AVTransport >> urn:schemas-upnp-org:service:AVTransport:1 >> SetPlayMode
\param RTKMR_Token MicroStack token
*/
void RTKMR_Response_AVTransport_SetPlayMode(const RTKMR_SessionToken RTKMR_Token)
{
    RTKMR_ResponseGeneric(RTKMR_Token,"urn:schemas-upnp-org:service:AVTransport:1","SetPlayMode","");
}

/*! \fn RTKMR_Response_AVTransport_Stop(const RTKMR_SessionToken RTKMR_Token)
\brief Response Method for AVTransport >> urn:schemas-upnp-org:service:AVTransport:1 >> Stop
\param RTKMR_Token MicroStack token
*/
void RTKMR_Response_AVTransport_Stop(const RTKMR_SessionToken RTKMR_Token)
{
    RTKMR_ResponseGeneric(RTKMR_Token,"urn:schemas-upnp-org:service:AVTransport:1","Stop","");
}

/*! \fn RTKMR_Response_ConnectionManager_ConnectionComplete(const RTKMR_SessionToken RTKMR_Token)
\brief Response Method for ConnectionManager >> urn:schemas-upnp-org:service:ConnectionManager:1 >> ConnectionComplete
\param RTKMR_Token MicroStack token
*/
void RTKMR_Response_ConnectionManager_ConnectionComplete(const RTKMR_SessionToken RTKMR_Token)
{
    RTKMR_ResponseGeneric(RTKMR_Token,"urn:schemas-upnp-org:service:ConnectionManager:1","ConnectionComplete","");
}

/*! \fn RTKMR_Response_ConnectionManager_GetCurrentConnectionIDs(const RTKMR_SessionToken RTKMR_Token, const char* unescaped_ConnectionIDs)
\brief Response Method for ConnectionManager >> urn:schemas-upnp-org:service:ConnectionManager:1 >> GetCurrentConnectionIDs
\param RTKMR_Token MicroStack token
\param unescaped_ConnectionIDs Value of argument ConnectionIDs \b     Note: Automatically Escaped
*/
void RTKMR_Response_ConnectionManager_GetCurrentConnectionIDs(const RTKMR_SessionToken RTKMR_Token, const char* unescaped_ConnectionIDs)
{
    char* body;
    char *ConnectionIDs = (char*)malloc(1+ILibXmlEscapeLength(unescaped_ConnectionIDs));
    
    ILibXmlEscape(ConnectionIDs,unescaped_ConnectionIDs);
    body = (char*)malloc(32+strlen(ConnectionIDs));
    sprintf(body,"<ConnectionIDs>%s</ConnectionIDs>",ConnectionIDs);
    RTKMR_ResponseGeneric(RTKMR_Token,"urn:schemas-upnp-org:service:ConnectionManager:1","GetCurrentConnectionIDs",body);
    free(body);
    free(ConnectionIDs);
}

/*! \fn RTKMR_Response_ConnectionManager_GetCurrentConnectionInfo(const RTKMR_SessionToken RTKMR_Token, const int RcsID, const int AVTransportID, const char* unescaped_ProtocolInfo, const char* unescaped_PeerConnectionManager, const int PeerConnectionID, const char* unescaped_Direction, const char* unescaped_Status)
\brief Response Method for ConnectionManager >> urn:schemas-upnp-org:service:ConnectionManager:1 >> GetCurrentConnectionInfo
\param RTKMR_Token MicroStack token
\param RcsID Value of argument RcsID
\param AVTransportID Value of argument AVTransportID
\param unescaped_ProtocolInfo Value of argument ProtocolInfo \b     Note: Automatically Escaped
\param unescaped_PeerConnectionManager Value of argument PeerConnectionManager \b     Note: Automatically Escaped
\param PeerConnectionID Value of argument PeerConnectionID
\param unescaped_Direction Value of argument Direction \b     Note: Automatically Escaped
\param unescaped_Status Value of argument Status \b     Note: Automatically Escaped
*/
void RTKMR_Response_ConnectionManager_GetCurrentConnectionInfo(const RTKMR_SessionToken RTKMR_Token, const int RcsID, const int AVTransportID, const char* unescaped_ProtocolInfo, const char* unescaped_PeerConnectionManager, const int PeerConnectionID, const char* unescaped_Direction, const char* unescaped_Status)
{
    char* body;
    char *ProtocolInfo = (char*)malloc(1+ILibXmlEscapeLength(unescaped_ProtocolInfo));
    char *PeerConnectionManager = (char*)malloc(1+ILibXmlEscapeLength(unescaped_PeerConnectionManager));
    char *Direction = (char*)malloc(1+ILibXmlEscapeLength(unescaped_Direction));
    char *Status = (char*)malloc(1+ILibXmlEscapeLength(unescaped_Status));
    
    ILibXmlEscape(ProtocolInfo,unescaped_ProtocolInfo);
    ILibXmlEscape(PeerConnectionManager,unescaped_PeerConnectionManager);
    ILibXmlEscape(Direction,unescaped_Direction);
    ILibXmlEscape(Status,unescaped_Status);
    body = (char*)malloc(233+strlen(ProtocolInfo)+strlen(PeerConnectionManager)+strlen(Direction)+strlen(Status));
    sprintf(body,"<RcsID>%d</RcsID><AVTransportID>%d</AVTransportID><ProtocolInfo>%s</ProtocolInfo><PeerConnectionManager>%s</PeerConnectionManager><PeerConnectionID>%d</PeerConnectionID><Direction>%s</Direction><Status>%s</Status>",RcsID,AVTransportID,ProtocolInfo,PeerConnectionManager,PeerConnectionID,Direction,Status);
    RTKMR_ResponseGeneric(RTKMR_Token,"urn:schemas-upnp-org:service:ConnectionManager:1","GetCurrentConnectionInfo",body);
    free(body);
    free(ProtocolInfo);
    free(PeerConnectionManager);
    free(Direction);
    free(Status);
}

/*! \fn RTKMR_Response_ConnectionManager_GetProtocolInfo(const RTKMR_SessionToken RTKMR_Token, const char* unescaped_Source, const char* unescaped_Sink)
\brief Response Method for ConnectionManager >> urn:schemas-upnp-org:service:ConnectionManager:1 >> GetProtocolInfo
\param RTKMR_Token MicroStack token
\param unescaped_Source Value of argument Source \b     Note: Automatically Escaped
\param unescaped_Sink Value of argument Sink \b     Note: Automatically Escaped
*/
void RTKMR_Response_ConnectionManager_GetProtocolInfo(const RTKMR_SessionToken RTKMR_Token, const char* unescaped_Source, const char* unescaped_Sink)
{
    char* body;
    char *Source = (char*)malloc(1+ILibXmlEscapeLength(unescaped_Source));
    char *Sink = (char*)malloc(1+ILibXmlEscapeLength(unescaped_Sink));
    
    ILibXmlEscape(Source,unescaped_Source);
    ILibXmlEscape(Sink,unescaped_Sink);
    body = (char*)malloc(31+strlen(Source)+strlen(Sink));
    sprintf(body,"<Source>%s</Source><Sink>%s</Sink>",Source,Sink);
    RTKMR_ResponseGeneric(RTKMR_Token,"urn:schemas-upnp-org:service:ConnectionManager:1","GetProtocolInfo",body);
    free(body);
    free(Source);
    free(Sink);
}

/*! \fn RTKMR_Response_ConnectionManager_PrepareForConnection(const RTKMR_SessionToken RTKMR_Token, const int ConnectionID, const int AVTransportID, const int RcsID)
\brief Response Method for ConnectionManager >> urn:schemas-upnp-org:service:ConnectionManager:1 >> PrepareForConnection
\param RTKMR_Token MicroStack token
\param ConnectionID Value of argument ConnectionID
\param AVTransportID Value of argument AVTransportID
\param RcsID Value of argument RcsID
*/
void RTKMR_Response_ConnectionManager_PrepareForConnection(const RTKMR_SessionToken RTKMR_Token, const int ConnectionID, const int AVTransportID, const int RcsID)
{
    char* body;
    
    body = (char*)malloc(109);
    sprintf(body,"<ConnectionID>%d</ConnectionID><AVTransportID>%d</AVTransportID><RcsID>%d</RcsID>",ConnectionID,AVTransportID,RcsID);
    RTKMR_ResponseGeneric(RTKMR_Token,"urn:schemas-upnp-org:service:ConnectionManager:1","PrepareForConnection",body);
    free(body);
}

/*! \fn RTKMR_Response_RenderingControl_GetMute(const RTKMR_SessionToken RTKMR_Token, const int CurrentMute)
\brief Response Method for RenderingControl >> urn:schemas-upnp-org:service:RenderingControl:1 >> GetMute
\param RTKMR_Token MicroStack token
\param CurrentMute Value of argument CurrentMute
*/
void RTKMR_Response_RenderingControl_GetMute(const RTKMR_SessionToken RTKMR_Token, const int CurrentMute)
{
    char* body;
    
    body = (char*)malloc(29);
    sprintf(body,"<CurrentMute>%d</CurrentMute>",(CurrentMute!=0?1:0));
    RTKMR_ResponseGeneric(RTKMR_Token,"urn:schemas-upnp-org:service:RenderingControl:1","GetMute",body);
    free(body);
}

/*! \fn RTKMR_Response_RenderingControl_GetVolume(const RTKMR_SessionToken RTKMR_Token, const unsigned short CurrentVolume)
\brief Response Method for RenderingControl >> urn:schemas-upnp-org:service:RenderingControl:1 >> GetVolume
\param RTKMR_Token MicroStack token
\param CurrentVolume Value of argument CurrentVolume
*/
void RTKMR_Response_RenderingControl_GetVolume(const RTKMR_SessionToken RTKMR_Token, const unsigned short CurrentVolume)
{
    char* body;
    
    body = (char*)malloc(38);
    sprintf(body,"<CurrentVolume>%u</CurrentVolume>",CurrentVolume);
    RTKMR_ResponseGeneric(RTKMR_Token,"urn:schemas-upnp-org:service:RenderingControl:1","GetVolume",body);
    free(body);
}

/*! \fn RTKMR_Response_RenderingControl_GetVolumeDB(const RTKMR_SessionToken RTKMR_Token, const short CurrentVolume)
\brief Response Method for RenderingControl >> urn:schemas-upnp-org:service:RenderingControl:1 >> GetVolumeDB
\param RTKMR_Token MicroStack token
\param CurrentVolume Value of argument CurrentVolume
*/
void RTKMR_Response_RenderingControl_GetVolumeDB(const RTKMR_SessionToken RTKMR_Token, const short CurrentVolume)
{
    char* body;
    
    body = (char*)malloc(38);
    sprintf(body,"<CurrentVolume>%d</CurrentVolume>",CurrentVolume);
    RTKMR_ResponseGeneric(RTKMR_Token,"urn:schemas-upnp-org:service:RenderingControl:1","GetVolumeDB",body);
    free(body);
}

/*! \fn RTKMR_Response_RenderingControl_GetVolumeDBRange(const RTKMR_SessionToken RTKMR_Token, const short MinValue, const short MaxValue)
\brief Response Method for RenderingControl >> urn:schemas-upnp-org:service:RenderingControl:1 >> GetVolumeDBRange
\param RTKMR_Token MicroStack token
\param MinValue Value of argument MinValue
\param MaxValue Value of argument MaxValue
*/
void RTKMR_Response_RenderingControl_GetVolumeDBRange(const RTKMR_SessionToken RTKMR_Token, const short MinValue, const short MaxValue)
{
    char* body;
    
    body = (char*)malloc(55);
    sprintf(body,"<MinValue>%d</MinValue><MaxValue>%d</MaxValue>",MinValue,MaxValue);
    RTKMR_ResponseGeneric(RTKMR_Token,"urn:schemas-upnp-org:service:RenderingControl:1","GetVolumeDBRange",body);
    free(body);
}

/*! \fn RTKMR_Response_RenderingControl_ListPresets(const RTKMR_SessionToken RTKMR_Token, const char* unescaped_CurrentPresetNameList)
\brief Response Method for RenderingControl >> urn:schemas-upnp-org:service:RenderingControl:1 >> ListPresets
\param RTKMR_Token MicroStack token
\param unescaped_CurrentPresetNameList Value of argument CurrentPresetNameList \b     Note: Automatically Escaped
*/
void RTKMR_Response_RenderingControl_ListPresets(const RTKMR_SessionToken RTKMR_Token, const char* unescaped_CurrentPresetNameList)
{
    char* body;
    char *CurrentPresetNameList = (char*)malloc(1+ILibXmlEscapeLength(unescaped_CurrentPresetNameList));
    
    ILibXmlEscape(CurrentPresetNameList,unescaped_CurrentPresetNameList);
    body = (char*)malloc(48+strlen(CurrentPresetNameList));
    sprintf(body,"<CurrentPresetNameList>%s</CurrentPresetNameList>",CurrentPresetNameList);
    RTKMR_ResponseGeneric(RTKMR_Token,"urn:schemas-upnp-org:service:RenderingControl:1","ListPresets",body);
    free(body);
    free(CurrentPresetNameList);
}

/*! \fn RTKMR_Response_RenderingControl_SelectPreset(const RTKMR_SessionToken RTKMR_Token)
\brief Response Method for RenderingControl >> urn:schemas-upnp-org:service:RenderingControl:1 >> SelectPreset
\param RTKMR_Token MicroStack token
*/
void RTKMR_Response_RenderingControl_SelectPreset(const RTKMR_SessionToken RTKMR_Token)
{
    RTKMR_ResponseGeneric(RTKMR_Token,"urn:schemas-upnp-org:service:RenderingControl:1","SelectPreset","");
}

/*! \fn RTKMR_Response_RenderingControl_SetMute(const RTKMR_SessionToken RTKMR_Token)
\brief Response Method for RenderingControl >> urn:schemas-upnp-org:service:RenderingControl:1 >> SetMute
\param RTKMR_Token MicroStack token
*/
void RTKMR_Response_RenderingControl_SetMute(const RTKMR_SessionToken RTKMR_Token)
{
    RTKMR_ResponseGeneric(RTKMR_Token,"urn:schemas-upnp-org:service:RenderingControl:1","SetMute","");
}

/*! \fn RTKMR_Response_RenderingControl_SetVolume(const RTKMR_SessionToken RTKMR_Token)
\brief Response Method for RenderingControl >> urn:schemas-upnp-org:service:RenderingControl:1 >> SetVolume
\param RTKMR_Token MicroStack token
*/
void RTKMR_Response_RenderingControl_SetVolume(const RTKMR_SessionToken RTKMR_Token)
{
    RTKMR_ResponseGeneric(RTKMR_Token,"urn:schemas-upnp-org:service:RenderingControl:1","SetVolume","");
}

/*! \fn RTKMR_Response_RenderingControl_SetVolumeDB(const RTKMR_SessionToken RTKMR_Token)
\brief Response Method for RenderingControl >> urn:schemas-upnp-org:service:RenderingControl:1 >> SetVolumeDB
\param RTKMR_Token MicroStack token
*/
void RTKMR_Response_RenderingControl_SetVolumeDB(const RTKMR_SessionToken RTKMR_Token)
{
    RTKMR_ResponseGeneric(RTKMR_Token,"urn:schemas-upnp-org:service:RenderingControl:1","SetVolumeDB","");
}



void RTKMR_SendEventSink(
void *WebReaderToken,
int IsInterrupt,
struct packetheader *header,
char *buffer,
int *p_BeginPointer,
int EndPointer,
int done,
void *subscriber,
void *upnp,
int *PAUSE)	
{
    if(done!=0 && ((struct SubscriberInfo*)subscriber)->Disposing==0)
    {
        sem_wait(&(((struct RTKMR_DataObject*)upnp)->EventLock));
        --((struct SubscriberInfo*)subscriber)->RefCount;
        if(((struct SubscriberInfo*)subscriber)->RefCount==0)
        {
            LVL3DEBUG(printf("\r\n\r\nSubscriber at [%s] %d.%d.%d.%d:%d was/did UNSUBSCRIBE while trying to send event\r\n\r\n",((struct SubscriberInfo*)subscriber)->SID,(((struct SubscriberInfo*)subscriber)->Address&0xFF),((((struct SubscriberInfo*)subscriber)->Address>>8)&0xFF),((((struct SubscriberInfo*)subscriber)->Address>>16)&0xFF),((((struct SubscriberInfo*)subscriber)->Address>>24)&0xFF),((struct SubscriberInfo*)subscriber)->Port);)
            RTKMR_DestructSubscriberInfo(((struct SubscriberInfo*)subscriber));
        }
        else if(header==NULL)
        {
            LVL3DEBUG(printf("\r\n\r\nCould not deliver event for [%s] %d.%d.%d.%d:%d UNSUBSCRIBING\r\n\r\n",((struct SubscriberInfo*)subscriber)->SID,(((struct SubscriberInfo*)subscriber)->Address&0xFF),((((struct SubscriberInfo*)subscriber)->Address>>8)&0xFF),((((struct SubscriberInfo*)subscriber)->Address>>16)&0xFF),((((struct SubscriberInfo*)subscriber)->Address>>24)&0xFF),((struct SubscriberInfo*)subscriber)->Port);)
            // Could not send Event, so unsubscribe the subscriber
            ((struct SubscriberInfo*)subscriber)->Disposing = 1;
            RTKMR_ExpireSubscriberInfo(upnp,subscriber);
        }
        sem_post(&(((struct RTKMR_DataObject*)upnp)->EventLock));
    }
}
void RTKMR_SendEvent_Body(void *upnptoken,char *body,int bodylength,struct SubscriberInfo *info)
{
    struct RTKMR_DataObject* RTKMR_Object = (struct RTKMR_DataObject*)upnptoken;
    struct sockaddr_in dest;
    int packetLength;
    char *packet;
    int ipaddr;
    
    memset(&dest,0,sizeof(dest));
    dest.sin_addr.s_addr = info->Address;
    dest.sin_port = htons(info->Port);
    dest.sin_family = AF_INET;
    ipaddr = info->Address;
    
    packet = (char*)malloc(info->PathLength + bodylength + 483);
    packetLength = sprintf(packet,"NOTIFY %s HTTP/1.1\r\nSERVER: %s, UPnP/1.0, Intel MicroStack/1.0.1868\r\nHOST: %d.%d.%d.%d:%d\r\nContent-Type: text/xml; charset=\"utf-8\"\r\nNT: upnp:event\r\nNTS: upnp:propchange\r\nSID: %s\r\nSEQ: %d\r\nContent-Length: %d\r\n\r\n<?xml version=\"1.0\" encoding=\"utf-8\"?><e:propertyset xmlns:e=\"urn:schemas-upnp-org:event-1-0\"><e:property><%s></e:property></e:propertyset>",info->Path,RTKMR_PLATFORM,(ipaddr&0xFF),((ipaddr>>8)&0xFF),((ipaddr>>16)&0xFF),((ipaddr>>24)&0xFF),info->Port,info->SID,info->SEQ,bodylength+137,body);
    ++info->SEQ;
    
    ++info->RefCount;
    ILibWebClient_PipelineRequestEx(RTKMR_Object->EventClient,&dest,packet,packetLength,0,NULL,0,0,&RTKMR_SendEventSink,info,upnptoken);
}
void RTKMR_SendEvent(void *upnptoken, char* body, const int bodylength, const char* eventname)
{
    struct SubscriberInfo *info = NULL;
    struct RTKMR_DataObject* RTKMR_Object = (struct RTKMR_DataObject*)upnptoken;
    struct sockaddr_in dest;
    LVL3DEBUG(struct timeval tv;)
    
    if(RTKMR_Object==NULL)
    {
        free(body);
        return;
    }
    sem_wait(&(RTKMR_Object->EventLock));
    if(strncmp(eventname,"AVTransport",11)==0)
    {
        info = RTKMR_Object->HeadSubscriberPtr_AVTransport;
    }
    if(strncmp(eventname,"ConnectionManager",17)==0)
    {
        info = RTKMR_Object->HeadSubscriberPtr_ConnectionManager;
    }
    if(strncmp(eventname,"RenderingControl",16)==0)
    {
        info = RTKMR_Object->HeadSubscriberPtr_RenderingControl;
    }
    
    memset(&dest,0,sizeof(dest));
    while(info!=NULL)
    {
        if(!RTKMR_SubscriptionExpired(info))
        {
            RTKMR_SendEvent_Body(upnptoken,body,bodylength,info);
        }
        else
        {
            //Remove Subscriber
            #if defined(WIN32) || defined(_WIN32_WCE)
            LVL3DEBUG(printf("\r\n\r\nTIMESTAMP: %d\r\n",GetTickCount()/1000);)
            #else
            LVL3DEBUG(gettimeofday(&tv,NULL);)
            LVL3DEBUG(printf("\r\n\r\nTIMESTAMP: %d\r\n",tv.tv_sec);)
            #endif
            LVL3DEBUG(printf("Did not renew [%s] %d.%d.%d.%d:%d UNSUBSCRIBING <%d>\r\n\r\n",((struct SubscriberInfo*)info)->SID,(((struct SubscriberInfo*)info)->Address&0xFF),((((struct SubscriberInfo*)info)->Address>>8)&0xFF),((((struct SubscriberInfo*)info)->Address>>16)&0xFF),((((struct SubscriberInfo*)info)->Address>>24)&0xFF),((struct SubscriberInfo*)info)->Port,info);)
        }
        
        info = info->Next;
    }
    
    sem_post(&(RTKMR_Object->EventLock));
}

/*! \fn RTKMR_SetState_AVTransport_LastChange(RTKMR_MicroStackToken upnptoken, char* val)
\brief Sets the state of LastChange << urn:schemas-upnp-org:service:AVTransport:1 << AVTransport \para
\b Note: Must be called at least once prior to start
\param upnptoken The MicroStack token
\param val The new value of the state variable
*/
void RTKMR_SetState_AVTransport_LastChange(RTKMR_MicroStackToken upnptoken, char* val)
{
    struct RTKMR_DataObject *RTKMR_Object = (struct RTKMR_DataObject*)upnptoken;
    char* body;
    int bodylength;
    char* valstr;
    valstr = (char*)malloc(ILibXmlEscapeLength(val)+1);
    ILibXmlEscape(valstr,val);
    if (RTKMR_Object->AVTransport_LastChange != NULL) free(RTKMR_Object->AVTransport_LastChange);
    RTKMR_Object->AVTransport_LastChange = valstr;
    body = (char*)malloc(30 + (int)strlen(valstr));
    bodylength = sprintf(body,"%s>%s</%s","LastChange",valstr,"LastChange");
    RTKMR_SendEvent(upnptoken,body,bodylength,"AVTransport");
    free(body);
}

/*! \fn RTKMR_SetState_ConnectionManager_SourceProtocolInfo(RTKMR_MicroStackToken upnptoken, char* val)
\brief Sets the state of SourceProtocolInfo << urn:schemas-upnp-org:service:ConnectionManager:1 << ConnectionManager \para
\b Note: Must be called at least once prior to start
\param upnptoken The MicroStack token
\param val The new value of the state variable
*/
void RTKMR_SetState_ConnectionManager_SourceProtocolInfo(RTKMR_MicroStackToken upnptoken, char* val)
{
    struct RTKMR_DataObject *RTKMR_Object = (struct RTKMR_DataObject*)upnptoken;
    char* body;
    int bodylength;
    char* valstr;
    valstr = (char*)malloc(ILibXmlEscapeLength(val)+1);
    ILibXmlEscape(valstr,val);
    if (RTKMR_Object->ConnectionManager_SourceProtocolInfo != NULL) free(RTKMR_Object->ConnectionManager_SourceProtocolInfo);
    RTKMR_Object->ConnectionManager_SourceProtocolInfo = valstr;
    body = (char*)malloc(46 + (int)strlen(valstr));
    bodylength = sprintf(body,"%s>%s</%s","SourceProtocolInfo",valstr,"SourceProtocolInfo");
    RTKMR_SendEvent(upnptoken,body,bodylength,"ConnectionManager");
    free(body);
}

/*! \fn RTKMR_SetState_ConnectionManager_SinkProtocolInfo(RTKMR_MicroStackToken upnptoken, char* val)
\brief Sets the state of SinkProtocolInfo << urn:schemas-upnp-org:service:ConnectionManager:1 << ConnectionManager \para
\b Note: Must be called at least once prior to start
\param upnptoken The MicroStack token
\param val The new value of the state variable
*/
void RTKMR_SetState_ConnectionManager_SinkProtocolInfo(RTKMR_MicroStackToken upnptoken, char* val)
{
    struct RTKMR_DataObject *RTKMR_Object = (struct RTKMR_DataObject*)upnptoken;
    char* body;
    int bodylength;
    char* valstr;
    valstr = (char*)malloc(ILibXmlEscapeLength(val)+1);
    ILibXmlEscape(valstr,val);
    if (RTKMR_Object->ConnectionManager_SinkProtocolInfo != NULL) free(RTKMR_Object->ConnectionManager_SinkProtocolInfo);
    RTKMR_Object->ConnectionManager_SinkProtocolInfo = valstr;
    body = (char*)malloc(42 + (int)strlen(valstr));
    bodylength = sprintf(body,"%s>%s</%s","SinkProtocolInfo",valstr,"SinkProtocolInfo");
    RTKMR_SendEvent(upnptoken,body,bodylength,"ConnectionManager");
    free(body);
}

/*! \fn RTKMR_SetState_ConnectionManager_CurrentConnectionIDs(RTKMR_MicroStackToken upnptoken, char* val)
\brief Sets the state of CurrentConnectionIDs << urn:schemas-upnp-org:service:ConnectionManager:1 << ConnectionManager \para
\b Note: Must be called at least once prior to start
\param upnptoken The MicroStack token
\param val The new value of the state variable
*/
void RTKMR_SetState_ConnectionManager_CurrentConnectionIDs(RTKMR_MicroStackToken upnptoken, char* val)
{
    struct RTKMR_DataObject *RTKMR_Object = (struct RTKMR_DataObject*)upnptoken;
    char* body;
    int bodylength;
    char* valstr;
    valstr = (char*)malloc(ILibXmlEscapeLength(val)+1);
    ILibXmlEscape(valstr,val);
    if (RTKMR_Object->ConnectionManager_CurrentConnectionIDs != NULL) free(RTKMR_Object->ConnectionManager_CurrentConnectionIDs);
    RTKMR_Object->ConnectionManager_CurrentConnectionIDs = valstr;
    body = (char*)malloc(50 + (int)strlen(valstr));
    bodylength = sprintf(body,"%s>%s</%s","CurrentConnectionIDs",valstr,"CurrentConnectionIDs");
    RTKMR_SendEvent(upnptoken,body,bodylength,"ConnectionManager");
    free(body);
}

/*! \fn RTKMR_SetState_RenderingControl_LastChange(RTKMR_MicroStackToken upnptoken, char* val)
\brief Sets the state of LastChange << urn:schemas-upnp-org:service:RenderingControl:1 << RenderingControl \para
\b Note: Must be called at least once prior to start
\param upnptoken The MicroStack token
\param val The new value of the state variable
*/
void RTKMR_SetState_RenderingControl_LastChange(RTKMR_MicroStackToken upnptoken, char* val)
{
    struct RTKMR_DataObject *RTKMR_Object = (struct RTKMR_DataObject*)upnptoken;
    char* body;
    int bodylength;
    char* valstr;
    valstr = (char*)malloc(ILibXmlEscapeLength(val)+1);
    ILibXmlEscape(valstr,val);
    if (RTKMR_Object->RenderingControl_LastChange != NULL) free(RTKMR_Object->RenderingControl_LastChange);
    RTKMR_Object->RenderingControl_LastChange = valstr;
    body = (char*)malloc(30 + (int)strlen(valstr));
    bodylength = sprintf(body,"%s>%s</%s","LastChange",valstr,"LastChange");
    RTKMR_SendEvent(upnptoken,body,bodylength,"RenderingControl");
    free(body);
}



void RTKMR_DestroyMicroStack(void *object)
{
    struct RTKMR_DataObject *upnp = (struct RTKMR_DataObject*)object;
    struct SubscriberInfo  *sinfo,*sinfo2;
    RTKMR_SendByeBye(upnp);
    
    sem_destroy(&(upnp->EventLock));
    
    free(upnp->AVTransport_LastChange);
    free(upnp->ConnectionManager_SourceProtocolInfo);
    free(upnp->ConnectionManager_SinkProtocolInfo);
    free(upnp->ConnectionManager_CurrentConnectionIDs);
    free(upnp->RenderingControl_LastChange);
    
    
    free(upnp->AddressList);
    free(upnp->NOTIFY_SEND_socks);
    free(upnp->UUID);
    free(upnp->Serial);
    free(upnp->DeviceDescription);
    
    sinfo = upnp->HeadSubscriberPtr_AVTransport;
    while(sinfo!=NULL)
    {
        sinfo2 = sinfo->Next;
        RTKMR_DestructSubscriberInfo(sinfo);
        sinfo = sinfo2;
    }
    sinfo = upnp->HeadSubscriberPtr_ConnectionManager;
    while(sinfo!=NULL)
    {
        sinfo2 = sinfo->Next;
        RTKMR_DestructSubscriberInfo(sinfo);
        sinfo = sinfo2;
    }
    sinfo = upnp->HeadSubscriberPtr_RenderingControl;
    while(sinfo!=NULL)
    {
        sinfo2 = sinfo->Next;
        RTKMR_DestructSubscriberInfo(sinfo);
        sinfo = sinfo2;
    }
    
    
    #if defined(WIN32) || defined(_WIN32_WCE)
    WSACleanup();
    #endif
}
int RTKMR_GetLocalPortNumber(RTKMR_SessionToken token)
{
    return(ILibWebServer_GetPortNumber(((struct ILibWebServer_Session*)token)->Parent));
}
void RTKMR_SessionReceiveSink(
struct ILibWebServer_Session *sender,
int InterruptFlag,
struct packetheader *header,
char *bodyBuffer,
int *beginPointer,
int endPointer,
int done)
{
    if(header!=NULL && done !=0 && InterruptFlag==0)
    {
        RTKMR_ProcessHTTPPacket(sender,header,bodyBuffer,beginPointer==NULL?0:*beginPointer,endPointer);
        if(beginPointer!=NULL) {*beginPointer = endPointer;}
    }
}
void RTKMR_SessionSink(struct ILibWebServer_Session *SessionToken, void *user)
{
    SessionToken->OnReceive = &RTKMR_SessionReceiveSink;
    SessionToken->User = user;
}
RTKMR_MicroStackToken RTKMR_CreateMicroStack(void *Chain, const char* FriendlyName,const char* UDN, const char* SerialNumber, const int NotifyCycleSeconds, const unsigned short PortNum)

{
    struct RTKMR_DataObject* RetVal = (struct RTKMR_DataObject*)malloc(sizeof(struct RTKMR_DataObject));
    char* DDT;
    
    #if defined(WIN32) || defined(_WIN32_WCE)
    WORD wVersionRequested;
    WSADATA wsaData;
    srand((int)GetTickCount());
    #ifdef WINSOCK1
    wVersionRequested = MAKEWORD( 1, 1 );	
    #elif WINSOCK2
    wVersionRequested = MAKEWORD( 2, 0 );
    #endif
    if (WSAStartup( wVersionRequested, &wsaData ) != 0) {exit(1);}
    #else
    struct timeval tv;
    gettimeofday(&tv,NULL);
    srand((int)tv.tv_sec);
    #endif
    
    RTKMR_Init(RetVal,NotifyCycleSeconds,PortNum);
    RetVal->ForceExit = 0;
    RetVal->PreSelect = &RTKMR_MasterPreSelect;
    RetVal->PostSelect = &RTKMR_MasterPostSelect;
    RetVal->Destroy = &RTKMR_DestroyMicroStack;
    RetVal->InitialNotify = 0;
    if (UDN != NULL)
    {
        RetVal->UUID = (char*)malloc((int)strlen(UDN)+6);
        sprintf(RetVal->UUID,"uuid:%s",UDN);
        RetVal->UDN = RetVal->UUID + 5;
    }
    if (SerialNumber != NULL)
    {
        RetVal->Serial = (char*)malloc((int)strlen(SerialNumber)+1);
        strcpy(RetVal->Serial,SerialNumber);
    }
    
    RetVal->DeviceDescription = (char*)malloc(10+RTKMR_DeviceDescriptionTemplateLengthUX+ (int)strlen(FriendlyName)  + (((int)strlen(RetVal->Serial) + (int)strlen(RetVal->UUID)) * 1));
    
    
    RetVal->WebServerTimer = ILibCreateLifeTime(Chain);
    
    RetVal->HTTPServer = ILibWebServer_Create(Chain,UPNP_HTTP_MAXSOCKETS,PortNum,&RTKMR_SessionSink,RetVal);
    RetVal->WebSocketPortNumber=(int)ILibWebServer_GetPortNumber(RetVal->HTTPServer);
    
    
    
    ILibAddToChain(Chain,RetVal);
    RetVal->EventClient = ILibCreateWebClient(5,Chain);
    RetVal->Chain = Chain;
    RetVal->UpdateFlag = 0;
    
    DDT = ILibDecompressString((char*)RTKMR_DeviceDescriptionTemplate,RTKMR_DeviceDescriptionTemplateLength,RTKMR_DeviceDescriptionTemplateLengthUX);
    RetVal->DeviceDescriptionLength = sprintf(RetVal->DeviceDescription,DDT,FriendlyName,RetVal->Serial,RetVal->UDN);
    
    free(DDT);

    RetVal->AVTransport_LastChange = malloc(sizeof(char *));
    *(RetVal->AVTransport_LastChange) = ""; 
    RetVal->ConnectionManager_SourceProtocolInfo = malloc(sizeof(char *));
    *(RetVal->ConnectionManager_SourceProtocolInfo) = "";
    RetVal->ConnectionManager_SinkProtocolInfo = malloc(sizeof(char *));
    *(RetVal->ConnectionManager_SinkProtocolInfo) = "";
    RetVal->ConnectionManager_CurrentConnectionIDs = malloc(sizeof(char *));
    *(RetVal->ConnectionManager_CurrentConnectionIDs) = "";
    RetVal->RenderingControl_LastChange = malloc(sizeof(char *));
    *(RetVal->RenderingControl_LastChange) = "";
    
    sem_init(&(RetVal->EventLock),0,1);
    return(RetVal);
}











